Package: ARPALData
Type: Package
Title: Retrieving and Analyzing Air Quality and Weather Data from ARPA
        Lombardia
Version: 2.0.1
Description: Contains functions for retrieving, managing, and analyzing air quality and weather data from the Regione Lombardia open database (<https://www.dati.lombardia.it/>).
    Data are collected by ARPA Lombardia (Lombardia Environmental Protection Agency), Italy, through its ground monitoring network (<https://www.dati.lombardia.it/stories/s/auv9-c2sj>).
    See the website <https://www.arpalombardia.it/> for further information on ARPA Lombardia's activities and history.
    Data quality (e.g., missing values, extreme values, and graphical mapping) has been checked in collaboration with members of ARPA Lombardia's air quality control office.
    The package provides observations since 1989 (for weather) and 1968 (for air quality), and these data are updated daily by the regional agency.
    A full description of the package is available in the companion paper Maranzano \& Algieri (2024), "ARPALData: an R package for retrieving and analyzing air quality
    and weather data from ARPA Lombardia (Italy)", Environmental and Ecological Statistics, <doi:10.1007/s10651-024-00599-6>.
License: GPL (>= 2)
Imports:
        tidyselect,dplyr,utils,rlang,lubridate,readr,stringr,tm,tidyr,eurostat,sf,ggplot2,tibble,aweek,curl,future,future.apply,httr2,jsonlite
Suggests: tidyverse, knitr, rmarkdown
Encoding: UTF-8
Language: en-US
RoxygenNote: 7.3.3
Authors@R: 
    c(person(given = "Paolo",
             family = "Maranzano",
             role = c("aut", "cre", "cph"),
             email = "pmaranzano.ricercastatistica@gmail.com",
             comment = c(ORCID = "0000-0002-9228-2759")),
      person(given = "Andrea",
             family = "Algieri",
             role = c("aut", "cph"),
             email = "a.algieri@arpalombardia.it"))
NeedsCompilation: no
Packaged: 2026-07-16 21:07:03 UTC; paulm
Author: Paolo Maranzano [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0002-9228-2759>),
  Andrea Algieri [aut, cph]
Maintainer: Paolo Maranzano <pmaranzano.ricercastatistica@gmail.com>
Repository: CRAN
Date/Publication: 2026-07-16 21:40:07 UTC
Built: R 4.5.2; ; 2026-07-16 22:42:38 UTC; unix

export("get_ARPA_Lombardia_AQ_data","get_ARPA_Lombardia_W_data","get_ARPA_Lombardia_AQ_municipal_data",
       "get_ARPA_Lombardia_AQ_registry","get_ARPA_Lombardia_W_registry","get_ARPA_Lombardia_AQ_municipal_registry",
       "get_ARPA_Lombardia_zoning","get_Lombardia_geospatial",
       "is_ARPALdf", "is_ARPALdf_AQ", "is_ARPALdf_AQ_mun", "is_ARPALdf_W",
       "map_Lombardia_stations", "Time_aggregate","ARPALdf_Summary","ARPALdf_Summary_map",
       "registry_KNN_dist")
import("ggplot2")
importFrom("dplyr", "filter", "bind_rows", "case_when","recode","select")
importFrom("tidyselect", "where","contains","all_of","matches")
importFrom("utils", "data", "download.file","unzip")
importFrom("tidyr", "%>%")
importFrom("rlang", ".data", "check_installed","is_installed")
importFrom("lubridate", "ymd", "dmy_hms", "ymd_hms","interval","time_length","as_datetime","make_date","round_date","days_in_month","month","dmonths")
importFrom("stringr","str_to_title","str_replace_all","str_split","str_extract")
importFrom("readr", "read_csv","cols")
importFrom("eurostat", "get_eurostat_geospatial")
importFrom("sf", "st_as_sf")
importFrom("aweek","week2date","date2week")
importFrom("tibble","tibble","rownames_to_column")
importFrom("stats", "mad", "median", "sd","cor","lag","quantile")
importFrom("tm", "removePunctuation", "removeNumbers", "stripWhitespace")
importFrom("curl", "curl_fetch_disk","has_internet")
importFrom("future","plan","multisession","multicore","availableCores")
importFrom("future.apply","future_apply")

### Modified on 2026-06-15: import the new JSON/SODA download dependencies.
importFrom("httr2", "request", "req_url_query", "req_error", "req_headers", "req_perform", "resp_status", "resp_body_string")
importFrom("jsonlite", "fromJSON")

#' Download weather/meteorological data from ARPA Lombardia website
#'
#' More detailed description.
#'
#' @description 'get_ARPA_Lombardia_W_data' returns observed air weather measurements collected by
#' ARPA Lombardia ground detection system for Lombardy region in Northern Italy.
#' Available meteorological variables are: temperature (Celsius degrees), rainfall (mm), wind speed (m/s),
#' wind direction (degrees), relative humidity (%), global solar radiation (W/m2), and snow height (cm).
#' Data are available from 1989 and are updated up to the current date.
#' For more information, visit the section 'Idro-Nivo-Meteo' at the website:
#' https://www.dati.lombardia.it/stories/s/auv9-c2sj
#'
#' @param ID_station Numeric value. ID of the station to consider. Using ID_station = NULL, all the available
#' stations are selected. Default is ID_station = NULL.
#' @param Date_begin Character vector of the first date-time to download. Format can be either "YYYY-MM-DD" or "YYYY-MM-DD hh:mm:ss". Default is Date_begin = "2022-01-01".
#' @param Date_end Character vector of the last date-time to download. Format can be either "YYYY-MM-DD" or "YYYY-MM-DD hh:mm:ss". Default is Date_end = "2022-12-31".
#' @param Frequency Temporal aggregation frequency. It can be "10mins", "hourly", "daily", "weekly",
#' "monthly". Default is Frequency = "10mins"
#' @param Var_vec Character vector of variables to aggregate. If NULL (default) all the variables are averaged,
#' except for 'Temperature' and 'Snow_height', which are cumulated.
#' @param Fns_vec Character vector of aggregation function to apply to the selected variables.
#' Available functions are mean, median, min, max, sum, qPP (PP-th percentile), sd, var,
#' vc (variability coefficient), skew (skewness) and kurt (kurtosis). Attention: for Wind Speed and
#' Wind Speed Gust only mean, min and max are available; for Wind Direction and Wind Direction Gust
#' only mean is available.
#' @param by_sensor Logical value (TRUE or FALSE). If 'by_sensor = TRUE', the function returns the observed concentrations
#' by sensor code, while if 'by_sensor = FALSE' (default) it returns the observed concentrations by station.
#' @param verbose Logical value (TRUE or FALSE). Toggle warnings and messages. If 'verbose = TRUE' (default) the function
#' prints on the screen some messages describing the progress of the tasks. If 'verbose = FALSE' any message about
#' the progression is suppressed.
#' @param parallel Logical value (TRUE or FALSE). If 'parallel = FALSE' (default), data downloading is performed using a sequential/serial approach and additional parameters 'parworkers' and 'parfuturetype' are ignored.
#' When 'parallel = TRUE', data downloading is performed using parallel computing through the Futureverse setting.
#' More detailed information about parallel computing in the Futureverse is available at the following websites:
#' https://future.futureverse.org/ and https://cran.r-project.org/web/packages/future.apply/vignettes/future.apply-1-overview.html
#' @param parworkers Numeric integer value. If 'parallel = TRUE' (parallel mode active), the user can declare the number of parallel workers to be activated using 'parworkers = integer number'. By default ('parworkers = NULL'), the number of active workers is half of the available local cores.
#' @param parfuturetype Character vector. If 'parallel = TRUE' (parallel mode active), the user can declare the parallel strategy to be used according to the Futureverse syntax through 'parfuturetype'. By default, the 'multisession' (background R sessions on local machine) is used. In alternative, the 'multicore' (forked R processes on local machine. Not supported by Windows and RStudio) setting can be used.
#'
#'
#' @return A data frame of class 'data.frame' and 'ARPALdf'. The object is fully compatible with Tidyverse.
#'
#' @examples
#' \donttest{
#' ## Download all the (10 minutes frequency) weather measurements at station 100
#' ## between August 2022 and April 2024.
#' get_ARPA_Lombardia_W_data(ID_station = 100, Date_begin = "2022-08-01",
#'                           Date_end = "2024-04-30", Frequency = "10mins")
#' ## Download all the (daily frequency) weather measurements at station 1974 during 2025 and 2026
#' get_ARPA_Lombardia_W_data(ID_station = 1974, Date_begin = "2025-01-01",
#'                           Date_end = "2026-12-31", Frequency = "daily")
#' }
#'
#' @export

get_ARPA_Lombardia_W_data <-
  function(ID_station = NULL, Date_begin = "2021-01-01", Date_end = "2022-12-31",
           Frequency = "10mins", Var_vec = NULL, Fns_vec = NULL, by_sensor = FALSE, verbose = TRUE,
           parallel = FALSE, parworkers = NULL, parfuturetype = "multisession") {

    ###  Welcome message
    if (verbose == TRUE) {
      cat("Retrieving desired ARPA Lombardia dataset: started at", as.character(Sys.time()), "\n")
    }

    ############################
    ########## Checks ##########
    ############################

    ##### Check for internet connection
    if(!curl::has_internet()) {
      message("Internet connection not available at the moment.\nPlease check your internet connection. If the problem persists, please contact the package maintainer.")
      return(invisible(NULL))
    }

    ##### Check if packages for JSON/SODA download are installed
    ### Modified on 2026-06-15: replace the socratadata dependency check with httr2/jsonlite checks.
    rlang::check_installed("httr2",
                           reason = "Package \"httr2\" must be installed to download data from ARPA Lombardia Open Database.")
    rlang::check_installed("jsonlite",
                           reason = "Package \"jsonlite\" must be installed to parse JSON data from ARPA Lombardia Open Database.")

    ##### Check if Futureverse is installed
    rlang::check_installed("future",
                           reason = "Package \"future\" must be installed to download (parallel) data from ARPA Lombardia Open Database.")
    rlang::check_installed("future.apply",
                           reason = "Package \"future\" must be installed to download (parallel) data from ARPA Lombardia Open Database.")

    ##### Define %notin%
    '%notin%' <- Negate('%in%')

    ##### Checks if by_sensor setup properly
    if (by_sensor %notin% c(0,1,FALSE,TRUE)) {
      stop("Wrong setup for 'by_sensor'. Use 1 or 0 or TRUE or FALSE.",
           call. = FALSE)
    }

    ##### Checks if parallel setup properly
    if (parallel %notin% c(FALSE,TRUE)) {
      stop("Wrong setup for 'parallel'. Use TRUE or FALSE.",
           call. = FALSE)
    }
    if (parfuturetype %notin% c("multisession","multicore")) {
      stop("Wrong setup for 'parallel'. Use TRUE or FALSE.",
           call. = FALSE)
    }

    ##### Control checks on Var_vec and Fns_vec
    # Wind_speed and Wind_direction must be selected together
    if (!is.na(match("Wind_speed",Var_vec)) & is.na(match("Wind_direction",Var_vec))) {
      stop("It's not possible to select only Wind_speed: also Wind_direction must be included in the list of selected variables! Change the values of 'Var_vec' and 'Var_funs'",
           call. = FALSE)
    }
    if (is.na(match("Wind_speed",Var_vec)) & !is.na(match("Wind_direction",Var_vec))) {
      stop("It's not possible to select only Wind_direction: also Wind_speed must be included in the list of selected variables! Change the values of 'Var_vec' and 'Var_funs'",
           call. = FALSE)
    }
    # Wind direction can only be averaged
    if(sum(Var_vec %in% c("Wind_direction","Wind_direction_gust") & Fns_vec != "mean") > 0) {
      stop("Error: on Wind_direction and Wind_direction_gust is possible to calculate only the average value. Use 'mean' in 'Fns_vec.'",
           call. = FALSE)
    }
    # Wind speed can only be averaged, maximized or minimized
    if(sum(Var_vec %in% c("Wind_speed","Wind_speed_gust") & Fns_vec %notin% c("mean","min","max")) > 0) {
      stop("Error: on Wind_speed and Wind_speed_gust is possible to calculate only mean, max or min values. Use 'mean' or 'max' or 'min' in 'Fns_vec.'",
           call. = FALSE)
    }



    ######################################
    ########## Downloading data ##########
    ######################################

    ### Registry
    Metadata <- W_metadata_reshape()
    # Metadata <- Metadata %>%
    #   dplyr::select(-c(.data$Altitude,.data$Province,
    #                    .data$DateStart,.data$DateStop,
    #                    .data$Latitude,.data$Longitude))

    ### Checks if ID_station is valid (in the list of active stations)
    if (!is.null(ID_station) & all(ID_station %notin% Metadata$IDStation)) {
      stop("ID_station NOT in the list of active stations. Change ID_station or use ID_station = NULL",
           call. = FALSE)
    }

    Metadata <- Metadata %>%
      dplyr::filter((.data$DateStop >= Date_begin | is.na(.data$DateStop)) &
                      .data$DateStart <= Date_begin)
    if (!is.null(ID_station)) {
      Metadata <- Metadata %>%
        dplyr::filter(.data$IDStation %in% ID_station)
    }
    if (!is.null(Var_vec)) {
      Metadata <- Metadata %>%
        dplyr::filter(.data$Measure %in% Var_vec)
    }

    ##### Download using Socrata API
    # Attention: this is a correction for issues with adaptive database of ARPA Lombardia
    begin <- Sys.time()
    Today <- Sys.time()
    Past <- Today - lubridate::dmonths(7) - lubridate::ddays(1)
    Past <- lubridate::round_date(Past, "day")
    ### Modified on 2026-06-15: according to the Open Data Lombardia Idro-Nivo-Meteo page,
    ### the current tabular dataset contains the last seven months of weather data.
    ### Earlier requests are therefore routed to the historical parameter-specific datasets,
    ### while only the last-seven-month window is routed to the current unified dataset.
    Break_date_current <- Past
    if (Date_end >= Break_date_current & Date_begin <= Break_date_current) {
      Meteo1 <- W_download_past(Metadata,Date_begin,Break_date_current)
      Meteo2 <- try(
        expr = {
          W_download_current(Metadata,Break_date_current + lubridate::minutes(5),Date_end)
        },silent = TRUE
      )
      if (is.element("try-error", attr(Meteo2,"class"))) {
        Meteo2 <- NULL
      }
      ### Modified on 2026-06-15: removed the obsolete warning about the temporary
      ### unavailability of weather data after 2024-02-01. The download logic now
      ### follows the Open Data Lombardia Idro-Nivo-Meteo structure: historical
      ### parameter-specific datasets are used before the current seven-month
      ### tabular dataset window, while recent data are queried from the current
      ### unified endpoint.
      Meteo <- bind_rows(Meteo1,Meteo2)
    }
    if (Date_end < Break_date_current & Date_begin < Break_date_current) {
      Meteo <- W_download_past(Metadata,Date_begin,Date_end)
    }
    if (Date_end > Break_date_current & Date_begin > Break_date_current) {
      Meteo <- W_download_current(Metadata,Date_begin,Date_end,Date_end)
    }
    end <- Sys.time()
    end - begin





    #####################################
    ########## Processing data ##########
    #####################################
    if (verbose == TRUE) {
      cat("Processing data: started at", as.character(Sys.time()), "\n")
    }
    ### Change variable names
    Meteo <- Meteo %>%
      dplyr::filter(!is.na(.data$data),
                    .data$stato %in% c("VA")) %>%
      dplyr::select(IDSensor = .data$idsensore, Date = .data$data, Value = .data$valore,
                    Status = .data$stato, Operator = .data$operatore) %>%
      dplyr::mutate(
        # Date = lubridate::ymd_hms(.data$Date),
        Value = as.numeric(.data$Value),
        IDSensor = as.character(.data$IDSensor)
      )
    ### Add metadata
    Meteo <- dplyr::right_join(Meteo,Metadata, by = "IDSensor")
    ### Cleaning
    if (by_sensor %in% c(1,TRUE)) {
      Meteo <- Meteo %>%
        dplyr::select(.data$Date,.data$IDStation,.data$NameStation,.data$IDSensor,
                      .data$Measure,.data$Value) %>%
        dplyr::mutate(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric),
                                    ~ dplyr::na_if(.,-9999))) %>%
        dplyr::mutate(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric),
                                    ~ dplyr::na_if(.,NaN)))
    } else if (by_sensor %in% c(0,FALSE)) {
      Meteo <- Meteo %>%
        dplyr::mutate(Measure = dplyr::case_when(.data$Measure == "Wind_direction" & .data$Operator == "Max" ~ "Wind_direction_gust",
                                                 .data$Measure == "Wind_speed" & .data$Operator == "Max" ~ "Wind_speed_gust",
                                                 TRUE ~ as.character(.data$Measure))) %>%
        # dplyr::select(-c(.data$IDSensor, .data$Operator, .data$Status, .data$Unit_meas)) %>%
        dplyr::select(.data$Date,.data$IDStation,.data$NameStation,.data$Measure,.data$Value) %>%
        tidyr::pivot_wider(names_from = .data$Measure, values_from = .data$Value,
                           values_fn = function(x) mean(x,na.rm=T)
                           ) %>%
        dplyr::mutate(dplyr::across(dplyr::matches(c("Wind_direction","Wind_direction_gust")), ~ round(.x,0))) %>%
        dplyr::mutate(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric),
                                    ~ dplyr::na_if(.,-9999))) %>%
        dplyr::mutate(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric),
                                    ~ dplyr::na_if(.,NaN)))
    }
    Meteo[is.na(Meteo)] <- NA
    Meteo[is.nan_df(Meteo)] <- NA

    ### Add dataset attributes
    attr(Meteo, "class") <- c("ARPALdf","ARPALdf_W","tbl_df","tbl","data.frame")

    ### Cleaning
    if (is.null(Var_vec) & is.null(Fns_vec)) {
      vv <- c("Rainfall","Temperature","Relative_humidity","Global_radiation","Water_height",
              "Snow_height","Wind_speed","Wind_speed_gust","Wind_direction")
      vv <- vv[vv %in% names(Meteo)]
      fv <- ifelse(vv == "Rainfall" | vv == "Snow_height", "sum", "mean")
      fv <- ifelse(vv == "Wind_speed_gust", "max", "mean")
    } else {
      vv <- Var_vec
      fv <- Fns_vec
    }

    # Checks if all the variables are available for the selected stations
    if (all(dplyr::all_of(vv) %in% names(Meteo)) == F) {
      stop("One or more variables are not available for the selected stations! Change the values of 'Var_vec'",
           call. = FALSE)
    }

    if (by_sensor %in% c(0,FALSE)) {
      ### Aggregating dataset
      if (Frequency != "10mins") {
        if (verbose==T) {
          cat("Aggregating ARPA Lombardia data: started at", as.character(Sys.time()), "\n")
        }
        Meteo <- Meteo %>%
          Time_aggregate(Frequency = Frequency, Var_vec = Var_vec, Fns_vec = Fns_vec, verbose = verbose) %>%
          dplyr::arrange(.data$NameStation, .data$Date)
      } else {
        Meteo <- Meteo %>%
          dplyr::select(.data$Date,.data$IDStation,.data$NameStation,vv) %>%
          dplyr::arrange(.data$NameStation, .data$Date)
      }

      ### Regularizing dataset: same number of timestamps for each station and variable
      if (verbose == TRUE) {
        cat("Regularizing ARPA data: started at", as.character(Sys.time()), "\n")
      }

      Meteo <- Meteo %>%
        dplyr::arrange(.data$Date) %>%
        dplyr::filter(.data$Date >= Date_begin,
                      .data$Date <= Date_end) %>%
        tidyr::pivot_longer(cols = -c(.data$Date,.data$IDStation,.data$NameStation),
                            names_to = "Measure", values_to = "Value") %>%
        dplyr::mutate(Date = case_when(Frequency %in% c("10mins","hourly") ~ as.character(format(x = .data$Date, format = "%Y-%m-%d %H:%M:%S")),
                                       TRUE ~ as.character(format(x = .data$Date, format = "%Y-%m-%d")))) %>%
        tidyr::pivot_wider(names_from = .data$Date, values_from = .data$Value) %>%
        tidyr::pivot_longer(cols = -c(.data$Measure,.data$IDStation,.data$NameStation),
                            names_to = "Date", values_to = "Value") %>%
        tidyr::pivot_wider(names_from = .data$Measure, values_from = .data$Value)

      if (Frequency %notin% c("10mins","hourly")) {
        Meteo <- Meteo %>%
          dplyr::mutate(Date = lubridate::ymd(.data$Date)) %>%
          dplyr::arrange(.data$IDStation,.data$Date)
      } else {
        Meteo <- Meteo %>%
          dplyr::mutate(Date = lubridate::ymd_hms(.data$Date)) %>%
          dplyr::arrange(.data$IDStation,.data$Date)
      }

      freq_unit <- dplyr::case_when(Frequency == "10mins" ~ "10 min",
                                    Frequency == "hourly" ~ "hours",
                                    Frequency == "daily" ~ "days",
                                    Frequency == "weekly" ~ "weeks",
                                    Frequency == "monthly" ~ "months")

      structure(list(Meteo = Meteo))
      attr(Meteo, "class") <- c("ARPALdf","ARPALdf_W","tbl_df","tbl","data.frame")
      attr(Meteo, "frequency") <- Frequency
      attr(Meteo, "units") <- freq_unit
    } else if (by_sensor %in% c(1,TRUE)) {
      Meteo <- Meteo %>%
        dplyr::arrange(.data$Date) %>%
        dplyr::filter(.data$Date >= Date_begin,
                      .data$Date <= Date_end)

      structure(list(Meteo = Meteo))
      attr(Meteo, "class") <- c("ARPALdf","ARPALdf_W","tbl_df","tbl","data.frame")
      attr(Meteo, "frequency") <- "10mins"
      attr(Meteo, "units") <- "10 min"
    }

    if (verbose == TRUE) {
      cat("Processing data: ended at", as.character(Sys.time()), "\n")
    }

    if (verbose == TRUE) {
      cat("Retrieving desired ARPA Lombardia dataset: ended at", as.character(Sys.time()), "\n")
    }

    invisible(gc())
    return(Meteo)

  }

#' @keywords internal
#' @noRd

as_difftime <- function (timeseries, format = "%X", units = "auto") {
  if (units %in% c("months", "years")) {
    if (!is.numeric(timeseries)) {
      stop("'timeseries' must be numeric for units = '", units, "'", sep = "")
    } else {
      structure(timeseries, units = units, class = "difftime")
    }
  } else base::as.difftime(timeseries, format, units)
}

#' Download geospatial data of Lombardy from Eurostat
#'
#' @description 'get_Lombardia_geospatial' returns the polygonal (shape file) object containing the geometries
#' of Lombardy. Shapefile are available at different NUTS levels (https://ec.europa.eu/eurostat/web/nuts/background):
#' 'LAU' for the shapefile of municipalities of Lombardy, 'NUTS3' for the shapefile of provinces of Lombardy
#' and 'NUTS2' for the shapefile of Lombardy.
#'
#' @param NUTS_level NUTS level required: use "NUTS2" for regional geometries, "NUTS3" for provincial geometries,
#' or "LAU" for municipal geometries. Default NUTS_level = "LAU".
#'
#' @return A data frame of class 'data.frame', "sf" and 'ARPALdf'.
#'
#' @examples
#' \donttest{
#' shape <- get_Lombardia_geospatial(NUTS_level = "LAU")
#' }
#'
#' @export

get_Lombardia_geospatial <- function(NUTS_level = "LAU") {

  '%notin%' <- Negate('%in%')

  ##### Checks
  if (NUTS_level %notin% c("NUTS2","NUTS3","LAU")) {
    stop("Selected NUTS not available: use one of 'NUTS2', 'NUTS3' and 'LAU'",call. = FALSE)
  }

  ##### Check for internet connection
  if(!curl::has_internet()) {
    message("Internet connection not available at the moment.\nPlease check your internet connection. If the problem persists, please contact the package maintainer.")
    return(invisible(NULL))
  }

  ##### Read Lombardy municipality shapefile bundled with the package
  ### Modified on 2026-06-26: use the package-bundled Lombardy municipality
  ### shapefile stored in inst/extdata, instead of checking and downloading
  ### the same zip archive from the ARPALData GitHub repository at runtime.
  temp1 <- ARPAL_extdata_path("Shape_Comuni_Lombardia.zip")
  temp2 <- tempfile()
  unzip(zipfile = temp1, exdir = temp2)
  your_SHP_file<-list.files(temp2, pattern = ".shp$",full.names=TRUE)

  # Read the shapefile
  Lombardia <- sf::read_sf(your_SHP_file) %>%
    sf::st_as_sf(crs = 4326) %>%
    dplyr::rename(Prov_name = .data$NOME_PRO,City = .data$NOME_COM, City_code_ISTAT = .data$ISTAT) %>%
    dplyr::mutate(Prov_name = case_when(.data$Prov_name == "VARESE" ~ "Varese",
                                        .data$Prov_name == "COMO" ~ "Como",
                                        .data$Prov_name == "LECCO" ~ "Lecco",
                                        .data$Prov_name == "SONDRIO" ~ "Sondrio",
                                        .data$Prov_name == "BERGAMO" ~ "Bergamo",
                                        .data$Prov_name == "BRESCIA" ~ "Brescia",
                                        .data$Prov_name == "PAVIA" ~ "Pavia",
                                        .data$Prov_name == "LODI" ~ "Lodi",
                                        .data$Prov_name == "CREMONA" ~ "Cremona",
                                        .data$Prov_name == "MANTOVA" ~ "Mantova",
                                        .data$Prov_name == "MILANO" ~ "Milano",
                                        .data$Prov_name == "MONZA E DELLA BRIANZA" ~ "Monza e della Brianza")) %>%
    dplyr::select(.data$Prov_name, .data$City, .data$City_code_ISTAT, .data$geometry) %>%
    dplyr::mutate(dplyr::across(c(.data$City), toupper),
                  dplyr::across(c(.data$City), ~ gsub("\\-", " ", .x)),
                  dplyr::across(c(.data$City), ~ stringr::str_replace_all(.x, c("S\\."="San ","s\\."="San ",
                                                                                "V\\."="Via ","v\\."="Via ",
                                                                                " D\\`" = " D\\' ", " D\\` " = " D\\'",
                                                                                "D\\`" = " D\\'", "D\\'" = " D\\' "))),
                  dplyr::across(c(.data$City), tm::removePunctuation),
                  dplyr::across(c(.data$City), tm::removeNumbers),
                  dplyr::across(c(.data$City), tm::stripWhitespace)) %>%
    dplyr::mutate(City = dplyr::recode(.data$City,
                                       "CASASCO DINTELVI" = "CASASCO INTELVI",
                                       "CERANO DINTELVI" = "CERANO INTELVI",
                                       "SAN GIORGIO BIGARELLO" = "BIGARELLO",
                                       "PUEGNAGO DEL GARDA" = "PUEGNAGO SUL GARDA",
                                       "FELONICA" = "SERMIDE E FELONICA",
                                       "GERRE DE CAPRIOLI" = "GERRE DECAPRIOLI")) %>%
    dplyr::mutate(dplyr::across(c(.data$City), stringr::str_to_title),
                  dplyr::across(c(.data$City), ~ stringr::str_replace_all(.x,
                                                                          c(" D " = " D\\'",
                                                                            "Sermide E Felonica" = "Sermide e Felonica"))))

  # Associating NUTS codes (from Eurostat) to each observation
  Eurostat <- eurostat::get_eurostat_geospatial(output_class = "sf",resolution = 60, nuts_level = 3, year = 2016)
  Eurostat <- Eurostat %>%
    sf::st_as_sf(crs = 4326) %>%
    data.frame() %>%
    dplyr::filter(grepl("ITC4",.data$geo)) %>%
    dplyr::select(Prov_name = .data$NUTS_NAME, Prov_code_EUROSTAT = .data$NUTS_ID) %>%
    dplyr::mutate(Reg_name = "Lombardia", Reg_code_EUROSTAT = "ITC4")

  # Joining ISTAT metadata and Eurostat metadata
  Lombardia <- dplyr::left_join(Lombardia,Eurostat,by=c("Prov_name"))
  Lombardia <- Lombardia %>%
    dplyr::select(.data$Reg_name,.data$Reg_code_EUROSTAT,.data$Prov_name,.data$Prov_code_EUROSTAT,
                  .data$City,.data$City_code_ISTAT,.data$geometry)

  if (NUTS_level == "NUTS3") {
    Lombardia <- Lombardia %>%
      dplyr::group_by(.data$Reg_name,.data$Reg_code_EUROSTAT,.data$Prov_name,.data$Prov_code_EUROSTAT) %>%
      dplyr::summarise(.groups = "keep") %>%
      dplyr::ungroup()
  }

  if (NUTS_level == "NUTS2") {
    Lombardia <- Lombardia %>%
      dplyr::group_by(.data$Reg_name,.data$Reg_code_EUROSTAT) %>%
      dplyr::summarise(.groups = "keep") %>%
      dplyr::ungroup()
  }

  file.remove(your_SHP_file)

  attr(Lombardia, "class") <- c("ARPALdf","ARPALdf_AQ","tbl_df","tbl","data.frame","sf")

  return(Lombardia)
}

#' Summary statistics for a data frame of class 'ARPALdf'
#'
#' @description 'ARPALdf_Summary' returns many descriptive statistics summaring the data contained in a data frame
#' of class ARPALdf. Statistics are calculated at overall level (full sample), by station ID and by year.
#' For each variable are reported the basic positioning indices (min, max, mean, median, quantile) and
#' variability indices (range, standard deviation). Other reported statistics are the Pearson's linear correlation
#' by station and some graphical representation of the distribution (kernel density plot, histogram, boxplot).
#' In addition, the function returns useful data-quality information, such as gap length statistics (i.e. number of
#' missing observations for each variable by station and by year) and outlier detection tools
#' (e.g., Hampel filter and boxplot rule)
#'
#' @param Data Dataset of class 'ARPALdf' containing the data to be summarised.
#' @param by_IDStat Logical value (TRUE or FALSE). Use TRUE (default) to compute summary statistics by Station ID.
#' @param by_Year Logical value (TRUE or FALSE). Use TRUE (default) to compute summary statistics by year.
#' @param gap_length Logical value (TRUE or FALSE). Use TRUE (default) to compute summary statistics for the gap length of each variable.
#' @param correlation Logical value (TRUE or FALSE). Use TRUE (default) to compute linear correlation of available variables.
#' @param histogram Logical value (TRUE or FALSE). Use TRUE to plot the histogram of each variable. Default is FALSE.
#' @param density Logical value (TRUE or FALSE). Use TRUE to plot the kernel density plot of each variable. Default is FALSE.
#' @param outlier Logical value (TRUE or FALSE). Use TRUE to analyze extreme values of each variable
#' (boxplot and Hampel filter). Default is FALSE.
#' @param verbose Logical value (TRUE or FALSE). Toggle warnings and messages. If 'verbose = TRUE' (default) the function
#' prints on the screen some messages describing the progress of the tasks. If 'verbose = FALSE' any message about
#' the progression is suppressed.
#'
#' @return A list of data.frames containing summary descriptive statistics for a data frame of class 'ARPALdf'.
#' Summary statistics are computed for the overall sample (Descr), by Station ID (Descr_by_IDStat) and by
#' year (Descr_by_Year). Available statistics are: number of NAs, % of NAs over the total sample, number of null values,
#' number of negative values, minimum, mean, maximum and standard deviation.
#'
#' @examples
#' \donttest{
#' ## Download daily air quality data from all the stations for January 2020
#' d <- get_ARPA_Lombardia_AQ_data(ID_station = NULL, Date_begin = "2020-01-01",
#'         Date_end = "2020-01-31", Frequency = "daily")
#' ## Summarising observed data
#' sum_stats <- ARPALdf_Summary(Data = d)
#' }
#'
#' @export

ARPALdf_Summary <- function(Data, by_IDStat = TRUE, by_Year = TRUE, gap_length = TRUE, correlation = TRUE,
                            histogram = FALSE, density = FALSE, outlier = FALSE, verbose=TRUE) {

  ### Checks
  stopifnot("histogram must be TRUE or FALSE" = histogram == FALSE | histogram == TRUE)
  stopifnot("density must be TRUE or FALSE" = density == FALSE | density == TRUE)
  stopifnot("by_IDStat must be TRUE or FALSE" = by_IDStat == FALSE | by_IDStat == TRUE)
  stopifnot("by_Year must be TRUE or FALSE" = by_Year == FALSE | by_Year == TRUE)
  stopifnot("gap_length must be TRUE or FALSE" = gap_length == FALSE | gap_length == TRUE)
  stopifnot("Data is not of class 'ARPALdf'" = is_ARPALdf(Data = Data) == TRUE)

  ### Print message about dimensions
  n <- dim(Data)[1]
  m <- length(unique(Data$IDStation))
  t <- length(unique(Data$Date))
  t1 <- min(Data$Date)
  t2 <- max(Data$Date)
  if (verbose==TRUE) {
    cat("The dataset contains: \n")
    cat(paste0("   ** ",n," total observations \n"))
    cat(paste0("   ** ",m," stations/ground sites \n"))
    cat(paste0("   ** ",t," time stamps from ",t1," to ",t2,"\n"))
    cat("Inspect this object to obtain summary statistics: \n")
    cat("   ** on the whole sample (Descr) \n")
    if (by_IDStat == TRUE) {
      cat("   ** by Station ID (Descr_by_IDStat) \n")
      cat("Attention! NA values of the statistics indicate that data don't exist for a specific Station \n")}
    if (by_Year == TRUE) {
      cat("   ** by Year (Descr_by_year) \n")
      cat("Attention! NA values of the statistics indicate that data don't exist for a specific Year \n")}
  }

  ### Fix IDStation to integer
  Data <- Data %>%
    dplyr::mutate(IDStation = as.integer(.data$IDStation))

  ### Overall statistics
  NA_count <- Data %>%
    dplyr::summarise(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric) &
                                     !tidyselect::vars_select_helpers$where(is.integer), ~ sum(is.na(.x))))
  NA_count_perc <- Data %>%
    dplyr::summarise(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric) &
                                     !tidyselect::vars_select_helpers$where(is.integer),
                                   ~ round(sum(is.na(.x))/dim(Data)[1]*100,2)))
  mean_vals <- Data %>%
    dplyr::summarise(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric) &
                                     !tidyselect::vars_select_helpers$where(is.integer), ~ round(mean(.x,na.rm=T),2)))
  sd_vals <- Data %>%
    dplyr::summarise(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric) &
                                     !tidyselect::vars_select_helpers$where(is.integer), ~ round(sd(.x,na.rm=T),2)))
  min_vals <- Data %>%
    dplyr::summarise(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric) &
                                     !tidyselect::vars_select_helpers$where(is.integer), ~ round(min(.x,na.rm=T),2)))
  max_vals <- Data %>%
    dplyr::summarise(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric) &
                                     !tidyselect::vars_select_helpers$where(is.integer), ~ round(max(.x,na.rm=T),2)))
  null_vals <- Data %>%
    dplyr::summarise(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric) &
                                     !tidyselect::vars_select_helpers$where(is.integer), ~ sum(.x == 0, na.rm=T)))
  neg_vals <- Data %>%
    dplyr::summarise(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric) &
                                     !tidyselect::vars_select_helpers$where(is.integer), ~ sum(.x < 0, na.rm=T)))

  descriptives <- dplyr::bind_rows(NA_count,NA_count_perc,null_vals,neg_vals,min_vals,mean_vals,max_vals,sd_vals)
  descriptives <- t(descriptives)
  descriptives <- data.frame(descriptives) %>%
    tibble::rownames_to_column()
  colnames(descriptives) <- c("Var","NA_count","NA_perc","Null_count","Negative_count",
                              "Min","Mean","Max","Std.Dev.")
  if (is_ARPALdf_AQ(Data = Data) == T) {
    attr(descriptives, "class") <- c("ARPALdf","ARPALdf_AQ","tbl_df","tbl","data.frame")
  } else if (is_ARPALdf_W(Data = Data) == T) {
    attr(descriptives, "class") <- c("ARPALdf","ARPALdf_W","tbl_df","tbl","data.frame")
  }


  ### Statistics by Station ID
  if (by_IDStat == TRUE) {
    if (verbose==TRUE) {
      cat("Computing summary statistics by Station ID \n")
    }
    NA_count_stat <- Data %>%
      dplyr::group_by(.data$IDStation) %>%
      dplyr::summarise(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric) &
                                       !tidyselect::vars_select_helpers$where(is.integer), ~ sum(is.na(.x))))
    NA_count_perc_stat <- Data %>%
      dplyr::group_by(.data$IDStation) %>%
      dplyr::summarise(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric) &
                                       !tidyselect::vars_select_helpers$where(is.integer),
                                     ~ round(sum(is.na(.x))/dplyr::n()*100,2)))
    mean_vals_stat <- Data %>%
      dplyr::group_by(.data$IDStation) %>%
      dplyr::summarise(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric) &
                                       !tidyselect::vars_select_helpers$where(is.integer), ~ round(mean(.x, na.rm=T),2)))
    sd_vals_stat <- Data %>%
      dplyr::group_by(.data$IDStation) %>%
      dplyr::summarise(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric) &
                                       !tidyselect::vars_select_helpers$where(is.integer), ~ round(sd(.x, na.rm=T),2)))
    min_vals_stat <- Data %>%
      dplyr::group_by(.data$IDStation) %>%
      dplyr::summarise(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric) &
                                       !tidyselect::vars_select_helpers$where(is.integer), ~ round(min(.x,na.rm=T),2)))
    max_vals_stat <- Data %>%
      dplyr::group_by(.data$IDStation) %>%
      dplyr::summarise(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric) &
                                       !tidyselect::vars_select_helpers$where(is.integer), ~ round(max(.x,na.rm=T),2)))
    null_vals_stat <- Data %>%
      dplyr::group_by(.data$IDStation) %>%
      dplyr::summarise(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric) &
                                       !tidyselect::vars_select_helpers$where(is.integer), ~ sum(.x == 0, na.rm=T)))
    neg_vals_stat <- Data %>%
      dplyr::group_by(.data$IDStation) %>%
      dplyr::summarise(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric) &
                                       !tidyselect::vars_select_helpers$where(is.integer), ~ sum(.x < 0, na.rm=T)))
    Descr_by_IDStat <- list(NA_count_stat,NA_count_perc_stat,null_vals_stat,neg_vals_stat,
                            min_vals_stat,mean_vals_stat,max_vals_stat,sd_vals_stat)
    names(Descr_by_IDStat) <- c("NA_count_by_stat","NA_perc_by_stat","Null_count_by_stat","Negative_count_by_stat",
                                "Min_by_stat","Mean_by_stat","Max_by_stat","StdDev_by_stat")
    for (i in 1:length(Descr_by_IDStat)) {
      Descr_by_IDStat[[i]][is.nan_df(Descr_by_IDStat[[i]])] <- NA
      Descr_by_IDStat[[i]][is.infinite_df(Descr_by_IDStat[[i]])] <- NA
      if (is_ARPALdf_AQ(Data = Data) == TRUE) {
        attr(Descr_by_IDStat[[i]], "class") <- c("ARPALdf","ARPALdf_AQ","tbl_df","tbl","data.frame")
      } else if (is_ARPALdf_W(Data = Data) == TRUE) {
        attr(Descr_by_IDStat[[i]], "class") <- c("ARPALdf","ARPALdf_W","tbl_df","tbl","data.frame")
      }
    }
  }


  ### Statistics by Year
  if (by_Year == TRUE) {
    if (verbose == TRUE) {
      cat("Computing summary statistics by Year \n")
    }
    NA_count_year <- Data %>%
      dplyr::group_by(lubridate::year(.data$Date)) %>%
      dplyr::summarise(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric) &
                                       !tidyselect::vars_select_helpers$where(is.integer), ~ sum(is.na(.x))))
    NA_count_perc_year <- Data %>%
      dplyr::group_by(lubridate::year(.data$Date)) %>%
      dplyr::summarise(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric) &
                                       !tidyselect::vars_select_helpers$where(is.integer),
                                     ~ round(sum(is.na(.x))/dplyr::n()*100,2)))
    mean_vals_year <- Data %>%
      dplyr::group_by(lubridate::year(.data$Date)) %>%
      dplyr::summarise(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric) &
                                       !tidyselect::vars_select_helpers$where(is.integer), ~ round(mean(.x, na.rm=T),2)))
    sd_vals_year <- Data %>%
      dplyr::group_by(lubridate::year(.data$Date)) %>%
      dplyr::summarise(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric) &
                                       !tidyselect::vars_select_helpers$where(is.integer), ~ round(sd(.x, na.rm=T),2)))
    min_vals_year <- Data %>%
      dplyr::group_by(lubridate::year(.data$Date)) %>%
      dplyr::summarise(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric) &
                                       !tidyselect::vars_select_helpers$where(is.integer), ~ round(min(.x,na.rm=T),2)))
    max_vals_year <- Data %>%
      dplyr::group_by(lubridate::year(.data$Date)) %>%
      dplyr::summarise(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric) &
                                       !tidyselect::vars_select_helpers$where(is.integer), ~ round(max(.x,na.rm=T),2)))
    null_vals_year <- Data %>%
      dplyr::group_by(lubridate::year(.data$Date)) %>%
      dplyr::summarise(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric) &
                                       !tidyselect::vars_select_helpers$where(is.integer), ~ sum(.x == 0, na.rm=T)))
    neg_vals_year <- Data %>%
      dplyr::group_by(lubridate::year(.data$Date)) %>%
      dplyr::summarise(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric) &
                                       !tidyselect::vars_select_helpers$where(is.integer), ~ sum(.x < 0, na.rm=T)))
    Descr_by_year <- list(NA_count_year,NA_count_perc_year,null_vals_year,neg_vals_year,
                          min_vals_year, mean_vals_year,max_vals_year,sd_vals_year)
    names(Descr_by_year) <- c("NA_count_by_year","NA_perc_by_year","Null_count_by_year","Negative_count_by_year",
                              "Min_by_year","Mean_by_year","Max_by_year","StdDev_by_year")
    for (i in 1:length(Descr_by_year)) {
      Descr_by_year[[i]][is.nan_df(Descr_by_year[[i]])] <- NA
      Descr_by_year[[i]][is.infinite_df(Descr_by_year[[i]])] <- NA
      if (is_ARPALdf_AQ(Data = Data) == TRUE) {
        attr(Descr_by_year[[i]], "class") <- c("ARPALdf","ARPALdf_AQ","tbl_df","tbl","data.frame")
      } else if (is_ARPALdf_W(Data = Data) == TRUE) {
        attr(Descr_by_year[[i]], "class") <- c("ARPALdf","ARPALdf_W","tbl_df","tbl","data.frame")
      }
    }
  }


  ### Gap length by variable
  if (gap_length == TRUE) {
    if (verbose == TRUE) {
      cat("Computing gap lengths statistics by variable \n")
    }
    vars <- Data %>%
      dplyr::select(tidyselect::vars_select_helpers$where(is.double) &
                      tidyselect::vars_select_helpers$where(is.numeric) &
                      !tidyselect::vars_select_helpers$where(is.integer)) %>%
      colnames()
    gap_length <- vector("list", length = length(vars))
    for (v in 1:length(vars)) {
      var <- vars[v]
      gl <- Data %>%
        dplyr::select(.data$Date,.data$IDStation,.data$NameStation,var = var) %>%
        dplyr::filter(!is.na(.data$var)) %>%
        dplyr::arrange(.data$IDStation, .data$NameStation, .data$Date) %>%
        dplyr::group_by(.data$IDStation,.data$NameStation) %>%
        ### Modified on 2026-06-15: replace a multi-row dplyr::summarise() call with
        ### group-wise dplyr::mutate() + dplyr::lag(), which is compatible with recent
        ### dplyr versions and computes consecutive time gaps within each station.
        dplyr::mutate(gap = lubridate::interval(dplyr::lag(.data$Date), .data$Date)) %>%
        dplyr::mutate(gap = lubridate::time_length(.data$gap,unit = attributes(Data)$units)) %>%
        dplyr::filter(!is.na(.data$gap), .data$gap > 0) %>%
        dplyr::summarise(.groups = "keep",
                         min_gap = min(.data$gap),
                         q25_gap = quantile(.data$gap,probs = 0.25),
                         mean_gap = round(mean(.data$gap),3),
                         median_gap = quantile(.data$gap,probs = 0.50),
                         q75_gap = quantile(.data$gap,probs = 0.75),
                         max_gap = max(.data$gap),
                         sd_gap_length = round(sd(.data$gap),3),
                         Length1 = sum(.data$gap == 1),
                         Length2 = sum(.data$gap == 2),
                         Length24 = sum(.data$gap == 24)) %>%
        as.data.frame() %>%
        dplyr::mutate(dplyr::across(tidyselect::contains("gap"),
                                    ~ as_difftime(.x,units = attributes(Data)$units)))
      colnames(gl) <- c("IDStation","NameStation",
                        paste0(var,"_min_gap"),paste0(var,"_q25_gap"),
                        paste0(var,"_mean_gap"),paste0(var,"_median_gap"),
                        paste0(var,"_q75_gap"),paste0(var,"_max_gap"),paste0(var,"_sd_gap"),
                        paste0(var,"_freq_gap1"),paste0(var,"_freq_gap2"),paste0(var,"_freq_gap24"))
      if (is_ARPALdf_AQ(Data = Data) == TRUE) {
        attr(gl, "class") <- c("ARPALdf","ARPALdf_AQ","tbl_df","tbl","data.frame")
      } else if (is_ARPALdf_W(Data = Data) == TRUE) {
        attr(gl, "class") <- c("ARPALdf","ARPALdf_W","tbl_df","tbl","data.frame")
      }
      gap_length[[v]] <- gl
    }
    names(gap_length) <- vars
  }


  ### Histogram
  if (histogram == TRUE) {
    if (verbose == TRUE) {
      cat("Graphics: plotting histogram of each variable \n")
    }
    hist_plot <- Data %>%
      dplyr::select(tidyselect::vars_select_helpers$where(is.numeric) &
                      !tidyselect::vars_select_helpers$where(is.integer)) %>%
      tidyr::pivot_longer(cols = dplyr::everything()) %>%
      ggplot2::ggplot(aes(.data$value)) +
      ggplot2::geom_histogram(bins = 30, fill="blue") +
      ggplot2::facet_wrap(~ .data$name, scales = "free")
    print(hist_plot)
  }


  ### Kernel density plot
  if (density == TRUE) {
    if (verbose == TRUE) {
      cat("Graphics: plotting density of each variable \n")
    }
    dens_plot <- Data %>%
      dplyr::select(tidyselect::vars_select_helpers$where(is.numeric) &
                      !tidyselect::vars_select_helpers$where(is.integer)) %>%
      tidyr::pivot_longer(cols = dplyr::everything()) %>%
      ggplot2::ggplot(aes(.data$value)) +
      ggplot2::geom_density(fill="blue",alpha = 0.5) +
      ggplot2::facet_wrap(~ .data$name, scales = "free")
    print(dens_plot)
  }


  ### Outlier analysis
  if (outlier == TRUE) {
    if (verbose == TRUE) {
      cat("Computing outlier statisics for each variable \n")
    }
    Data_long <- Data %>%
      tidyr::pivot_longer(cols = tidyselect::vars_select_helpers$where(is.numeric) &
                            !tidyselect::vars_select_helpers$where(is.integer),
                          names_to = "Vars", values_to = "Value")
    boxp <- ggplot(Data_long, aes(y = .data$Value)) +
      geom_boxplot(aes(fill = .data$Vars)) +
      coord_flip() +
      facet_wrap(~ .data$Vars, scales = "free") +
      labs(title = "Boxplot on the whole sample by variable") +
      theme(legend.position = "")
    print(boxp)
    if (verbose == TRUE) {
      cat("Computing Hampel filter for each variable \n")
      cat("Reports the % of observations above +3*MAD and below -3*MAD \n")
    }
    out <- Data %>%
      dplyr::summarise(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric) &
                                       !tidyselect::vars_select_helpers$where(is.integer), ~ Hampel_flt(.x))) %>%
      list()
    hampel <- matrix(NA,nrow = dim(out[[1]])[2], ncol = 7)
    for (i in 1:dim(out[[1]])[2]) {
      hampel[i,2] <- round(as.numeric(out[[1]][[i]]$I_low),2)
      hampel[i,3] <- round(length(out[[1]][[i]]$outlier_ind_low),2)
      hampel[i,4] <- round(length(out[[1]][[i]]$outlier_ind_low)/dim(Data)[1]*100,2)
      hampel[i,5] <- round(as.numeric(out[[1]][[i]]$I_upp),2)
      hampel[i,6] <- round(length(out[[1]][[i]]$outlier_ind_upp),2)
      hampel[i,7] <- round(length(out[[1]][[i]]$outlier_ind_upp)/dim(Data)[1]*100,2)
    }
    hampel <- data.frame(hampel)
    hampel[,1] <- names(out[[1]])
    colnames(hampel) <- c("Variable","Lower_bound","Obs_below_low_count","Obs_below_low_perc","Upper_bound","Obs_above_upp_count","Obs_above_upp_perc")
    if (is_ARPALdf_AQ(Data = Data) == T) {
      attr(hampel, "class") <- c("ARPALdf","ARPALdf_AQ","tbl_df","tbl","data.frame")
    } else if (is_ARPALdf_W(Data = Data) == T) {
      attr(hampel, "class") <- c("ARPALdf","ARPALdf_W","tbl_df","tbl","data.frame")
    }
  }


  ### Correlation analysis
  if (correlation == TRUE) {
    if (verbose == TRUE) {
      cat("Computing linear correlation analysis for available variable \n")
    }
    stz <- unique(Data$IDStation)
    name_stz <- unique(Data$NameStation)
    cor_matrix <- vector("list", length = length(stz))
    for (s in 1:length(stz)) {
      cor_matrix[[s]] <- Data %>%
        dplyr::filter(.data$IDStation == stz[s]) %>%
        dplyr::select(tidyselect::vars_select_helpers$where(is.double) &
                        tidyselect::vars_select_helpers$where(is.numeric)) %>%
        stats::cor(use = "pairwise.complete.obs") %>%
        as.data.frame() %>%
        tibble::rownames_to_column(var = "Var1") %>%
        tidyr::pivot_longer(cols = -.data$Var1, names_to = "Var2", values_to = "corr") %>%
        dplyr::mutate(IDStation = stz[s], NameStation = name_stz[s]) %>%
        dplyr::filter(.data$Var1 != .data$Var2) %>%
        tidyr::pivot_wider(names_from = c("Var1","Var2"),values_from = "corr")
    }
    cor_matrix <- bind_rows(cor_matrix)
    if (is_ARPALdf_AQ(Data = Data) == TRUE) {
      attr(cor_matrix, "class") <- c("ARPALdf","ARPALdf_AQ","tbl_df","tbl","data.frame")
    } else if (is_ARPALdf_W(Data = Data) == TRUE) {
      attr(cor_matrix, "class") <- c("ARPALdf","ARPALdf_W","tbl_df","tbl","data.frame")
    }
  }


  ### Output list
  ret_list <- list(Descr = descriptives)
  if (exists("Descr_by_IDStat", inherits = FALSE)) {
    ret_list <- c(ret_list, Descr_by_IDStat = list(Descr_by_IDStat))
  }
  if (exists("Descr_by_year", inherits = FALSE)) {
    ret_list <- c(ret_list, Descr_by_year = list(Descr_by_year))
  }
  if (exists("hampel", inherits = FALSE)) {
    ret_list <- c(ret_list, Hampel = list(hampel))
  }
  if (exists("gap_length", inherits = FALSE)) {
    ret_list <- c(ret_list, Gap_length = list(gap_length))
  }
  if (exists("cor_matrix", inherits = FALSE)) {
    ret_list <- c(ret_list, Cor_matrix = list(cor_matrix))
  }


  ### Output
  return(ret_list)
}

#' Download air quality data at manucipal from ARPA Lombardia website
#'
#' More detailed description.
#'
#' @description 'get_ARPA_Lombardia_AQ_municipal_data' returns the air quality levels at municipal level estimated by
#' ARPA Lombardia using a physico-chemical model which simulates air quality based on weather and geo-physical
#' variables. For each municipality of Lombardy, ARPA estimates the average (NO2_mean) and maximum daily (NO2_max_day)
#' level of NO2, the daily maximum (Ozone_max_day) and the 8-hours moving window maximum (Ozone_max_8h) of Ozone
#' and the average levels of PM10 (PM10_mean) and PM2.5 (PM2.5_mean).
#' Data are available from 2011 and are updated up to the current date.
#' For more information, visit the section 'Stime comunali dell'aria' at the website:
#' https://www.dati.lombardia.it/stories/s/auv9-c2sj
#'
#' @param ID_station Numeric value. ID of the station to consider. Using ID_station = NULL, all the available
#' stations are selected. Default is ID_station = NULL.
#' @param Date_begin Character vector of the first date-time to download. Format can be either "YYYY-MM-DD" or "YYYY-MM-DD hh:mm:ss". Default is Date_begin = "2022-01-01".
#' @param Date_end Character vector of the last date-time to download. Format can be either "YYYY-MM-DD" or "YYYY-MM-DD hh:mm:ss". Default is Date_end = "2022-12-31".
#' @param Frequency Temporal aggregation frequency. It can be "daily", "weekly", "monthly" or "yearly".
#' Default is Frequency = "daily".
#' @param Var_vec Character vector of variables to aggregate. If NULL (default) all the variables are averaged.
#' @param Fns_vec Character vector of aggregation function to apply to the selected variables.
#' Available functions are mean, median, min, max, sum, qPP (PP-th percentile), sd, var,
#' vc (variability coefficient), skew (skewness) and kurt (kurtosis).
#' @param by_sensor Logical value (TRUE or FALSE). If 'by_sensor = TRUE', the function returns the observed concentrations by sensor code, while
#' if 'by_sensor = FALSE' (default) it returns the observed concentrations by station.
#' @param verbose Logical value (TRUE or FALSE). Toggle warnings and messages. If 'verbose = TRUE' (default) the function
#' prints on the screen some messages describing the progress of the tasks. If 'verbose = FALSE' any message about
#' the progression is suppressed.
#' @param parallel Logical value (TRUE or FALSE). If 'parallel = FALSE' (default), data downloading is performed using a sequential/serial approach and additional parameters 'parworkers' and 'parfuturetype' are ignored.
#' When 'parallel = TRUE', data downloading is performed using parallel computing through the Futureverse setting.
#' More detailed information about parallel computing in the Futureverse is available at the following websites:
#' https://future.futureverse.org/ and https://cran.r-project.org/web/packages/future.apply/vignettes/future.apply-1-overview.html
#' @param parworkers Numeric integer value. If 'parallel = TRUE' (parallel mode active), the user can declare the number of parallel workers to be activated using 'parworkers = integer number'. By default ('parworkers = NULL'), the number of active workers is half of the available local cores.
#' @param parfuturetype Character vector. If 'parallel = TRUE' (parallel mode active), the user can declare the parallel strategy to be used according to the Futureverse syntax through 'parfuturetype'. By default, the 'multisession' (background R sessions on local machine) is used. In alternative, the 'multicore' (forked R processes on local machine. Not supported by Windows and RStudio) setting can be used.
#'
#' @return A data frame of class 'data.frame' and 'ARPALdf'. The object is fully compatible with Tidyverse.
#' The column 'NameStation' identifies the name of each municipality. The column 'IDStation' is an ID code
#' (assigned from ARPA) uniquely identifying each municipality.
#'
#' @examples
#' \donttest{
#' ### Modified on 2026-06-15: keep examples small to avoid stressing the
#' ### public Open Data Lombardia API during CRAN checks.
#' ## Download daily municipal concentrations in January 2025 for city number 100451.
#' get_ARPA_Lombardia_AQ_municipal_data(ID_station=100451,Date_begin = "2025-01-01",
#'                                      Date_end = "2025-01-07", Frequency="daily")
#' ## Download monthly concentrations of NO2 (average and maximum) observed in early 2026
#' ## at city number 100451.
#' get_ARPA_Lombardia_AQ_municipal_data(ID_station=100451,
#'                                      Date_begin = "2026-01-01", Date_end = "2026-03-31",
#'                                      Frequency="monthly",
#'                                      Var_vec=c("NO2_mean","NO2_mean"), Fns_vec=c("mean","max"))
#' ## Download daily concentrations observed in March 2023 at city number 100451.
#' ## Data are reported by sensor.
#' get_ARPA_Lombardia_AQ_municipal_data(ID_station=100451,
#'                                      Date_begin = "2023-03-01", Date_end = "2023-03-07",
#'                                      by_sensor = TRUE)
#' }
#'
#' @export

get_ARPA_Lombardia_AQ_municipal_data <-
  function(ID_station = NULL, Date_begin = "2021-01-01", Date_end = "2022-12-31", Frequency = "daily",
           Var_vec = NULL, Fns_vec = NULL, by_sensor = FALSE, verbose=TRUE,
           parallel = FALSE, parworkers = NULL, parfuturetype = "multisession") {

    ###  Welcome message
    if (verbose == TRUE) {
      cat("Retrieving desired ARPA Lombardia dataset: started at", as.character(Sys.time()), "\n")
    }

    ############################
    ########## Checks ##########
    ############################

    ##### Check for internet connection
    if(!curl::has_internet()) {
      message("Internet connection not available at the moment.\nPlease check your internet connection. If the problem persists, please contact the package maintainer.")
      return(invisible(NULL))
    }

    ##### Check if packages for JSON/SODA download are installed
    ### Modified on 2026-06-15: replace the socratadata dependency check with httr2/jsonlite checks.
    rlang::check_installed("httr2",
                           reason = "Package \"httr2\" must be installed to download data from ARPA Lombardia Open Database.")
    rlang::check_installed("jsonlite",
                           reason = "Package \"jsonlite\" must be installed to parse JSON data from ARPA Lombardia Open Database.")

    ##### Check if Futureverse is installed
    rlang::check_installed("future",
                           reason = "Package \"future\" must be installed to download (parallel) data from ARPA Lombardia Open Database.")
    rlang::check_installed("future.apply",
                           reason = "Package \"future\" must be installed to download (parallel) data from ARPA Lombardia Open Database.")

    ##### Define %notin%
    '%notin%' <- Negate('%in%')

    ##### Checks if by_sensor setup properly
    if (by_sensor %notin% c(0,1,FALSE,TRUE)) {
      stop("Wrong setup for 'by_sensor'. Use 1 or 0 or TRUE or FALSE.",
           call. = FALSE)
    }

    ##### Checks if parallel setup properly
    if (parallel %notin% c(FALSE,TRUE)) {
      stop("Wrong setup for 'parallel'. Use TRUE or FALSE.",
           call. = FALSE)
    }
    if (parfuturetype %notin% c("multisession","multicore")) {
      stop("Wrong setup for 'parallel'. Use TRUE or FALSE.",
           call. = FALSE)
    }

    ##### Checks Input frequency
    stopifnot("Frequency cannot be hourly as municipal estimates are provided on a daily basis" = (Frequency != "hourly") == T)


    #############################################
    ########## Preparation to download ##########
    #############################################

    ##### Define %notin%
    '%notin%' <- Negate('%in%')

    ### Registry/Metadata
    Metadata <- AQ_municipal_metadata_reshape()
    Metadata <- Metadata %>%
      dplyr::select(-c(.data$Province,.data$DateStart,.data$DateStop))

    ### Checks if ID_station is valid (in the list of active stations)
    if (!is.null(ID_station) & all(ID_station %notin% Metadata$IDStation)) {
      stop("ID_station NOT in the list of active stations. Change ID_station or use ID_station = NULL",
           call. = FALSE)
    }

    if (!is.null(ID_station)) {
      Metadata <- Metadata %>%
        dplyr::filter(.data$IDStation %in% ID_station)
    }

    ##### Splitting strategy for improving download speed
    n_blocks <- 36
    # Check dates format
    if (is.null(lubridate::guess_formats(x = Date_begin, orders = c("ymd HMS","ymd")))) {
      stop("Wrong format for 'Date_begin'. Use 'YYYY-MM-DD' or 'YYYY-MM-DD hh:mm:ss'", call. = FALSE)
    }
    if (is.null(lubridate::guess_formats(x = Date_end, orders = c("ymd HMS","ymd")))) {
      stop("Wrong format for 'Date_end'. Use 'YYYY-MM-DD' or 'YYYY-MM-DD hh:mm:ss'", call. = FALSE)
    }
    if (is.na(lubridate::ymd_hms(Date_begin, quiet = TRUE))) {
      Date_begin <- lubridate::ymd_hms(paste0(Date_begin," 00:00:00"))
    }
    if (is.na(lubridate::ymd_hms(Date_end, quiet = TRUE))) {
      Date_end <- lubridate::ymd_hms(paste0(Date_end," 23:00:00"))
    }
    ### Check for the presence of breaking dates (URLs change in several years)
    if (lubridate::year(Date_begin) == lubridate::year(Date_end)) {
      break_years <- lubridate::year(Date_begin)
    } else {
      break_years <- seq(from = lubridate::year(Date_begin), to = lubridate::year(Date_end), by = 1)
    }

    ##### Check online availability of the resources for the specified years
    if (verbose == TRUE) {
      cat("Checking availability of online resources: started at", as.character(Sys.time()), "\n")
    }
    URLs <- res_check <- numeric(length = length(break_years))
    for (yr in 1:length(break_years)) {
      URLs[yr] <- url <- url_AQ_year_dataset(Stat_type = "AQ_municipal_check", Year = break_years[yr])
      temp <- tempfile()
      res <- suppressWarnings(try(curl::curl_fetch_disk(url, temp), silent = TRUE))
      URLs[yr] <- url <- url_AQ_year_dataset(Stat_type = "AQ_municipal", Year = break_years[yr])
      if(res$status_code != 200) {
        message(paste0("The internet resource for year ", break_years[yr]," is not available at the moment. Status code: ",res$status_code,".\nPlease, try later. If the problem persists, please contact the package maintainer."))
      } else {
        res_check[yr] <- 1
      }
    }
    if (verbose == TRUE) {
      if (sum(res_check) == length(break_years)) {
        message("All the online resources are available.\n")
      }
      if (sum(res_check) > 0 & sum(res_check) < length(break_years)) {
        message("Part of the required online resources are not available. Please, try with a new request.\n")
        return(invisible(NULL))
      }
      if (sum(res_check) == 0) {
        message("None of the required online resources are available. Please, try with a new request.\n")
        return(invisible(NULL))
      }
    }

    ### Downloading data
    if (verbose == TRUE) {
      cat("Downloading data: started at", as.character(Sys.time()), "\n")
    }

    # clean RAM
    invisible(gc())



    ######################################
    ########## Downloading data ##########
    ######################################

    ### Filter valid URLs and dates
    ### Modified on 2026-06-15: keep municipal years greater than or equal to 2011,
    ### because url_AQ_year_dataset() maps Year >= 2025 to the current municipal
    ### JSON endpoint. Restricting to 2011:2025 drops valid 2026+ requests and
    ### makes the date-block sequence fail when by_sensor = TRUE.
    URLs_22on <- URLs[break_years >= 2011]
    break_years_22on <- break_years[break_years >= 2011]
    ### Modified on 2026-06-15: stop early if no valid municipal year is available.
    if (length(break_years_22on) == 0) {
      warning("No valid ARPA Lombardia municipal air-quality year was found for the requested dates.",
              call. = FALSE)
      Aria <- tibble::tibble(Date = as.Date(character()),
                             IDStation = numeric(),
                             NameStation = character())
      attr(Aria, "class") <- c("ARPALdf","ARPALdf_AQ_mun","tbl_df","tbl","data.frame")
      attr(Aria, "frequency") <- Frequency
      attr(Aria, "units") <- dplyr::case_when(Frequency == "daily" ~ "days",
                                      Frequency == "weekly" ~ "weeks",
                                      Frequency == "monthly" ~ "months",
                                      Frequency == "yearly" ~ "years")
      return(Aria)
    }

    if (lubridate::ymd("2011-01-01") > Date_begin) {
      Date_begin_22on <- lubridate::ymd("2011-01-01")
    } else {
      Date_begin_22on <- Date_begin
    }

    ### Building URLs/links in Socrata format using sequences of dates
    if (length(break_years_22on) == 1) {
      break_dates <- paste0(break_years_22on,"-12-31 23:00:00")
    } else {
      break_dates <- paste0(break_years_22on[-length(break_years_22on)],"-12-31 23:00:00")
    }
    dates_seq_end <- c(Date_begin_22on,lubridate::ymd_hms(break_dates),Date_end)
    dates_seq_end <- unique(dates_seq_end)
    dates_seq_end <- dates_seq_end[-1]
    dates_seq_end <- dates_seq_end[dates_seq_end <= Date_end & dates_seq_end >= Date_begin_22on]
    dates_seq_begin <- c(Date_begin_22on,lubridate::ymd_hms(break_dates) + lubridate::hours(1))
    dates_seq_begin <- unique(dates_seq_begin)
    dates_seq_begin <- dates_seq_begin[dates_seq_begin <= Date_end & dates_seq_begin >= Date_begin_22on]

    URL_blocks <- vector(mode = "list", length = length(dates_seq_begin))
    for (b in 1:length(dates_seq_begin)) {
      seq_temp <- seq(dates_seq_begin[b],
                      dates_seq_end[b],
                      length.out = n_blocks + 1)

      seq_begin_temp <- seq_temp[-length(seq_temp)]
      seq_begin_temp <- lubridate::round_date(seq_begin_temp, unit = "hour")
      seq_begin_temp <- stringr::str_replace(string = seq_begin_temp, pattern = " ", replacement = "T")

      seq_end_temp <- c(seq_temp[-c(1,length(seq_temp))] - lubridate::hours(1),seq_temp[length(seq_temp)])
      seq_end_temp <- lubridate::round_date(seq_end_temp, unit = "hour")
      seq_end_temp <- stringr::str_replace(string = seq_end_temp, pattern = " ", replacement = "T")

      if (is.null(ID_station)) {
        str_sensor <- NULL
      } else {
        str_sensor <- paste0("AND idsensore in(",paste0(sapply(X = Metadata$IDSensor, function(x) paste0("'",x,"'")),collapse = ","),")")
      }

      URL_blocks[[b]] <- data.frame(seq_begin_temp,seq_end_temp) %>%
        dplyr::mutate(link = paste0(URLs_22on[b],"?$where=data between '", seq_begin_temp, "' and '", seq_end_temp,"'", str_sensor)) %>%
        dplyr::select(.data$link)
    }
    URL_blocks <- dplyr::bind_rows(URL_blocks)

    ### Preparing parallel computation: explicitly open multisession/multicore workers by switching plan
    if (parallel == TRUE) {
      future::plan(future::multisession, workers = 12)
      if (is.null(parworkers)) {
        parworkers <- future::availableCores()/2
      }
      eval(parse(text = paste0("future::plan(future::",parfuturetype,", workers = ",parworkers,")")))
      if (verbose == TRUE) {
        message("Start parallel computing: number of parallel workers = ", future::nbrOfWorkers())
      }
    }

    ### Modified on 2026-06-15: replace socratadata::soc_read with the internal JSON/SODA reader.
    Aria <- dplyr::bind_rows(
      future.apply::future_apply(X = as.matrix(URL_blocks[,1]), MARGIN = 1, FUN = function(x) {
        ARPAL_read_socrata_json(url = as.character(x))
      })
    )

    ### Preparing parallel computation: explicitly open multisession/multicore workers by switching plan
    if (parallel == TRUE) {
      # Explicitly close multisession/multicore workers by switching plan
      future::plan(future::sequential)
      if (verbose == TRUE) {
        message("Stop parallel computing: number of parallel workers = ", future::nbrOfWorkers())
      }
    }

    ### Processing data
    Aria <- Aria %>%
      dplyr::select(IDSensor = .data$idsensore,
                    Date = .data$data,
                    Value = .data$valore,
                    Operator = .data$idoperatore) %>%
      dplyr::mutate(IDSensor = as.numeric(.data$IDSensor),
                    Value = as.numeric(.data$Value),
                    ### Modified on 2026-06-15: parse Socrata JSON date/datetime strings robustly.
                    Date = ARPAL_parse_socrata_date(.data$Date),
                    Operator = dplyr::case_when(.data$Operator == 1 ~ "mean",
                                                .data$Operator == 3 ~ "max",
                                                .data$Operator == 11 ~ "max_8h",
                                                .data$Operator == 12 ~ "max_day"))

    ### Ending parallel computation: explicitly close multisession/multicore workers by switching plan
    if (parallel == TRUE) {
      future::plan(future::sequential)
      if (verbose == TRUE) {
        message("Stop parallel computing: number of parallel workers = ", future::nbrOfWorkers())
      }
    }

    # clean RAM
    invisible(gc())



    #####################################
    ########## Processing data ##########
    #####################################

    ### Add metadata
    ### Modified on 2026-06-15: align IDSensor types before the join, because
    ### empty JSON/SODA responses and recent dplyr versions no longer silently
    ### coerce numeric and character join keys.
    Aria <- Aria %>%
      dplyr::mutate(IDSensor = as.character(.data$IDSensor))
    Metadata <- Metadata %>%
      dplyr::mutate(IDSensor = as.character(.data$IDSensor))
    Aria <- dplyr::right_join(Aria, Metadata, by = "IDSensor")

    ### Cleaning
    if (by_sensor %in% c(1,TRUE)) {
      Aria <- Aria %>%
        dplyr::mutate(Pollutant = paste0(.data$Pollutant,"_",.data$Operator)) %>%
        dplyr::filter(!is.na(.data$Date)) %>%
        dplyr::select(-c(.data$Operator)) %>%
        dplyr::mutate(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric),
                                    ~ dplyr::na_if(.,-9999))) %>%
        dplyr::mutate(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric),
                                    ~ dplyr::na_if(.,NaN))) %>%
        dplyr::select(.data$Date,.data$IDStation,.data$NameStation,.data$IDSensor,
                      .data$Pollutant,.data$Value)
    } else if (by_sensor %in% c(0,FALSE)) {
      Aria <- Aria %>%
        dplyr::mutate(Pollutant = paste0(.data$Pollutant,"_",.data$Operator)) %>%
        dplyr::filter(!is.na(.data$Date)) %>%
        dplyr::select(-c(.data$IDSensor,.data$Operator)) %>%
        dplyr::mutate(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric),
                                    ~ dplyr::na_if(.,-9999))) %>%
        dplyr::mutate(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric),
                                    ~ dplyr::na_if(.,NaN))) %>%
        tidyr::pivot_wider(names_from = .data$Pollutant, values_from = .data$Value,
                           values_fn = function(x) mean(x,na.rm=T)) # Mean (without NA) of a NA vector = NaN
    }
    Aria[is.na(Aria)] <- NA
    Aria[is.nan_df(Aria)] <- NA

    ### Add dataset attributes
    attr(Aria, "class") <- c("ARPALdf","ARPALdf_AQ","tbl_df","tbl","data.frame")

    ### Modified on 2026-06-15: when by_sensor = TRUE municipal data are kept
    ### in long format (Pollutant, Value). Therefore the wide-variable checks
    ### based on NO2_mean, PM10_mean, and similar columns must be skipped.
    if (by_sensor %in% c(1, TRUE)) {
      Aria <- Aria %>%
        dplyr::arrange(.data$Date) %>%
        dplyr::filter(.data$Date >= Date_begin,
                      .data$Date <= Date_end)

      structure(list(Aria = Aria))
      attr(Aria, "class") <- c("ARPALdf","ARPALdf_AQ_mun","tbl_df","tbl","data.frame")
      attr(Aria, "frequency") <- "daily"
      attr(Aria, "units") <- "days"

      if (verbose == TRUE) {
        cat("Processing data: ended at", as.character(Sys.time()), "\n")
        cat("Retrieving desired ARPA Lombardia dataset: ended at", as.character(Sys.time()), "\n")
      }

      return(Aria)
    }

    ###
    if (is.null(Var_vec) & is.null(Fns_vec)) {
      vv <- c("NO2_mean","NO2_max_day","Ozone_max_8h","Ozone_max_day","PM10_mean","PM2.5_mean")
      vv <- vv[vv %in% names(Aria)]
      fv <- rep("mean",length(vv))
    } else {
      vv <- Var_vec
      fv <- Fns_vec
    }

    ### Modified on 2026-06-15: avoid regularization failures when an online
    ### ARPA Lombardia municipal endpoint/query returns no usable pollutant
    ### columns after filtering. This keeps the function behavior stable for
    ### empty or temporarily unavailable API responses.
    if (nrow(Aria) == 0 || length(vv) == 0) {
      if (verbose == TRUE) {
        warning("No ARPA Lombardia municipal air-quality data were returned for the selected stations, pollutants and dates.",
                call. = FALSE)
      }

      if (is.null(Var_vec) & is.null(Fns_vec)) {
        vv_empty <- c("NO2_mean","NO2_max_day","Ozone_max_8h","Ozone_max_day",
                      "PM10_mean","PM2.5_mean")
      } else {
        vv_empty <- Var_vec
      }

      if (by_sensor %in% c(1, TRUE)) {
        Aria <- tibble::tibble(
          Date = as.Date(character()),
          IDStation = numeric(),
          NameStation = character(),
          IDSensor = character(),
          Pollutant = character(),
          Value = numeric()
        )
        attr(Aria, "class") <- c("ARPALdf","ARPALdf_AQ","tbl_df","tbl","data.frame")
        attr(Aria, "frequency") <- "daily"
        attr(Aria, "units") <- "days"
      } else {
        Aria <- tibble::tibble(
          Date = as.Date(character()),
          IDStation = numeric(),
          NameStation = character()
        )
        for (v in vv_empty) {
          Aria[[v]] <- numeric()
        }

        freq_unit <- dplyr::case_when(Frequency == "daily" ~ "days",
                                      Frequency == "weekly" ~ "weeks",
                                      Frequency == "monthly" ~ "months",
                                      Frequency == "yearly" ~ "years")

        attr(Aria, "class") <- c("ARPALdf","ARPALdf_AQ_mun","tbl_df","tbl","data.frame")
        attr(Aria, "frequency") <- Frequency
        attr(Aria, "units") <- freq_unit
      }

      if (verbose == TRUE) {
        cat("Processing data: ended at", as.character(Sys.time()), "\n")
        cat("Retrieving desired ARPA Lombardia dataset: ended at", as.character(Sys.time()), "\n")
      }

      return(Aria)
    }

    # Checks if all the pollutants are available for the selected stations
    if (all(dplyr::all_of(vv) %in% names(Aria)) == FALSE) {
      stop("One or more pollutants are not available for the selected stations! Change the values of 'Var_vec'",
           call. = FALSE)
    }

    if (by_sensor %in% c(0,FALSE)) {
      ### Aggregating dataset
      if (Frequency != "daily") {
        if (verbose==T) {
          cat("Aggregating ARPA Lombardia data: started at", as.character(Sys.time()), "\n")
        }
        Aria <- Aria %>%
          Time_aggregate(Frequency = Frequency, Var_vec = Var_vec, Fns_vec = Fns_vec, verbose = verbose) %>%
          dplyr::arrange(.data$NameStation, .data$Date)
      } else {
        Aria <- Aria %>%
          dplyr::select(.data$Date,.data$IDStation,.data$NameStation,vv) %>%
          dplyr::arrange(.data$NameStation, .data$Date)
      }

      ### Regularizing dataset: same number of timestamps for each station and variable
      if (verbose == TRUE) {
        cat("Regularizing ARPA Lombardia data: started at", as.character(Sys.time()), "\n")
      }
      freq_unit <- dplyr::case_when(Frequency == "daily" ~ "days",
                                    Frequency == "weekly" ~ "weeks",
                                    Frequency == "monthly" ~ "months",
                                    Frequency == "yearly" ~ "years")

      Aria <- Aria %>%
        dplyr::arrange(.data$Date) %>%
        dplyr::filter(.data$Date >= Date_begin,
                      .data$Date <= Date_end) %>%
        tidyr::pivot_longer(cols = -c(.data$Date,.data$IDStation,.data$NameStation),
                            names_to = "Measure", values_to = "Value") %>%
        tidyr::pivot_wider(names_from = .data$Date, values_from = .data$Value) %>%
        tidyr::pivot_longer(cols = -c(.data$Measure,.data$IDStation,.data$NameStation),
                            names_to = "Date", values_to = "Value") %>%
        tidyr::pivot_wider(names_from = .data$Measure, values_from = .data$Value) %>%
        dplyr::mutate(Date = lubridate::ymd(.data$Date)) %>%
        dplyr::arrange(.data$IDStation,.data$Date)

      structure(list(Aria = Aria))
      attr(Aria, "class") <- c("ARPALdf","ARPALdf_AQ_mun","tbl_df","tbl","data.frame")
      attr(Aria, "frequency") <- Frequency
      attr(Aria, "units") <- freq_unit

    } else if (by_sensor %in% c(1,TRUE)) {
      Aria <- Aria %>%
        dplyr::arrange(.data$Date) %>%
        dplyr::filter(.data$Date >= Date_begin,
                      .data$Date <= Date_end)

      structure(list(Aria = Aria))
      attr(Aria, "class") <- c("ARPALdf","ARPALdf_AQ","tbl_df","tbl","data.frame")
      attr(Aria, "frequency") <- "daily"
      attr(Aria, "units") <- "days"
    }

    if (verbose == TRUE) {
      cat("Processing data: ended at", as.character(Sys.time()), "\n")
    }

    if (verbose == TRUE) {
      cat("Retrieving desired ARPA Lombardia dataset: ended at", as.character(Sys.time()), "\n")
    }

    return(Aria)
  }

#' @keywords internal
#' @noRd

ARPAL_parse_socrata_datetime <- function(x) {

  ### Modified on 2026-06-15: parse Socrata JSON date/datetime strings
  ### robustly and return POSIXct values. This helper is used in functions
  ### that expose sub-daily observations, where preserving the hour/minute
  ### component is required.

  ### Convert the input to character to make the parser robust to the classes
  ### returned by jsonlite::fromJSON(), readr, or intermediate package code.
  x <- as.character(x)

  ### Socrata JSON endpoints commonly return timestamps in ISO-like form, such
  ### as "2026-03-01T12:30:00.000". lubridate::ymd_hms() handles these strings
  ### and returns POSIXct values. Warnings are suppressed because a fallback is
  ### intentionally applied to date-only strings.
  out <- suppressWarnings(lubridate::ymd_hms(x, tz = "UTC", quiet = TRUE))

  ### Values not parsed by ymd_hms() but still present are generally date-only
  ### strings. They are interpreted as midnight UTC of the corresponding date.
  idx <- is.na(out) & !is.na(x) & nzchar(x)

  if (any(idx)) {
    out[idx] <- suppressWarnings(
      as.POSIXct(lubridate::ymd(x[idx], quiet = TRUE), tz = "UTC")
    )
  }

  ### Return POSIXct values. This keeps hourly and 10-minute outputs as proper
  ### date-time columns instead of character vectors.
  out
}

#' Check if a given object belongs to class 'ARPALdf'
#'
#' @description 'is_ARPALdf' checks if the input object belongs to the class 'ARPALdf'
#'
#' @param Data Object to check if the class of a dataframe is 'ARPALdf', i.e. ARPAL dataframe.
#'
#' @return The function returns 'True' if the object is of class 'ARPALdf' and it returns 'False' if the
#'     object isn't of class 'ARPALdf'
#'
#' @examples
#' d <- get_ARPA_Lombardia_AQ_registry()
#' is_ARPALdf(d)
#'
#' @export

is_ARPALdf <- function(Data) {
  is.element("ARPALdf", attr(Data,"class"))
}

#' Identifies the K-nearest neighbors (stations) to all the monitoring sites included in a given ARPALdf
#' registry data.frame. The neighbors are identified computing the Euclidean distance among the sites'
#' coordinates.
#'
#' @description For each element included in reg_X, it identifies the k-nearest neighboring locations
#' (among those included in reg_Y) according to an Euclidean distance metric. reg_X and reg_Y
#' must be two 'ARPALdf' objects obtained using get_ARPA_Lombardia_xxx_registry'.
#'
#' @param reg_X Dataset of class 'ARPALdf' containing the stations list obtained as registry
#' (from 'get_ARPA_Lombardia_xxx_registry' command). The object must contain the following
#' colums: 'IDStation','NameStation','Longitude' and 'Latitude'.
#' @param reg_Y Dataset of class 'ARPALdf' containing the stations list obtained as registry
#' (from 'get_ARPA_Lombardia_xxx_registry' command). The object must contain the following
#' colums: 'IDStation','NameStation','Longitude' and 'Latitude'.
#' @param k Integer value. Represents the number of neighbors the user wants to identify.
#'
#' @return A data.frame object having the same length of reg_X. For each row (stations in reg_X)
#' it contains the name and the IDStation code for the k-nearest neighbors.
#'
#' @examples
#' if (require("tidyverse")) {
#'     regAQ <- get_ARPA_Lombardia_AQ_registry()
#'     regAQ <- regAQ %>% filter(Pollutant %in% c("PM10","Ammonia"))
#'     regW <- get_ARPA_Lombardia_W_registry()
#'     registry_KNN_dist(regAQ,regW,k=2)
#' }
#'
#' @export

registry_KNN_dist <- function(reg_X,reg_Y,k=1){

  # For each element included in reg_X, it identifies the k-nearest neighbors (among those included in reg_Y)
  # according to an Euclidean distance metric

  reg_X <- reg_X %>%
    dplyr::distinct(.data$IDStation,.data$NameStation,.data$Longitude,.data$Latitude) %>%
    sf::st_as_sf(coords = c("Longitude", "Latitude"), crs=4326) %>%
    dplyr::arrange(.data$IDStation)
  reg_Y <- reg_Y %>%
    dplyr::distinct(.data$IDStation,.data$NameStation,.data$Longitude,.data$Latitude) %>%
    sf::st_as_sf(coords = c("Longitude", "Latitude"), crs=4326) %>%
    dplyr::arrange(.data$IDStation)

  ### Computing distance
  dist_matrix <- sf::st_distance(reg_X,reg_Y)
  ### Identifying the k-NN IDStation of reg_Y
  min_dist_idx <- apply(dist_matrix,MARGIN = 1, FUN = k_min_vals_vec_idx, k=k)
  min_dist <- apply(dist_matrix,MARGIN = 1, FUN = k_min_vals_vec, k=k)
  if (k == 1) {
    min_dist_idx <- t(as.matrix(min_dist_idx))
    min_dist <- t(as.matrix(min_dist))
  }
  ### Extracting for each k the corresponding ID and NameStation of reg_Y
  knn_list <- list(length = k)
  for (j in 1:k) {
    knn_prog <- data.frame(cbind(reg_Y[min_dist_idx[j,],]$NameStation,
                                 reg_Y[min_dist_idx[j,],]$IDStation,
                                 round(min_dist[j,],2)))
    colnames(knn_prog) <- c(paste0("reg_Y_nn",j,"_name"),
                            paste0("reg_Y_nn",j,"_ID"),
                            paste0("reg_Y_nn",j,"_dist"))
    knn_list[[j]] <- knn_prog
  }
  ### Selecting columns from reg_X
  reg_X_name <- reg_X %>%
    dplyr::select(reg_X_name = .data$NameStation, reg_X_ID = .data$IDStation)
  ### Combining dataframes
  output_tab <- dplyr::bind_cols(reg_X_name,knn_list)

  return(list(output_tab))
}

#' Download ARPA Lombardia zoning geometries
#'
#' @description 'get_ARPA_Lombardia_zoning' returns the geometries (polygonal shape file) and a map of
#' the ARPA zoning of Lombardy. The zoning reflects the main orographic characteristics of the territory.
#' Lombardy region is classified into seven type of areas: large urbanized areas, urbanized areas in rural
#' contexts, rural areas, mountainous areas and valley bottom.
#' For more information, visit the section 'Zonizzazione ARPA Lombardia' at the websites
#' https://www.arpalombardia.it/temi-ambientali/aria/rete-di-rilevamento/classificazione-zone/ and
#' https://www.arpalombardia.it/temi-ambientali/aria/mappa-della-zonizzazione/
#'
#' @param plot_map Logical value (FALSE or TRUE). If plot_map = TRUE, the ARPA Lombardia zoning is represented
#' on a map, if plot_mat = FALSE only the geometry (polygon shapefile) is stored in the output.
#' Default is plot_map = TRUE.
#' @param title Title of the plot. Deafult is 'ARPA Lombardia zoning'
#' @param line_type Linetype for the zones' borders. Default is 1.
#' @param line_size Size of the line for the zones. Default is 1.
#' @param xlab x-axis label. Default is 'Longitude'.
#' @param ylab y-axis label. Default is 'Latitude'.
#'
#' @return The function returns an object of class 'sf' containing the polygon borders of the seven zones used by
#' ARPA Lombardia to classify the regional territory. If plot_map = 1, it also returns a map of the zoning.
#'
#' @examples
#' zones <- get_ARPA_Lombardia_zoning(plot_map = TRUE)
#'
#' @export

get_ARPA_Lombardia_zoning <-
  function(plot_map = TRUE, title = "ARPA Lombardia zoning", line_type = 1,
           line_size = 1, xlab = "Longitude", ylab = "Latitude") {

    ##### Define %notin%
    '%notin%' <- Negate('%in%')

    ##### Checks if by_sensor setup properly
    if (plot_map %notin% c(FALSE,TRUE)) {
      stop("Wrong setup for 'plot_map'. Use 1 or 0 or TRUE or FALSE.",
           call. = FALSE)
    }

    ##### Read zoning shapefile bundled with the package
    ### Modified on 2026-06-26: use the package-bundled ARPA zoning shapefile
    ### stored in inst/extdata, instead of checking and downloading the same
    ### zip archive from the ARPALData GitHub repository at runtime.
    temp1 <- ARPAL_extdata_path("ARPA_zoning_shape.zip")
    temp2 <- tempfile()
    unzip(zipfile = temp1, exdir = temp2)
    your_SHP_file <- list.files(temp2, pattern = ".shp$",full.names=TRUE)

    # Read and reshape the shapefile
    Zoning <- sf::read_sf(your_SHP_file) %>%
      sf::st_as_sf(crs = 4326) %>%
      dplyr::mutate(Zone = case_when(.data$COD_ZONA == "A" ~ "Urbanized Plain",
                                     .data$COD_ZONA == "Agg_BG" ~ "Metropolitan area of Bergamo",
                                     .data$COD_ZONA == "Agg_BS" ~ "Metropolitan area of Brescia",
                                     .data$COD_ZONA == "Agg_MI" ~ "Metropolitan area of Milano",
                                     .data$COD_ZONA == "B" ~ "Rural Plain",
                                     .data$COD_ZONA == "C" ~ "Mountain",
                                     .data$COD_ZONA == "D" ~ "Valley floor")) %>%
      dplyr::select(Cod_Zone = .data$COD_ZONA, .data$Zone, .data$geometry)

    if (plot_map == 1) {
      geo_plot <- Zoning %>%
        ggplot2::ggplot() +
        ggplot2::geom_sf(aes(fill = .data$Zone),linetype = line_type, size = line_size) +
        ggplot2::labs(title = title, x = xlab, y = ylab) +
        ggplot2::theme_bw() +
        ggplot2::theme(legend.position="bottom") +
        ggplot2::guides(fill=guide_legend(nrow=2,byrow=TRUE)) +
        ggplot2::scale_x_continuous(labels = function(x) paste0(x, '\u00B0', "E")) +
        ggplot2::scale_y_continuous(labels = function(x) paste0(x, '\u00B0', "N"))
      print(geo_plot)
    }

    file.remove(your_SHP_file)

    attr(Zoning, "class") <- c("ARPALdf","tbl_df","tbl","data.frame","sf")

    return(Zoning)
  }

#' @keywords internal
#' @noRd

ARPAL_read_socrata_json <- function(url,
                                    where = NULL,
                                    page_size = 10000000,
                                    max_pages = Inf,
                                    api_key_id = Sys.getenv("ARPALDATA_SOCRATA_APP_TOKEN", unset = Sys.getenv("ARPALDATA_SODATA_APP_TOKEN", unset = ""))) {

  ### Modified on 2026-06-15: replace socratadata/RSocrata downloads with a
  ### minimal httr2 + jsonlite reader for the Socrata/SODA JSON API. The public
  ### ARPALData functions keep their original URL-building logic; this helper
  ### only executes the request, handles pagination, parses JSON, and returns a
  ### tibble with the raw Socrata columns expected by downstream code.

  ### Modified on 2026-06-15: set the default SODA page size to 10000000 as
  ### requested for large ARPA Lombardia downloads. If an endpoint contains more
  ### rows than page_size, the loop below still uses $offset to retrieve the
  ### remaining pages.

  ### Modified on 2026-06-15: read an optional Socrata app token from
  ### ARPALDATA_SOCRATA_APP_TOKEN, while keeping backward compatibility with
  ### ARPALDATA_SODATA_APP_TOKEN. The package works without a token because the
  ### ARPA Lombardia datasets are public; the token only helps users who perform
  ### large or frequent downloads.

  ### Normalize the URL supplied by the existing ARPALData URL-mapping functions.
  ### Older code may still generate .csv endpoints, whereas the new downloader
  ### reads the JSON API. Keeping the conversion here avoids changing the public
  ### functions and their internal URL builders.
  url <- as.character(url)[1]
  url <- stringr::str_replace(url, "\\.csv", ".json")

  ### Separate the base resource URL from an optional query string. Some internal
  ### functions already append a $where clause to the URL; this helper preserves
  ### that behavior by extracting and reusing the existing condition.
  url_split <- strsplit(url, "\\?", fixed = FALSE)[[1]]
  base_url <- url_split[1]
  query_string <- if (length(url_split) > 1) paste(url_split[-1], collapse = "?") else ""

  ### Decode the query string so that $where clauses encoded in URLs remain
  ### readable and can be passed back to httr2::req_url_query().
  query_parts <- if (nzchar(query_string)) strsplit(query_string, "&", fixed = TRUE)[[1]] else character()
  query_parts_decoded <- utils::URLdecode(query_parts)

  ### If the caller did not pass an explicit where argument, reuse the first
  ### $where/where clause found in the URL. This preserves the pre-existing
  ### package design, where filtering by sensor/date/year is often constructed
  ### upstream.
  where_from_url <- query_parts_decoded[grepl("^\\$where=|^where=", query_parts_decoded)]
  if (length(where_from_url) > 0 && is.null(where)) {
    where <- sub("^\\$?where=", "", where_from_url[1])
  }

  ### Storage for downloaded pages. Socrata returns at most $limit rows per
  ### request, so pagination is handled explicitly through $offset.
  out <- list()
  page <- 1
  offset <- 0

  repeat {

    ### Build a single SODA request. req_error(FALSE) prevents httr2 from
    ### stopping immediately on HTTP 4xx/5xx responses, allowing the function to
    ### print a package-specific error message containing the URL and response
    ### body.
    req <- httr2::request(base_url)
    req <- httr2::req_url_query(req, `$limit` = page_size, `$offset` = offset)
    req <- httr2::req_error(req, is_error = function(resp) FALSE)

    ### Add the server-side SoQL filter only when present. Empty where clauses
    ### are intentionally ignored so that metadata endpoints and unfiltered
    ### resource requests remain valid.
    if (!is.null(where) && nzchar(where)) {
      req <- httr2::req_url_query(req, `$where` = where)
    }

    ### Use the optional Socrata app token only when the user provides one. No
    ### secret or private credential is hard-coded in the package.
    if (!is.null(api_key_id) && nzchar(api_key_id)) {
      req <- httr2::req_headers(req, `X-App-Token` = api_key_id)
    }

    ### Modified on 2026-06-15: retry transient HTTP failures (for example,
    ### connection resets from the public Open Data Lombardia endpoint) before
    ### failing the request. These errors are not deterministic and may occur
    ### during examples or CRAN checks when public endpoints are under load.
    resp <- NULL
    req_error <- NULL
    for (req_attempt in seq_len(3)) {
      resp <- tryCatch(
        httr2::req_perform(req),
        error = function(e) {
          req_error <<- e
          NULL
        }
      )
      if (!is.null(resp)) {
        break
      }
      if (req_attempt < 3) {
        Sys.sleep(2 * req_attempt)
      }
    }
    if (is.null(resp)) {
      stop(
        paste0(
          "ARPA Lombardia / Socrata API request failed after 3 attempts.\n\nURL:\n",
          req$url, "\n\nError:\n", conditionMessage(req_error)
        ),
        call. = FALSE
      )
    }

    ### Read the HTTP status and response body. The body is kept as text until
    ### after status checks so that API error messages can be reported verbatim.
    status <- httr2::resp_status(resp)
    body <- httr2::resp_body_string(resp)

    ### Modified on 2026-06-15: fall back to an unauthenticated public request
    ### if the optional app token is invalid. This makes user-side token mistakes
    ### non-fatal for public data downloads.
    if (status == 403 && grepl("Invalid app_token", body, fixed = TRUE)) {
      req <- httr2::request(base_url)
      req <- httr2::req_url_query(req, `$limit` = page_size, `$offset` = offset)
      req <- httr2::req_error(req, is_error = function(resp) FALSE)

      if (!is.null(where) && nzchar(where)) {
        req <- httr2::req_url_query(req, `$where` = where)
      }

      ### Retry the unauthenticated fallback request as well, because transient
      ### connection resets may also occur here.
      resp <- NULL
      req_error <- NULL
      for (req_attempt in seq_len(3)) {
        resp <- tryCatch(
          httr2::req_perform(req),
          error = function(e) {
            req_error <<- e
            NULL
          }
        )
        if (!is.null(resp)) {
          break
        }
        if (req_attempt < 3) {
          Sys.sleep(2 * req_attempt)
        }
      }
      if (is.null(resp)) {
        stop(
          paste0(
            "ARPA Lombardia / Socrata API request failed after 3 attempts.\n\nURL:\n",
            req$url, "\n\nError:\n", conditionMessage(req_error)
          ),
          call. = FALSE
        )
      }
      status <- httr2::resp_status(resp)
      body <- httr2::resp_body_string(resp)
    }

    ### Stop on HTTP errors after the invalid-token fallback has been attempted.
    ### Including the full URL and body is useful for debugging changed Socrata
    ### field names, invalid SoQL clauses, or retired datasets.
    if (status >= 400) {
      stop(
        paste0(
          "ARPA Lombardia / Socrata API error. HTTP status: ", status,
          "\n\nURL:\n", req$url,
          "\n\nBody:\n", body
        ),
        call. = FALSE
      )
    }

    ### Parse the JSON page. jsonlite::fromJSON(flatten = TRUE) returns a data
    ### frame for ordinary Socrata row arrays and an empty object when the query
    ### returns no rows.
    tmp <- jsonlite::fromJSON(body, flatten = TRUE)

    ### No rows in the current page means that the download is complete. This
    ### also covers genuinely empty query results.
    if (length(tmp) == 0 || nrow(tmp) == 0) {
      break
    }

    ### Store the page as a tibble while preserving raw Socrata column names.
    ### Downstream ARPALData functions perform all semantic renaming and type
    ### conversion.
    out[[page]] <- tibble::as_tibble(tmp)

    ### If fewer rows than page_size are returned, this is the last page.
    if (nrow(tmp) < page_size) {
      break
    }

    ### max_pages is primarily a safety/debugging argument. In normal package
    ### use it remains Inf.
    if (page >= max_pages) {
      break
    }

    ### Move to the next SODA page.
    offset <- offset + page_size
    page <- page + 1
  }

  ### Modified on 2026-06-15: return an empty tibble with the raw Socrata
  ### columns expected by downstream ARPALData processing functions when an
  ### endpoint/query returns no rows. This prevents empty API responses from
  ### dropping required columns such as idsensore, data, valore, and stato.
  if (length(out) == 0) {
    return(tibble::tibble(
      idsensore = character(),
      data = character(),
      valore = character(),
      stato = character(),
      operatore = character(),
      ### Modified on 2026-06-15: include idoperatore for municipal AQ endpoints,
      ### which use this raw column in downstream processing.
      idoperatore = character()
    ))
  }

  ### Combine all downloaded pages into a single tibble.
  dplyr::bind_rows(out)
}

#' @keywords internal
#' @noRd

Wind_averaging <- function(ws,wd) {
  # Calculate the u and v wind components
  u_wind <- - ws * sin(2 * pi * wd/360)
  v_wind <- - ws * cos(2 * pi * wd/360)
  # Calculate the average wind vectors
  mean_u <- mean(u_wind, na.rm = T)
  mean_v <- mean(v_wind, na.rm = T)
  # Calculate the resultant vector average wind direction with atan2
  wd_average <- (atan2(mean_u, mean_v) * 360/2/pi)
  wd_average <- ifelse(wd_average > 180, wd_average - 180, wd_average + 180)
  # Calculate the vector average wind speed
  ws_average <- ((mean_u^2 + mean_v^2)^0.5)
  ### Output
  return(list(Wind_speed = ws_average, Wind_direction = wd_average))
}

#' Download air quality data from ARPA Lombardia website
#'
#' @description 'get_ARPA_Lombardia_AQ_data' returns observed air quality measurements collected by
#' ARPA Lombardia ground detection system for Lombardy region in Northern Italy.
#' Available airborne pollutant concentrations are: NO2, NOx, PM10, PM2.5, Ozone, Arsenic, Benzene,
#' Benzo-a-pirene, Ammonia, Sulfur Dioxide, Black Carbon, CO, Nikel, Cadmium and Lead.
#' Data are available from 1968 and are updated up to the current date.
#' For more information, visit the section 'Monitoraggio aria' at the website:
#' https://www.dati.lombardia.it/stories/s/auv9-c2sj
#'
#' @param ID_station Numeric value. ID of the station to consider. Using ID_station = NULL, all the available
#' stations are selected. Default is ID_station = NULL.
#' @param Date_begin Character vector of the first date-time to download. Format can be either "YYYY-MM-DD" or "YYYY-MM-DD hh:mm:ss". Default is Date_begin = "2022-01-01".
#' @param Date_end Character vector of the last date-time to download. Format can be either "YYYY-MM-DD" or "YYYY-MM-DD hh:mm:ss". Default is Date_end = "2022-12-31".
#' @param Frequency Temporal aggregation frequency. It can be "hourly", "daily", "weekly", "monthly" or "yearly".
#' Default is Frequency = "hourly".
#' @param Var_vec Character vector of variables to aggregate. If NULL (default) all the variables are averaged.
#' @param Fns_vec Character vector of aggregation function to apply to the selected variables.
#' Available functions are mean, median, min, max, sum, qPP (PP-th percentile), sd, var,
#' vc (variability coefficient), skew (skewness) and kurt (kurtosis).
#' @param by_sensor Logical value (TRUE or FALSE). If 'by_sensor = TRUE', the function returns the observed concentrations
#' by sensor code, while if 'by_sensor = FALSE' (default) it returns the observed concentrations by station.
#' @param verbose Logical value (TRUE or FALSE). Toggle warnings and messages. If 'verbose = TRUE' (default) the function
#' prints on the screen some messages describing the progress of the tasks. If 'verbose = FALSE' any message about
#' the progression is suppressed.
#' @param parallel Logical value (TRUE or FALSE). If 'parallel = FALSE' (default), data downloading is performed using a sequential/serial approach and additional parameters 'parworkers' and 'parfuturetype' are ignored.
#' When 'parallel = TRUE', data downloading is performed using parallel computing through the Futureverse setting.
#' More detailed information about parallel computing in the Futureverse is available at the following websites:
#' https://future.futureverse.org/ and https://cran.r-project.org/web/packages/future.apply/vignettes/future.apply-1-overview.html
#' @param parworkers Numeric integer value. If 'parallel = TRUE' (parallel mode active), the user can declare the number of parallel workers to be activated using 'parworkers = integer number'. By default ('parworkers = NULL'), the number of active workers is half of the available local cores.
#' @param parfuturetype Character vector. If 'parallel = TRUE' (parallel mode active), the user can declare the parallel strategy to be used according to the Futureverse syntax through 'parfuturetype'. By default, the 'multisession' (background R sessions on local machine) is used. In alternative, the 'multicore' (forked R processes on local machine. Not supported by Windows and RStudio) setting can be used.
#'
#' @return A data frame of class 'data.frame' and 'ARPALdf'. The object is fully compatible with Tidyverse.
#'
#' @examples
#' \donttest{
#' ## Download hourly air quality data for 2024 at station 501.
#' get_ARPA_Lombardia_AQ_data(ID_station=501, Date_begin = "2024-01-01",
#'                            Date_end = "2024-12-31", Frequency="hourly", parallel = TRUE)
#' ## Download (parallel) monthly data for NOx and NO2 observed between May and
#' ## August 2021 for all the stations active on the network. For NOx is computed
#' ## the 25th percentile, while for NO2 is computed the maximum concentration observed.
#' get_ARPA_Lombardia_AQ_data(ID_station=NULL,Date_begin = "2021-05-01",
#'                            Date_end = "2021-08-01", Frequency="monthly",Var_vec=c("NOx","NO2"),
#'                            Fns_vec=c("q25","max"), parallel = TRUE)
#' ## Download hourly air quality data by sensor from February to June 2026 at station 501.
#' get_ARPA_Lombardia_AQ_data(ID_station=501,Date_begin = "2026-02-01 00:00:00",
#'                            Date_end = "2026-06-01 23:00:00", by_sensor = TRUE)
#' }
#'
#' @export

get_ARPA_Lombardia_AQ_data <-
  function(ID_station = NULL, Date_begin = "2022-01-01", Date_end = "2022-12-31",
           Frequency = "hourly", Var_vec = NULL, Fns_vec = NULL, by_sensor = FALSE, verbose = TRUE,
           parallel = FALSE, parworkers = NULL, parfuturetype = "multisession") {

    ### Welcome message
    if (verbose == TRUE) {
      cat("Retrieving desired ARPA Lombardia dataset: started at", as.character(Sys.time()), "\n")
    }

    ############################
    ########## Checks ##########
    ############################

    ##### Check for internet connection
    if(!curl::has_internet()) {
      message("Internet connection not available at the moment.\nPlease check your internet connection. If the problem persists, please contact the package maintainer.")
      return(invisible(NULL))
    }

    ##### Check if packages for JSON/SODA download are installed
    ### Modified on 2026-06-15: replace the socratadata dependency check with httr2/jsonlite checks.
    rlang::check_installed("httr2",
                           reason = "Package \"httr2\" must be installed to download data from ARPA Lombardia Open Database.")
    rlang::check_installed("jsonlite",
                           reason = "Package \"jsonlite\" must be installed to parse JSON data from ARPA Lombardia Open Database.")

    ##### Check if Futureverse is installed
    rlang::check_installed("future",
                           reason = "Package \"future\" must be installed to download (parallel) data from ARPA Lombardia Open Database.")
    rlang::check_installed("future.apply",
                           reason = "Package \"future\" must be installed to download (parallel) data from ARPA Lombardia Open Database.")

    ##### Define %notin%
    '%notin%' <- Negate('%in%')

    ##### Checks if by_sensor setup properly
    if (by_sensor %notin% c(0,1,FALSE,TRUE)) {
      stop("Wrong setup for 'by_sensor'. Use 1 or 0 or TRUE or FALSE.",
           call. = FALSE)
    }

    ##### Checks if parallel setup properly
    if (parallel %notin% c(FALSE,TRUE)) {
      stop("Wrong setup for 'parallel'. Use TRUE or FALSE.",
           call. = FALSE)
    }
    if (parfuturetype %notin% c("multisession","multicore")) {
      stop("Wrong setup for 'parallel'. Use TRUE or FALSE.",
           call. = FALSE)
    }


    ######################################
    ########## Downloading data ##########
    ######################################

    ### Registry/Metadata
    Metadata <- AQ_metadata_reshape()
    Metadata <- Metadata %>%
      dplyr::select(-c(.data$Altitude,.data$Province,.data$City,
                       .data$DateStart,.data$DateStop,.data$Latitude,.data$Longitude,
                       .data$ARPA_zone,.data$ARPA_stat_type))

    ### Checks if ID_station is valid (in the list of active stations)
    '%notin%' <- Negate('%in%')
    if (!is.null(ID_station) & all(ID_station %notin% Metadata$IDStation)) {
      stop("ID_station NOT in the list of active stations. Change ID_station or use ID_station = NULL",
           call. = FALSE)
    }

    if (!is.null(ID_station)) {
      Metadata <- Metadata %>%
        dplyr::filter(.data$IDStation %in% ID_station)
    }
    if (!is.null(Var_vec)) {
      Metadata <- Metadata %>%
        dplyr::filter(.data$Pollutant %in% Var_vec)
    }

    if (verbose == TRUE) {
      cat("Downloading data from ARPA Lombardia: started at", as.character(Sys.time()), "\n")
    }

    ##### Splitting strategy for improving download speed
    n_blocks <- 12
    # Check dates format
    if (is.null(lubridate::guess_formats(x = Date_begin, orders = c("ymd HMS","ymd")))) {
      stop("Wrong format for 'Date_begin'. Use 'YYYY-MM-DD' or 'YYYY-MM-DD hh:mm:ss'", call. = FALSE)
    }
    if (is.null(lubridate::guess_formats(x = Date_end, orders = c("ymd HMS","ymd")))) {
      stop("Wrong format for 'Date_end'. Use 'YYYY-MM-DD' or 'YYYY-MM-DD hh:mm:ss'", call. = FALSE)
    }
    if (is.na(lubridate::ymd_hms(Date_begin, quiet = TRUE))) {
      Date_begin <- lubridate::ymd_hms(paste0(Date_begin," 00:00:00"), tz = "CET")
    }
    if (is.na(lubridate::ymd_hms(Date_end, quiet = TRUE))) {
      Date_end <- lubridate::ymd_hms(paste0(Date_end," 23:00:00"), tz = "CET")
    }
    ### Check for the presence of breaking dates (URLs change in several years)
    if (lubridate::year(Date_begin) == lubridate::year(Date_end)) {
      break_years <- lubridate::year(Date_begin)
    } else {
      break_years <- seq(from = lubridate::year(Date_begin), to = lubridate::year(Date_end), by = 1)
    }

    ##### Check online availability of the resources for the specified years
    if (verbose == TRUE) {
      cat("Checking availability of online resources: started at", as.character(Sys.time()), "\n")
    }
    URLs <- res_check <- numeric(length = length(break_years))
    for (yr in 1:length(break_years)) {
      URLs[yr] <- url <- url_AQ_year_dataset(Stat_type = "AQ", Year = break_years[yr])
      temp <- tempfile()
      res <- suppressWarnings(try(curl::curl_fetch_disk(url, temp), silent = TRUE))
      if(res$status_code != 200) {
        message(paste0("The internet resource for year ", break_years[yr]," is not available at the moment. Status code: ",res$status_code,".\nPlease, try later. If the problem persists, please contact the package maintainer."))
      } else {
        res_check[yr] <- 1
      }
    }
    if (verbose == TRUE) {
      if (sum(res_check) == length(break_years)) {
        message("All the online resources are available.\n")
      }
      if (sum(res_check) > 0 & sum(res_check) < length(break_years)) {
        message("Part of the required online resources are not available. Please, try with a new request.\n")
        return(invisible(NULL))
      }
      if (sum(res_check) == 0) {
        message("None of the required online resources are available. Please, try with a new request.\n")
        return(invisible(NULL))
      }
    }

    ### Downloading data
    if (verbose == TRUE) {
      cat("Downloading data: started at", as.character(Sys.time()), "\n")
    }

    ### Building URLs/links in Socrata format using sequences of dates
    if (length(break_years) == 1) {
      break_dates <- paste0(break_years,"-12-31 23:00:00")
    } else {
      break_dates <- paste0(break_years[-length(break_years)],"-12-31 23:00:00")
    }
    dates_seq_end <- c(Date_begin,lubridate::ymd_hms(break_dates),Date_end)
    dates_seq_end <- unique(dates_seq_end)
    dates_seq_end <- dates_seq_end[-1]
    dates_seq_end <- dates_seq_end[dates_seq_end <= Date_end & dates_seq_end >= Date_begin]
    dates_seq_begin <- c(Date_begin,lubridate::ymd_hms(break_dates) + lubridate::hours(1))
    dates_seq_begin <- unique(dates_seq_begin)
    dates_seq_begin <- dates_seq_begin[dates_seq_begin <= Date_end & dates_seq_begin >= Date_begin]
    URL_blocks <- vector(mode = "list", length = length(dates_seq_begin))
    for (b in 1:length(dates_seq_begin)) {
      seq_temp <- seq(lubridate::as_datetime(dates_seq_begin[b]),
                      lubridate::as_datetime(dates_seq_end[b]),
                      length.out = n_blocks + 1)

      seq_begin_temp <- seq_temp[-length(seq_temp)]
      seq_begin_temp <- lubridate::round_date(seq_begin_temp, unit = "hour")
      seq_begin_temp <- stringr::str_replace(string = seq_begin_temp, pattern = " ", replacement = "T")

      seq_end_temp <- c(seq_temp[-c(1,length(seq_temp))] - lubridate::hours(1),seq_temp[length(seq_temp)])
      seq_end_temp <- lubridate::round_date(seq_end_temp, unit = "hour")
      seq_end_temp <- stringr::str_replace(string = seq_end_temp, pattern = " ", replacement = "T")

      if (is.null(ID_station)) {
        str_sensor <- NULL
      } else {
        str_sensor <- paste0("AND idsensore in(",paste0(sapply(X = Metadata$IDSensor, function(x) paste0("'",x,"'")),collapse = ","),")")
      }

      URL_blocks[[b]] <- data.frame(seq_begin_temp,seq_end_temp) %>%
        dplyr::mutate(link = paste0(URLs[b],"?$where=data between '", seq_begin_temp, "' and '", seq_end_temp,"'", str_sensor)) %>%
        dplyr::select(.data$link)
    }
    URL_blocks <- dplyr::bind_rows(URL_blocks)

    ### Preparing parallel computation: explicitly open multisession/multicore workers by switching plan
    if (parallel == TRUE) {
      # Explicitly open multisession/multicore workers by switching plan
      future::plan(future::multisession, workers = 12)
      if (is.null(parworkers)) {
        parworkers <- future::availableCores()/2
      }
      eval(parse(text = paste0("future::plan(future::",parfuturetype,", workers = ",parworkers,")")))
      if (verbose == TRUE) {
        message("Start parallel computing: number of parallel workers = ", future::nbrOfWorkers())
      }
    }

    ### Download using Socrata API
    ### Modified on 2026-06-15: replace socratadata::soc_read with the internal JSON/SODA reader.
    Aria <- dplyr::bind_rows(
      future.apply::future_apply(X = as.matrix(URL_blocks), MARGIN = 1, FUN = function(x) {
        ARPAL_read_socrata_json(url = as.character(x))
      })
    )

    ### Ending parallel computation: explicitly close multisession/multicore workers by switching plan
    if (parallel == TRUE) {
      # Explicitly close multisession/multicore workers by switching plan
      future::plan(future::sequential)
      if (verbose == TRUE) {
        message("Stop parallel computing: number of parallel workers = ", future::nbrOfWorkers())
      }
    }



    #####################################
    ########## Processing data ##########
    #####################################
    if (verbose == TRUE) {
      cat("Processing data: started at", as.character(Sys.time()), "\n")
    }

    ### Change variable names
    Aria <- Aria %>%
      dplyr::select(IDSensor = .data$idsensore,
                    Date = .data$data,
                    Value = .data$valore) %>%
      dplyr::mutate(IDSensor = as.numeric(.data$IDSensor),
                    ### Modified on 2026-06-15: parse Socrata JSON datetime strings so AQ by_sensor outputs expose POSIXct dates.
                    Date = ARPAL_parse_socrata_datetime(.data$Date),
                    Value = as.numeric(.data$Value))
    # clean RAM
    invisible(gc())

    ### Add metadata
    Aria <- dplyr::right_join(Aria,Metadata, by = "IDSensor")

    ### Cleaning
    if (by_sensor %in% c(1,TRUE)) {
      Aria <- Aria %>%
        dplyr::filter(!is.na(.data$Date)) %>%
        dplyr::mutate(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric),
                                    ~ dplyr::na_if(.,-9999))) %>%
        dplyr::mutate(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric),
                                    ~ dplyr::na_if(.,NaN))) %>%
        dplyr::select(.data$Date,.data$IDStation,.data$NameStation,.data$IDSensor,
                      .data$Pollutant,.data$Value)
    } else if (by_sensor %in% c(0,FALSE)) {
      Aria <- Aria %>%
        dplyr::filter(!is.na(.data$Date)) %>%
        dplyr::select(-c(.data$IDSensor)) %>%
        dplyr::mutate(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric),
                                    ~ dplyr::na_if(.,-9999))) %>%
        dplyr::mutate(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric),
                                    ~ dplyr::na_if(.,NaN))) %>%
        tidyr::pivot_wider(names_from = .data$Pollutant, values_from = .data$Value,
                           values_fn = function(x) mean(x,na.rm=T)) # Mean (without NA) of a NA vector = NaN
    }

    ### Modified on 2026-06-15: qualify dplyr functions to avoid R CMD check
    ### notes on undefined global functions.
    Aria <- Aria %>%
      dplyr::mutate(
        dplyr::across(tidyselect::vars_select_helpers$where(is.numeric),
                      function(x) dplyr::na_if(x = x, y = NaN)),
        dplyr::across(tidyselect::vars_select_helpers$where(is.numeric),
                      function(x) dplyr::na_if(x = x, y = NA))
      )
    # Aria[is.na(Aria)] <- NA
    # Aria[is.nan_df(Aria)] <- NA
    # clean RAM
    invisible(gc())

    ### Add dataset attributes
    attr(Aria, "class") <- c("ARPALdf","tbl_df","tbl","data.frame")

    ###
    if (is.null(Var_vec) & is.null(Fns_vec)) {
      vv <- c("Ammonia","Arsenic","Benzene","Benzo_a_pyrene","BlackCarbon","Cadmium",
              "CO","Lead","Nikel","NO","NO2","NOx","Ozone","PM_tot","PM10","PM2.5","Sulfur_dioxide")
      vv <- vv[vv %in% names(Aria)]
      fv <- rep("mean",length(vv))
    } else {
      vv <- Var_vec
      fv <- Fns_vec
    }

    # Checks if all the pollutants are available for the selected stations
    if (all(dplyr::all_of(vv) %in% names(Aria)) == FALSE) {
      stop("One or more pollutants are not available for the selected stations! Change the values of 'Var_vec'",
           call. = FALSE)
    }

    if (by_sensor %in% c(0,FALSE)) {
      ### Aggregating dataset across time
      if (Frequency != "hourly") {
        if (verbose == TRUE) {
          cat("Aggregating ARPA Lombardia data: started at", as.character(Sys.time()), "\n")
        }
        Aria <- Aria %>%
          Time_aggregate(Frequency = Frequency, Var_vec = Var_vec, Fns_vec = Fns_vec, verbose = verbose) %>%
          dplyr::arrange(.data$NameStation, .data$Date)
      } else {
        Aria <- Aria %>%
          dplyr::select(.data$Date,.data$IDStation,.data$NameStation,vv) %>%
          dplyr::arrange(.data$NameStation, .data$Date)
      }

      ### Regularizing dataset: same number of timestamps for each station and variable
      if (verbose == TRUE) {
        cat("Regularizing ARPA Lombardia data: started at", as.character(Sys.time()), "\n")
      }
      freq_unit <- dplyr::case_when(Frequency == "hourly" ~ "hours",
                                    Frequency == "daily" ~ "days",
                                    Frequency == "weekly" ~ "weeks",
                                    Frequency == "monthly" ~ "months",
                                    Frequency == "yearly" ~ "years")

      Aria <- Aria %>%
        dplyr::arrange(.data$Date) %>%
        dplyr::filter(.data$Date >= Date_begin,
                      .data$Date <= Date_end) %>%
        tidyr::pivot_longer(cols = -c(.data$Date,.data$IDStation,.data$NameStation),
                            names_to = "Measure", values_to = "Value") %>%
        dplyr::mutate(Date = case_when(Frequency %in% c("hourly") ~ as.character(format(x = .data$Date, format = "%Y-%m-%d %H:%M:%S")),
                                       TRUE ~ as.character(format(x = .data$Date, format = "%Y-%m-%d")))) %>%
        tidyr::pivot_wider(names_from = .data$Date, values_from = .data$Value) %>%
        tidyr::pivot_longer(cols = -c(.data$Measure,.data$IDStation,.data$NameStation),
                            names_to = "Date", values_to = "Value") %>%
        tidyr::pivot_wider(names_from = .data$Measure, values_from = .data$Value)

      if (Frequency != "hourly") {
        Aria <- Aria %>%
          dplyr::mutate(Date = lubridate::ymd(.data$Date)) %>%
          dplyr::arrange(.data$IDStation,.data$Date)
      } else {
        Aria <- Aria %>%
          dplyr::mutate(Date = lubridate::ymd_hms(.data$Date)) %>%
          dplyr::arrange(.data$IDStation,.data$Date)
      }

      structure(list(Aria = Aria))
      attr(Aria, "class") <- c("ARPALdf","ARPALdf_AQ","tbl_df","tbl","data.frame")
      attr(Aria, "frequency") <- Frequency
      attr(Aria, "units") <- freq_unit
    } else if (by_sensor %in% c(1,TRUE)) {
      Aria <- Aria %>%
        dplyr::arrange(.data$Date) %>%
        dplyr::filter(.data$Date >= Date_begin,
                      .data$Date <= Date_end)
      structure(list(Aria = Aria))
      attr(Aria, "class") <- c("ARPALdf","ARPALdf_AQ","tbl_df","tbl","data.frame")
      attr(Aria, "frequency") <- "hourly"
      attr(Aria, "units") <- "hours"
    }

    if (verbose == TRUE) {
      cat("Processing data: ended at", as.character(Sys.time()), "\n")
    }

    if (verbose == TRUE) {
      cat("Retrieving desired ARPA Lombardia dataset: ended at", as.character(Sys.time()), "\n")
    }

    return(Aria)
  }

#' Generate a map of the selected stations
#'
#' @description 'get_ARPA_Lombardia_AQ_data' represents on a map (geometries/polygon of Lombardy) the location of
#' the stations contained in a data frame of class 'ARPALdf'. Data can be either a ARPALdf of observed data
#' (from 'get_ARPA_Lombardia_xxx' commands) and an ARPALdf obtained as registry
#' (from 'get_ARPA_Lombardia_xxx_registry' command).
#'
#' @param data Dataset of class 'ARPALdf' containing the stations to plot on the map. It can be either a
#' ARPALdf of observed data (from 'get_ARPA_Lombardia_xxx' commands) and an ARPALdf obtained as registry
#' (from 'get_ARPA_Lombardia_xxx_registry' command).
#' @param title Title of the plot. Deafult is 'Map of ARPA stations in Lombardy'
#' @param prov_line_type Linetype for Lombardy provinces. Default is 1.
#' @param prov_line_size Size of the line for Lombardy provinces. Default is 1.
#' @param col_points Color of the points. Default is 'blue'.
#' @param xlab x-axis label. Default is 'Longitude'.
#' @param ylab y-axis label. Default is 'Latitude'.
#'
#' @return A map of selected stations across the Lombardy region
#'
#' @examples
#' \donttest{
#' ## Map network from a dataset of measurements
#' # Download daily concentrations observed at all the stations in January 2020.
#' d <- get_ARPA_Lombardia_AQ_data(ID_station = NULL, Date_begin = "2020-01-01",
#'                                 Date_end = "2020-01-31", Frequency = "daily")
#' # Map the stations included in 'd'
#' map_Lombardia_stations(data = d, title = "Air quality stations in Lombardy")
#' ## Map network from a registry dataset
#' # Download registry for all the AQ stations in 2020.
#' r <- get_ARPA_Lombardia_AQ_registry()
#' # Map the stations included in 'r'
#' map_Lombardia_stations(data = r, title = "Air quality stations in Lombardy")
#' }
#'
#' @export

map_Lombardia_stations <-
  function(data, title = "Map of ARPA stations in Lombardy", prov_line_type = 1,
           prov_line_size = 1, col_points = "blue",
           xlab = "Longitude", ylab = "Latitude") {

    ##### Check for internet connection
    if(!curl::has_internet()) {
      message("Internet connection not available at the moment.\nPlease check your internet connection. If the problem persists, please contact the package maintainer.")
      return(invisible(NULL))
    }

    ##### Retrieve shapefile from Eurostat
    Lombardia <- get_Lombardia_geospatial(NUTS_level = "NUTS3")
    if (is.null(Lombardia)) {
      message("The map will not include the ground layer with Lombardy's shapefile. Only points/coordinates will be plot.")
    }

    ##### Data manipulation
    if(is_ARPALdf_AQ(Data = data) == T) {
      Stats <- get_ARPA_Lombardia_AQ_registry()
    } else if (is_ARPALdf_W(Data = data) == T) {
      Stats <- get_ARPA_Lombardia_W_registry()
    }

    data <- data.frame(IDStation = unique(data$IDStation))
    d <- dplyr::left_join(data,Stats,by="IDStation")
    d <- d %>%
      sf::st_as_sf(coords = c("Longitude", "Latitude"),crs = 4326)

    ##### Mapping
    geo_plot <- Lombardia %>%
      ggplot2::ggplot() +
      ggplot2::geom_sf(linetype = prov_line_type, size = prov_line_size) +
      ggplot2::geom_sf(data = d, col=col_points) +
      ggplot2::labs(title = title) +
      ggplot2::theme_bw() +
      ggplot2::scale_x_continuous(labels = function(x) paste0(x, '\u00B0', "E")) +
      ggplot2::scale_y_continuous(labels = function(x) paste0(x, '\u00B0', "N"))

    print(geo_plot)
  }

#' Check if a given object belongs to class 'ARPALdf_AQ'
#'
#' @description 'is_ARPALdf_AQ' checks if the input object belongs to the class 'ARPALdf_AQ'
#'
#' @param Data Object to check if the class of a dataframe is 'ARPALdf_AQ', i.e. ARPAL
#' dataframe for air quality data.
#'
#' @return The function returns 'True' if the object is of class 'ARPALdf_AQ' and it returns 'False' if the
#'     object isn't of class 'ARPALdf_AQ'
#'
#' @examples
#' d <- get_ARPA_Lombardia_AQ_registry()
#' is_ARPALdf_AQ(d)
#'
#' @export

is_ARPALdf_AQ <- function(Data) {
  is.element("ARPALdf_AQ", attr(Data,"class"))
}

#' @keywords internal
#' @noRd

Custom_summarise <- function(grouped_data,var_vec,fns_vec) {

  '%notin%' <- Negate('%in%')

  # Wind direction can only be averaged
  if(sum(var_vec %in% c("Wind_direction","Wind_direction_gust") & fns_vec != "mean") > 0) {
    stop("Error: on the Wind_direction and Wind_direction_gust is possible to calculate only the average value. Use 'mean' in 'Fns_vec.'",
         call. = FALSE)
  }

  # Wind speed can only be averaged, maximized or minimized
  if(sum(var_vec %in% c("Wind_speed","Wind_speed_gust") & fns_vec %notin% c("mean","min","max")) > 0) {
    stop("Error: on the Wind_speed and Wind_speed_gust is possible to calculate only mean, max or min values. Use 'mean' or 'max' or 'min' in 'Fns_vec.'",
         call. = FALSE)
  }

  # Checks if all the selected variables are available for the actual dataset
  if (all(dplyr::all_of(var_vec) %in% names(grouped_data)) == F) {
    stop("Error: one or more measures are not available for the selected stations! Change the values of 'Var_vec'",
         call. = FALSE)
  }

  vv_vec <- (duplicated(var_vec,fromLast = T) | duplicated(var_vec,fromLast = F))

  summ_data <- grouped_data %>%
    dplyr::summarise(dplyr::across(var_vec[vv_vec & var_vec %in% c("Wind_speed","Wind_speed_gust") & fns_vec=="mean"], ~ Wind_averaging(Wind_speed,Wind_direction)$Wind_speed,.names = "{.col}_mean"),
                     dplyr::across(var_vec[!vv_vec & var_vec %in% c("Wind_speed","Wind_speed_gust") & fns_vec=="mean"], ~ Wind_averaging(Wind_speed,Wind_direction)$Wind_speed),
                     dplyr::across(var_vec[vv_vec & var_vec %in% c("Wind_direction","Wind_direction_gust") & fns_vec=="mean"], ~ Wind_averaging(Wind_speed,Wind_direction)$Wind_direction,.names = "{.col}_mean"),
                     dplyr::across(var_vec[!vv_vec & var_vec %in% c("Wind_direction","Wind_direction_gust") & fns_vec=="mean"], ~ Wind_averaging(Wind_speed,Wind_direction)$Wind_direction),
                     dplyr::across(var_vec[vv_vec & var_vec %notin% c("Wind_speed","Wind_direction","Wind_speed_gust","Wind_direction_gust") & fns_vec=="mean"], ~ mean(.x, na.rm=T),.names = "{.col}_mean"),
                     dplyr::across(var_vec[!vv_vec & var_vec %notin% c("Wind_speed","Wind_direction","Wind_speed_gust","Wind_direction_gust") & fns_vec=="mean"], ~ mean(.x, na.rm=T)),
                     dplyr::across(var_vec[vv_vec & var_vec %notin% c("Wind_speed","Wind_direction","Wind_speed_gust","Wind_direction_gust") & fns_vec=="median"], ~ median(.x, na.rm=T),.names = "{.col}_median"),
                     dplyr::across(var_vec[!vv_vec & var_vec %notin% c("Wind_speed","Wind_direction","Wind_speed_gust","Wind_direction_gust") & fns_vec=="median"], ~ median(.x, na.rm=T)),
                     dplyr::across(var_vec[vv_vec & var_vec %notin% c("Wind_speed","Wind_direction","Wind_speed_gust","Wind_direction_gust") & grepl('\\bq[0-9]+$',fns_vec)],
                                     list(!!!quantilep(.data$.x,fns_vec[vv_vec & var_vec %notin% c("Wind_speed","Wind_direction","Wind_speed_gust","Wind_direction_gust") & grepl('\\bq[0-9]+$',fns_vec)])),
                                   .names = "{.col}_{.fn}"),
                     dplyr::across(var_vec[!vv_vec & var_vec %notin% c("Wind_speed","Wind_direction","Wind_speed_gust","Wind_direction_gust") & grepl('\\bq[0-9]+$',fns_vec)],
                                     list(!!!quantilep(.data$.x,fns_vec[!vv_vec & var_vec %notin% c("Wind_speed","Wind_direction","Wind_speed_gust","Wind_direction_gust") & grepl('\\bq[0-9]+$',fns_vec)]))),
                     dplyr::across(var_vec[vv_vec & var_vec %notin% c("Wind_speed","Wind_direction","Wind_speed_gust","Wind_direction_gust") & fns_vec=="sum"], ~ sum(.x, na.rm=T),.names = "{.col}_cum"),
                     dplyr::across(var_vec[!vv_vec & var_vec %notin% c("Wind_speed","Wind_direction","Wind_speed_gust","Wind_direction_gust") & fns_vec=="sum"], ~ sum(.x, na.rm=T)),
                     dplyr::across(var_vec[vv_vec & var_vec %notin% c("Wind_direction","Wind_direction_gust") & fns_vec=="min"], ~ min(.x, na.rm=T),.names = "{.col}_min"),
                     dplyr::across(var_vec[!vv_vec & var_vec %notin% c("Wind_direction","Wind_direction_gust") & fns_vec=="min"], ~ min(.x, na.rm=T)),
                     dplyr::across(var_vec[vv_vec & var_vec %notin% c("Wind_direction","Wind_direction_gust") & fns_vec=="max"], ~ max(.x, na.rm=T),.names = "{.col}_max"),
                     dplyr::across(var_vec[!vv_vec & var_vec %notin% c("Wind_direction","Wind_direction_gust") & fns_vec=="max"], ~ max(.x, na.rm=T)),
                     dplyr::across(var_vec[vv_vec & var_vec %notin% c("Wind_speed","Wind_direction","Wind_speed_gust","Wind_direction_gust") & fns_vec=="sd"], ~ sd(.x, na.rm=T),.names = "{.col}_sd"),
                     dplyr::across(var_vec[!vv_vec & var_vec %notin% c("Wind_speed","Wind_direction","Wind_speed_gust","Wind_direction_gust") & fns_vec=="sd"], ~ sd(.x, na.rm=T)),
                     dplyr::across(var_vec[vv_vec & var_vec %notin% c("Wind_speed","Wind_direction","Wind_speed_gust","Wind_direction_gust") & fns_vec=="var"], ~ var(.x, na.rm=T),.names = "{.col}_var"),
                     dplyr::across(var_vec[!vv_vec & var_vec %notin% c("Wind_speed","Wind_direction","Wind_speed_gust","Wind_direction_gust") & fns_vec=="var"], ~ var(.x, na.rm=T)),
                     dplyr::across(var_vec[vv_vec & var_vec %notin% c("Wind_speed","Wind_direction","Wind_speed_gust","Wind_direction_gust") & fns_vec=="vc"], ~ sd(.x, na.rm=T)/mean(.x, na.rm=T),.names = "{.col}_vc"),
                     dplyr::across(var_vec[!vv_vec & var_vec %notin% c("Wind_speed","Wind_direction","Wind_speed_gust","Wind_direction_gust") & fns_vec=="vc"], ~ sd(.x, na.rm=T)/mean(.x, na.rm=T)),
                     dplyr::across(var_vec[vv_vec & var_vec %notin% c("Wind_speed","Wind_direction","Wind_speed_gust","Wind_direction_gust") & fns_vec=="skew"], ~ mom_skew(.x, na.rm=T),.names = "{.col}_skew"),
                     dplyr::across(var_vec[!vv_vec & var_vec %notin% c("Wind_speed","Wind_direction","Wind_speed_gust","Wind_direction_gust") & fns_vec=="skew"], ~ mom_skew(.x, na.rm=T)),
                     dplyr::across(var_vec[vv_vec & var_vec %notin% c("Wind_speed","Wind_direction","Wind_speed_gust","Wind_direction_gust") & fns_vec=="kurt"], ~ mom_kurt(.x, na.rm=T),.names = "{.col}_kurt"),
                     dplyr::across(var_vec[!vv_vec & var_vec %notin% c("Wind_speed","Wind_direction","Wind_speed_gust","Wind_direction_gust") & fns_vec=="kurt"], ~ mom_kurt(.x, na.rm=T))) %>%
    dplyr::mutate(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric), ~ ifelse(is.infinite(.x), NA, .x)))

  # Drop non-useful columns concerning the quantiles
  if (sum(grepl('\\q[0-9]+$',names(summ_data))) > 0) {
    '%notin%' <- Negate('%in%')
    names_full <- names(summ_data)[grepl('\\q[0-9]+$',names(summ_data))]
    to_drop <- names_full[names_full %notin% paste0(var_vec,"_",fns_vec)]
    summ_data <- summ_data %>%
      dplyr::select(-to_drop)
  }

  return(summ_data)
}

#' @keywords internal
#' @noRd

ARPAL_parse_socrata_date <- function(x) {

  ### Modified on 2026-06-15: parse Socrata JSON date/datetime strings
  ### robustly and return a Date vector. Open Data Lombardia/Socrata
  ### endpoints may encode daily fields either as simple dates
  ### (for example, "2026-03-01") or as full timestamps
  ### (for example, "2026-03-01T00:00:00.000"). Using lubridate::ymd()
  ### alone would turn full timestamps into NA in some code paths.

  ### Convert the input to character first. This keeps the helper stable when
  ### the API parser returns character, Date, POSIXct, numeric-looking strings,
  ### or empty vectors.
  x <- as.character(x)

  ### First try the most informative parser, lubridate::ymd_hms(). This handles
  ### the standard Socrata datetime representation returned by JSON endpoints.
  ### Warnings are suppressed because failures are expected for pure YYYY-MM-DD
  ### strings and are handled in the fallback step below.
  out <- suppressWarnings(lubridate::ymd_hms(x, tz = "UTC", quiet = TRUE))

  ### Identify values that were not parsed as datetimes but are still non-empty
  ### input strings. These are usually pure dates and can be safely parsed with
  ### lubridate::ymd().
  idx <- is.na(out) & !is.na(x) & nzchar(x)

  ### Fallback parser for pure dates. The result is temporarily converted to
  ### POSIXct so that it can be assigned into the POSIXct vector created above.
  if (any(idx)) {
    out[idx] <- suppressWarnings(
      as.POSIXct(lubridate::ymd(x[idx], quiet = TRUE), tz = "UTC")
    )
  }

  ### Downstream daily/municipal metadata routines expect Date objects, not
  ### POSIXct objects. The final coercion also strips any irrelevant midnight
  ### time component returned by Socrata.
  as.Date(out)
}

#' @keywords internal
#' @noRd

W_download_past <- function(Metadata,Date_begin,Date_end,verbose = TRUE,
                            parallel = FALSE, parworkers = NULL, parfuturetype = "multisession") {
  ##### Splitting strategy for improving download speed
  n_blocks <- 12
  ### Modified on 2026-06-15: helper returning the raw weather columns
  ### expected downstream when no historical weather rows are available.
  W_empty_raw <- function() {
    tibble::tibble(
      idsensore = character(),
      data = character(),
      valore = character(),
      stato = character(),
      operatore = character(),
      idoperatore = character()
    )
  }

  # Check dates format
  if (is.null(lubridate::guess_formats(x = Date_begin, orders = c("ymd HMS","ymd")))) {
    stop("Wrong format for 'Date_begin'. Use 'YYYY-MM-DD' or 'YYYY-MM-DD hh:mm:ss'", call. = FALSE)
  }
  if (is.null(lubridate::guess_formats(x = Date_end, orders = c("ymd HMS","ymd")))) {
    stop("Wrong format for 'Date_end'. Use 'YYYY-MM-DD' or 'YYYY-MM-DD hh:mm:ss'", call. = FALSE)
  }
  if (is.na(lubridate::ymd_hms(Date_begin, quiet = TRUE))) {
    Date_begin <- lubridate::ymd_hms(paste0(Date_begin," 00:00:00"))
  }
  if (is.na(lubridate::ymd_hms(Date_end, quiet = TRUE))) {
    Date_end <- lubridate::ymd_hms(paste0(Date_end," 23:55:00"))
  }
  ### Check for the presence of breaking dates (URLs change in several years)
  if (lubridate::year(Date_begin) == lubridate::year(Date_end)) {
    break_years <- lubridate::year(Date_begin)
  } else {
    break_years <- seq(from = lubridate::year(Date_begin), to = lubridate::year(Date_end), by = 1)
  }

  ##### Check online availability of the resources for the specified years
  if (verbose == TRUE) {
    cat("Checking availability of online resources: started at", as.character(Sys.time()), "\n")
  }

  ##### Check for wind gust (same ID codes of wind speed but different link)
  ### Modified on 2026-06-15: drop empty/NA weather measures before
  ### building parameter-year URL combinations.
  measures <- unique(as.character(Metadata$Measure))
  measures <- measures[!is.na(measures) & nzchar(measures)]

  if (any(measures == "Wind_speed")) {
    measures <- unique(c(measures,"Wind_speed_gust"))
  }

  if (length(measures) == 0 || length(break_years) == 0) {
    message("No valid ARPA Lombardia weather parameter-year combination was found.")
    ### Modified on 2026-06-15: return a structured empty object instead of
    ### invisible(NULL), so downstream dplyr code can still find raw columns.
    return(W_empty_raw())
  }

  ##### Create combinations of weather parameters and years
  url_combs <- expand.grid(measures,break_years, stringsAsFactors = FALSE)
  colnames(url_combs) <- c("WPar","Year")

  URLs <- character(length = dim(url_combs)[1])
  res_check <- numeric(length = dim(url_combs)[1])

  ### Modified on 2026-06-15: keep track of unsupported metadata measures
  ### whose historical endpoint is not mapped by url_W_year_par(). This avoids
  ### the replacement length zero error raised by NULL switch() results.
  unsupported_url <- rep(FALSE, nrow(url_combs))

  for (yr in seq_len(nrow(url_combs))) {
    ### Modified on 2026-06-15: pass scalar values explicitly to the URL map.
    url_check <- url_W_year_par(WPar = paste0(as.character(url_combs$WPar[yr]),"_check"), Year = url_combs$Year[yr])
    url_data <- url_W_year_par(WPar = as.character(url_combs$WPar[yr]), Year = url_combs$Year[yr])

    if (length(url_check) == 0 || length(url_data) == 0 || is.na(url_check) || is.na(url_data)) {
      unsupported_url[yr] <- TRUE
      if (verbose == TRUE) {
        message(paste0("Skipping unsupported ARPA Lombardia weather parameter '",
                       as.character(url_combs$WPar[yr]), "' for year ", url_combs$Year[yr], "."))
      }
      next
    }

    URLs[yr] <- url <- url_check
    temp <- tempfile()
    res <- suppressWarnings(try(curl::curl_fetch_disk(url, temp), silent = TRUE))
    url_combs$link1[yr] <- URLs[yr] <- url <- url_data
    if(inherits(res, "try-error") || is.null(res$status_code) || res$status_code != 200) {
      status_code <- if (inherits(res, "try-error") || is.null(res$status_code)) NA else res$status_code
      message(paste0("The internet resource for ",url_combs[yr,"WPar"]," in year ", url_combs[yr,"Year"]," is not available at the moment. Status code: ",status_code,".\nPlease, try later. If the problem persists, please contact the package maintainer."))
    } else {
      res_check[yr] <- 1
    }
  }

  ### Modified on 2026-06-15: drop unsupported URL combinations before
  ### building Socrata query blocks.
  if (any(unsupported_url)) {
    url_combs <- url_combs[!unsupported_url, , drop = FALSE]
    res_check <- res_check[!unsupported_url]
  }

  if (nrow(url_combs) == 0) {
    message("None of the selected ARPA Lombardia weather parameters has a mapped online resource.")
    ### Modified on 2026-06-15: return a structured empty object instead of
    ### invisible(NULL), so downstream dplyr code can still find raw columns.
    return(W_empty_raw())
  }
  if (verbose == TRUE) {
    if (sum(res_check) == length(res_check)) {
      message("All the online resources are available.\n")
    }
    if (sum(res_check) > 0 & sum(res_check) < length(res_check)) {
      message("Part of the required online resources are not available. Please, try with a new request.\n")
      return(W_empty_raw())
    }
    if (sum(res_check) == 0) {
      message("None of the required online resources is available. Please, try with a new request.\n")
      return(W_empty_raw())
    }
  }


  ### Downloading data
  if (verbose == TRUE) {
    cat("Downloading data: started at", as.character(Sys.time()), "\n")
  }


  ### Building URLs/links in Socrata format using sequences of dates
  if (length(break_years) == 1) {
    break_dates <- paste0(break_years,"-12-31 23:00:00")
  } else {
    break_dates <- paste0(break_years[-length(break_years)],"-12-31 23:00:00")
  }
  dates_seq_end <- c(Date_begin,lubridate::ymd_hms(break_dates),Date_end)
  dates_seq_end <- unique(dates_seq_end)
  dates_seq_end <- dates_seq_end[-1]
  dates_seq_end <- dates_seq_end[dates_seq_end <= Date_end & dates_seq_end >= Date_begin]
  dates_seq_begin <- c(Date_begin,lubridate::ymd_hms(break_dates) + lubridate::hours(1))
  dates_seq_begin <- unique(dates_seq_begin)
  dates_seq_begin <- dates_seq_begin[dates_seq_begin <= Date_end & dates_seq_begin >= Date_begin]

  URL_blocks <- vector(mode = "list", length = dim(url_combs)[1])
  for (i in 1:dim(url_combs)[1]) {

    if (url_combs$WPar[i] == "Wind_speed_gust") {
      IDSensor_link <- Metadata %>%
        dplyr::filter(.data$Measure %in% "Wind_speed") %>%
        dplyr::pull(.data$IDSensor)
    } else {
      IDSensor_link <- Metadata %>%
        dplyr::filter(.data$Measure %in% url_combs$WPar[i]) %>%
        dplyr::pull(.data$IDSensor)
    }

    url_combs$str_sensor[i] <- paste0("AND idsensore in(",paste0(sapply(X = IDSensor_link, function(x) paste0("'",x,"'")),collapse = ","),")")

    dates_blocks <- vector(mode = "list", length = length(dates_seq_begin))
    for (b in 1:length(dates_seq_begin)) {
      seq_temp <- seq(lubridate::as_datetime(dates_seq_begin[b]),
                      lubridate::as_datetime(dates_seq_end[b]),
                      length.out = n_blocks + 1)

      seq_begin_temp <- seq_temp[-length(seq_temp)]
      seq_begin_temp <- lubridate::round_date(seq_begin_temp, unit = "hour")
      seq_begin_temp <- stringr::str_replace(string = seq_begin_temp, pattern = " ", replacement = "T")

      seq_end_temp <- c(seq_temp[-c(1,length(seq_temp))] - lubridate::hours(1),seq_temp[length(seq_temp)])
      seq_end_temp <- lubridate::round_date(seq_end_temp, unit = "hour")
      seq_end_temp <- stringr::str_replace(string = seq_end_temp, pattern = " ", replacement = "T")
      dates_blocks[[b]] <- data.frame(seq_begin_temp,seq_end_temp)
    }

    URL_blocks[[i]] <- dplyr::bind_cols(dplyr::bind_rows(dates_blocks),url_combs[i,]) %>%
      dplyr::mutate(link = paste0(.data$link1,"?$where=data between '", seq_begin_temp, "' and '", seq_end_temp,"'", .data$str_sensor)) %>%
      dplyr::select(.data$link)

  }
  URL_blocks <- dplyr::bind_rows(URL_blocks)


  ### Preparing parallel computation: explicitly open multisession/multicore workers by switching plan
  if (parallel == TRUE) {
    # Explicitly open multisession/multicore workers by switching plan
    future::plan(future::multisession, workers = 12)
    if (is.null(parworkers)) {
      parworkers <- future::availableCores()/2
    }
    eval(parse(text = paste0("future::plan(future::",parfuturetype,", workers = ",parworkers,")")))
    if (verbose == TRUE) {
      message("Start parallel computing: number of parallel workers = ", future::nbrOfWorkers())
    }
  }

  ##### Download using Socrata API
  # Attention: this is a correction for issues with wind speed: wind speed and wind speed gust have the same sensor ID
  # Solution: split the download (gust and non-gust) and add the column "operatore" (as in the previous versions)
  # Think about extending this part to all the weather parameters separately to speed up the download
  if (sum(url_combs$WPar %in% c("Wind_speed_gust")) > 0) {
    idx1 <- which(grepl(pattern = url_combs[url_combs$WPar %in% c("Wind_speed_gust"),"link1"],x = URL_blocks$link))
    ### Modified on 2026-06-15: replace socratadata::soc_read with the internal JSON/SODA reader.
    Meteo1 <- dplyr::bind_rows(
      future.apply::future_apply(X = as.matrix(URL_blocks[idx1,]), MARGIN = 1, FUN = function(x) {
        ARPAL_read_socrata_json(url = as.character(x))
      })
    )
    Meteo1$operatore <- "Max"
    idx2 <- which(!grepl(pattern = url_combs[url_combs$WPar %in% c("Wind_speed_gust"),"link1"],x = URL_blocks$link))
    ### Modified on 2026-06-15: replace socratadata::soc_read with the internal JSON/SODA reader.
    Meteo2 <- dplyr::bind_rows(
      future.apply::future_apply(X = as.matrix(URL_blocks[idx2,]), MARGIN = 1, FUN = function(x) {
        ARPAL_read_socrata_json(url = as.character(x))
      })
    )
    Meteo2$operatore <- "Average"
    Meteo <- dplyr::bind_rows(Meteo1, Meteo2)
  } else {
    ### Modified on 2026-06-15: replace socratadata::soc_read with the internal JSON/SODA reader.
    Meteo <- dplyr::bind_rows(
      future.apply::future_apply(X = as.matrix(URL_blocks), MARGIN = 1, FUN = function(x) {
        ARPAL_read_socrata_json(url = as.character(x))
      })
    )
    Meteo$operatore <- "Average"
  }

  return(Meteo)
}

#' @keywords internal
#' @noRd

AQ_metadata_reshape <-
  function() {

    '%notin%' <- Negate('%in%')

    ##### Check online availability for AQ metadata
    temp <- tempfile()
    res <- suppressWarnings(try(curl::curl_fetch_disk("https://www.dati.lombardia.it/resource/ib47-atvt.csv", temp), silent = TRUE))
    if(res$status_code != 200) {
      message(paste0("The internet resource for air quality stations metadata is not available at the moment. Status code: ",res$status_code,".\nPlease, try later. If the problem persists, please contact the package maintainer."))
      return(invisible(NULL))
    } else {
      ### Modified on 2026-06-15: replace socratadata::soc_read with the internal JSON/SODA reader.
      Metadata <- ARPAL_read_socrata_json(url = "https://www.dati.lombardia.it/resource/ib47-atvt.json")
    }

    Metadata <- Metadata %>%
      dplyr::rename(IDSensor = .data$idsensore, IDStation = .data$idstazione,
                    Pollutant = .data$nometiposensore, NameStation = .data$nomestazione,
                    Altitude = .data$quota, Province = .data$provincia, City = .data$comune,
                    DateStart = .data$datastart, DateStop = .data$datastop,
                    Latitude = .data$lat, Longitude = .data$lng) %>%
      dplyr::mutate(
                    ### Modified on 2026-06-15: parse Socrata JSON date/datetime strings robustly.
                    DateStart = ARPAL_parse_socrata_date(.data$DateStart),
                    DateStop = ARPAL_parse_socrata_date(.data$DateStop),
                    IDStation = as.numeric(.data$IDStation),
                    IDSensor = as.numeric(.data$IDSensor)) %>%
      dplyr::select(.data$IDSensor, .data$IDStation, .data$Pollutant, .data$NameStation, .data$Altitude,
                    .data$Province, .data$City, .data$DateStart, .data$DateStop, .data$Latitude, .data$Longitude) %>%
      dplyr::mutate(Pollutant = dplyr::recode(.data$Pollutant,
                                              "Ammoniaca" = "Ammonia",
                                              "Arsenico" = "Arsenic",
                                              "Benzene" = "Benzene",
                                              "Benzo(a)pirene" = "Benzo_a_pyrene",
                                              "Biossido di Azoto" = "NO2",
                                              "Monossido di Azoto" = "NO",
                                              "Ossidi di Azoto" = "NOx",
                                              "Biossido di Zolfo" = "Sulfur_dioxide",
                                              "BlackCarbon" = "BlackCarbon",
                                              "Monossido di Carbonio" = "CO",
                                              "Nikel" = "Nikel",
                                              "Ozono" = "Ozone",
                                              "Cadmio" = "Cadmium",
                                              "PM10 (SM2005)" = "PM10",
                                              "PM10" = "PM10",
                                              "Particelle sospese PM2.5" = "PM2.5",
                                              "Particolato Totale Sospeso" = "PM_tot",
                                              "Piombo" = "Lead"))

    ### Name stations
    # dplyr::across(c(.data$NameStation,.data$City), ~ stringi::stri_trans_general(str = .x, id="Latin-ASCII"))
    Metadata <- Metadata %>%
      dplyr::mutate(dplyr::across(c(.data$NameStation,.data$City), toupper),
                    dplyr::across(c(.data$NameStation,.data$City), ~ gsub("\\-", " ", .x)),
                    dplyr::across(c(.data$NameStation,.data$City), ~ stringr::str_replace_all(.x, c("S\\."="San ",
                                                                                                    "V\\."="Via ",
                                                                                                    "V\\.LE" = "Viale",
                                                                                                    " D\\`" = " D\\' ",
                                                                                                    " D\\` " = " D\\'",
                                                                                                    "D\\`" = " D\\'",
                                                                                                    "D\\'" = " D\\' ",
                                                                                                    "P\\.ZZA" = "Piazza",
                                                                                                    "C\\.SO" = "Corso",
                                                                                                    "LOC\\." = "Localita"))),
                    dplyr::across(c(.data$NameStation,.data$City), tm::removePunctuation),
                    dplyr::across(c(.data$NameStation,.data$City), tm::removeNumbers),
                    dplyr::across(c(.data$NameStation,.data$City), tm::stripWhitespace),
                    dplyr::across(c(.data$NameStation,.data$City), stringr::str_to_title),
                    dplyr::across(c(.data$NameStation,.data$City), ~ stringr::str_replace_all(.x, c(" D " = " D\\'",
                                                                                                    " Xi " = " XI ",
                                                                                                    " Xxv " = " XXV ",
                                                                                                    " Via " = " - Via ",
                                                                                                    " Corso " = " - Corso ",
                                                                                                    " Localita " = rlang::as_utf8_character(" - Localit\u00e0 "),
                                                                                                    " Piazza " = " - Piazza ",
                                                                                                    " Via Le " = " Viale ",
                                                                                                    "F Turati" = "Turati",
                                                                                                    "Via Serafino Delluomo" = "Via Serafino Dell'Uomo",
                                                                                                    "Via Dellartigianato" = "Via Dell'Artigianato",
                                                                                                    "Montanaso Lombardo Sp" = "Montanaso Lombardo SP202",
                                                                                                    "Sp Casa Dellalpino" = " - SP27 Casa Dell'Alpino",
                                                                                                    "Spino D'Adda Sp" = "Spino D'Adda SP1",
                                                                                                    "Villasanta - Via A Volta" = "Villasanta - Via Volta",
                                                                                                    "Vimercate - Via Dellospedale" = "Vimercate - Via Dell'Ospedale",
                                                                                                    "Ss Sempione" = "SS Sempione"))))


    ### Add extra information from ARPA offices (uploaded on Paolo Maranzano's GitHub page)
    # ARPA_zone = ARPA Lombardia zoning of the region: https://www.arpalombardia.it/Pages/Aria/Rete-di-rilevamento/Zonizzazione.aspx
    # ARPA_stat_type = stations type: https://www.arpalombardia.it/Pages/Aria/Rete-di-rilevamento/Criteri-di-rilevamento/Tipologia-delle-stazioni.aspx?firstlevel=Ieri

    ##### Read further AQ metadata bundled with the package
    ### Modified on 2026-06-26: use the package-bundled AQ stations metadata
    ### stored in inst/extdata, instead of checking and downloading the same
    ### file from the ARPALData GitHub repository at runtime.
    Metadata_ARPA_file <- ARPAL_extdata_path("AQ_stations_ARPA_Lombardia.csv")
    Metadata_ARPA <- readr::read_csv(Metadata_ARPA_file, show_col_types = FALSE)
    Metadata_ARPA <- Metadata_ARPA %>%
      dplyr::select(.data$IDStation,.data$ARPA_zone,.data$ARPA_stat_type) %>%
      dplyr::distinct()
    Metadata <- dplyr::left_join(Metadata,Metadata_ARPA,by = c("IDStation"))
    Metadata <- Metadata %>%
      filter(.data$IDStation %notin% c(518,602,603,612,694,698,700))
    # Galliate 518 (NO) --> Fuori regione, chiusa nel 2017
    # Melegnano 602 --> Chiusa dal 2017 --> Chiusa dal 2017
    # Filago via Fermi Marne (612) --> Chiusa dal 2017
    # Castiraga 603 --> Chiusa dal 2017
    # Salionze 694 (VR) --> Fuori regione
    # Ceneselli 698 (RO) --> Fuori regione
    # Melara 700 (RO) --> Fuori regione, chiusa dal 2018


    structure(list(Metadata = Metadata))
    attr(Metadata, "class") <- c("ARPALdf","tbl_df","tbl","data.frame")
    return(Metadata)
  }

AQ_municipal_metadata_reshape <-
  function() {

    ##### Check online availability for AQ municipal metadata
    temp <- tempfile()
    res <- suppressWarnings(try(curl::curl_fetch_disk("https://www.dati.lombardia.it/resource/5rep-i3mj.csv", temp), silent = TRUE))
    if(res$status_code != 200) {
      message(paste0("The internet resource for municipal air quality stations metadata is not available at the moment. Status code: ",res$status_code,".\nPlease, try later. If the problem persists, please contact the package maintainer."))
      return(invisible(NULL))
    } else {
      ### Modified on 2026-06-15: replace RSocrata::read.socrata with the internal JSON/SODA reader.
      Metadata <- ARPAL_read_socrata_json(url = "https://www.dati.lombardia.it/resource/5rep-i3mj.json")
    }

    # if(res$status_code != 200) {
    #   message(paste0("The internet resource for air quality of municipalities metadata is not available at the moment, try later.\nIf the problem persists, please contact the package maintainer."))
    #   return(invisible(NULL))
    # } else {
    #   ### Modified on 2026-06-15: registry download now uses ARPAL_read_socrata_json().
    # }

    Metadata <- Metadata %>%
      dplyr::rename(IDSensor = .data$idsensore, IDStation = .data$idstazione,
                    Pollutant = .data$nometiposensore,
                    Province = .data$provincia, NameStation = .data$comune,
                    DateStart = .data$datastart, DateStop = .data$datastop) %>%
      dplyr::mutate(
        ### Modified on 2026-06-15: parse Socrata JSON date/datetime strings robustly.
        DateStart = ARPAL_parse_socrata_date(.data$DateStart),
        DateStop = ARPAL_parse_socrata_date(.data$DateStop)
      ) %>%
      dplyr::select(.data$IDSensor, .data$IDStation, .data$Pollutant,
                    .data$Province, .data$NameStation, .data$DateStart, .data$DateStop) %>%
      dplyr::mutate(Pollutant = dplyr::recode(.data$Pollutant,
                                              "Ammoniaca" = "Ammonia",
                                              "Arsenico" = "Arsenic",
                                              "Benzene" = "Benzene",
                                              "Benzo(a)pirene" = "Benzo_a_pyrene",
                                              "Biossido di Azoto" = "NO2",
                                              "Monossido di Azoto" = "NO",
                                              "Ossidi di Azoto" = "NOx",
                                              "Biossido di Zolfo" = "Sulfur_dioxide",
                                              "BlackCarbon" = "BlackCarbon",
                                              "Monossido di Carbonio" = "CO",
                                              "Nikel" = "Nikel",
                                              "Ozono" = "Ozone",
                                              "Cadmio" = "Cadmium",
                                              "PM10 (SM2005)" = "PM10",
                                              "PM10" = "PM10",
                                              "Particelle sospese PM2.5" = "PM2.5",
                                              "Particolato Totale Sospeso" = "PM_tot",
                                              "Piombo" = "Lead"))

    ### Name stations (municipalities)
    Metadata <- Metadata %>%
      dplyr::mutate(dplyr::across(c(.data$NameStation), toupper),
                    dplyr::across(c(.data$NameStation), ~ gsub("\\-", " ", .x)),
                    dplyr::across(c(.data$NameStation), ~ stringr::str_replace_all(.x, c("S\\."="San ","s\\."="San ",
                                                                                         "V\\."="Via ","v\\."="Via ",
                                                                                         " D\\`" = " D\\' ", " D\\` " = " D\\'",
                                                                                         "D\\`" = " D\\'", "D\\'" = " D\\' "))),
                    dplyr::across(c(.data$NameStation), tm::removePunctuation),
                    dplyr::across(c(.data$NameStation), tm::removeNumbers),
                    dplyr::across(c(.data$NameStation), tm::stripWhitespace)) %>%
      dplyr::mutate(NameStation = dplyr::recode(.data$NameStation,
                                                "CASASCO DINTELVI" = "CASASCO INTELVI",
                                                "CERANO DINTELVI" = "CERANO INTELVI",
                                                "SAN GIORGIO BIGARELLO" = "BIGARELLO",
                                                "PUEGNAGO DEL GARDA" = "PUEGNAGO SUL GARDA",
                                                "FELONICA" = "SERMIDE E FELONICA",
                                                "GERRE DE CAPRIOLI" = "GERRE DECAPRIOLI")) %>%
      dplyr::mutate(dplyr::across(c(.data$NameStation), stringr::str_to_title),
                    dplyr::across(c(.data$NameStation), ~ stringr::str_replace_all(.x,
                                                                                   c(" D " = " D\\'",
                                                                                     "Sermide E Felonica" = "Sermide e Felonica"))))

    structure(list(Metadata = Metadata))
    attr(Metadata, "class") <- c("ARPALdf","tbl_df","tbl","data.frame")
    return(Metadata)
  }

#' @keywords internal
#' @noRd

Hampel_flt <- function(var) {
  I_low <- max(median(var,na.rm = T) - 3*mad(var,na.rm = T), min(var,na.rm=T))
  I_upp <- min(median(var,na.rm = T) + 3*mad(var,na.rm = T), max(var,na.rm=T))
  outlier_ind_low <- which(var < I_low)
  outlier_ind_upp <- which(var > I_upp)

  return(list(I_low = I_low,I_upp = I_upp,
              outlier_ind_upp = outlier_ind_upp,
              outlier_ind_low = outlier_ind_low))
}

# The Hampel Filter detects and removes the outliers of the input signal by using the Hampel
# identifier. The Hampel identifier is a variation of the three-sigma rule of statistics,
# which is robust against outliers. For each sample of the input signal, the block computes the
# median of a window composed of the current sample and Length−12 adjacent samples on each side of
# the current sample. Len is the window length you specify through the Window length parameter.
# The block also estimates the standard deviation of each sample about its window median by using
# the median absolute deviation. If a sample differs from the median by more than the threshold
# multiplied by the standard deviation, the filter replaces the sample with the median.

#' @keywords internal
#' @noRd

quantilep <- function(x, perc_str) {
  ### Modified on 2026-06-15: return an empty named list when no quantile
  ### functions are requested. This preserves the previous behavior inside
  ### dynamic dplyr summaries and avoids assigning one name to a zero-length
  ### list when Fns_vec does not contain strings such as "q25" or "q75".
  if (length(perc_str) == 0) {
    return(stats::setNames(list(), character(0)))
  }

  p <- suppressWarnings(as.numeric(sub(".*q", "", perc_str)) / 100)
  p <- p[is.finite(p)]

  if (length(p) == 0) {
    return(stats::setNames(list(), character(0)))
  }

  ### Modified on 2026-06-15: replace purrr::map_chr(), purrr::map(),
  ### purrr::partial() and purrr::set_names() with base R equivalents, so
  ### the package does not need to import purrr only for this small helper.
  p_names <- paste0("q", p * 100)
  p_funs <- lapply(p, function(pp) {
    force(pp)
    function(x, ...) {
      stats::quantile(x, probs = pp, na.rm = TRUE, ...)
    }
  })
  names(p_funs) <- p_names

  return(p_funs)
}

#' Check if a given object belongs to class 'ARPALdf_W'
#'
#' @description 'is_ARPALdf_W' checks if the input object belongs to the class 'ARPALdf_W'
#'
#' @param Data Object to check if the class of a dataframe is 'ARPALdf_W', i.e. ARPAL
#' dataframe for weather data.
#'
#' @return The function returns 'True' if the object is of class 'ARPALdf_W' and it returns 'False' if the
#'     object isn't of class 'ARPALdf_W'
#'
#' @examples
#' d <- get_ARPA_Lombardia_W_registry()
#' is_ARPALdf_W(d)
#'
#' @export
is_ARPALdf_W <- function(Data) {
  is.element("ARPALdf_W", attr(Data,"class"))
}

#' @keywords internal
#' @noRd

mom_skew <- function(x, na.rm = FALSE) {
  if (is.matrix(x)) {
    apply(x, 2, FUN = function(x) {
      if (na.rm) {
        x <- x[!is.na(x)]
      }
      (sum((x - mean(x))^3)/n)/(sum((x - mean(x))^2)/n)^(3/2)
      })
  } else if (is.vector(x)) {
    if (na.rm) {
      x <- x[!is.na(x)]
    }
    n <- length(x)
    (sum((x - mean(x))^3)/n)/(sum((x - mean(x))^2)/n)^(3/2)
  } else if (is.data.frame(x)) {
    sapply(X = x, FUN = function(x) {
      if (na.rm) {
        x <- x[!is.na(x)]
      }
      (sum((x - mean(x))^3)/n)/(sum((x - mean(x))^2)/n)^(3/2)
    })
  }
}

#' @keywords internal
#' @noRd

url_AQ_year_dataset <-
  function(Stat_type, Year) {
    url <- switch(Stat_type,
                  ##### Air quality from ground network
                  ### Modified on 2026-06-15: use the current-year endpoint for
                  ### Dati sensori aria and the consolidated historical endpoint for
                  ### Dati sensori aria dal 2018.
                  AQ = dplyr::case_when(Year >= 2026 ~ "https://www.dati.lombardia.it/resource/nicp-bhqi.json",
                                        Year %in% 2018:2025 ~ "https://www.dati.lombardia.it/resource/g2hp-ar79.json",
                                        Year %in% 2010:2017 ~ "https://www.dati.lombardia.it/resource/nr8w-tj77.json",
                                        Year %in% 2000:2009 ~ "https://www.dati.lombardia.it/resource/cthp-zqrr.json",
                                        Year %in% 1968:1999 ~ "https://www.dati.lombardia.it/resource/evzn-32bs.json"),

                  ##### Municipal data
                  AQ_municipal = dplyr::case_when(Year >= 2025 ~ "https://www.dati.lombardia.it/resource/ysm5-jwrn.json",
                                                  Year == 2024 ~ "https://www.dati.lombardia.it/resource/qyg8-q6gd.json",
                                                  Year == 2023 ~ "https://www.dati.lombardia.it/resource/q25y-843e.json",
                                                  Year == 2022 ~ "https://www.dati.lombardia.it/resource/fqaz-7ste.json",
                                                  Year == 2021 ~ "https://www.dati.lombardia.it/resource/7iq4-hq9t.json",
                                                  Year %in% 2017:2020 ~ "https://www.dati.lombardia.it/resource/aip6-75fk.json",
                                                  ### Modified on 2026-06-15: replace the older zipped municipal archive with the
                                                  ### current Socrata JSON endpoint for Dati Stime Comunali 2011-2016.
                                                  Year %in% 2011:2016 ~ "https://www.dati.lombardia.it/resource/miv5-s38a.json"),

                  ##### Web pages of the municipal data
                  ### Modified on 2026-06-15: align municipal-data check URLs with
                  ### the current Open Data Lombardia ARPA catalogue/story page.
                  ### In particular, 2024 and 2023 have year-specific pages, 2021
                  ### uses dataset id 7iq4-hq9t, and 2017--2020 are exposed as a
                  ### single consolidated dataset.
                  AQ_municipal_check = dplyr::case_when(Year >= 2025 ~ "https://www.dati.lombardia.it/Ambiente/Dati-stime-comunali/ysm5-jwrn",
                                                        Year == 2024 ~ "https://www.dati.lombardia.it/Ambiente/Dati-Stime-Comunali-2024/qyg8-q6gd",
                                                        Year == 2023 ~ "https://www.dati.lombardia.it/Ambiente/Dati-Stime-comunali-2023/q25y-843e",
                                                        Year == 2022 ~ "https://www.dati.lombardia.it/Ambiente/Dati-stime-comunali-2022/fqaz-7ste",
                                                        Year == 2021 ~ "https://www.dati.lombardia.it/Ambiente/Dati-Stime-Comunali-2021/7iq4-hq9t",
                                                        Year %in% 2017:2020 ~ "https://www.dati.lombardia.it/Ambiente/Dati-Stime-Comunali-2017-2020/aip6-75fk",
                                                        ### Modified on 2026-06-15: update the 2011-2016 municipal check page to
                                                        ### the current Open Data Lombardia dataset id miv5-s38a.
                                                        Year %in% 2011:2016 ~ "https://www.dati.lombardia.it/Ambiente/Dati-Stime-Comunali-2011-2016/miv5-s38a")
    )

    return(url)
  }

#' @keywords internal
#' @noRd

url_W_year_par <-
  function(WPar,Year) {

    ### Modified on 2026-06-15: validate scalar inputs before switch().
    ### This prevents failures when empty/NA metadata values reach the URL map.
    WPar <- as.character(WPar)
    Year <- suppressWarnings(as.integer(Year))
    if (length(WPar) != 1 || length(Year) != 1 || is.na(WPar) || is.na(Year)) {
      return(NA_character_)
    }

    ### Modified on 2026-06-15: normalize weather-parameter aliases before
    ### selecting the Socrata endpoint. Only ASCII labels are used in R code
    ### to comply with CRAN portability checks.
    WPar <- dplyr::recode(WPar,
                          "Wind_direction_gust" = "Wind_speed_gust",
                          "Wind_direction_gust_check" = "Wind_speed_gust_check",
                          "Raffica Vento" = "Wind_speed_gust",
                          "Raffica Vento_check" = "Wind_speed_gust_check",
                          "Velocita Vento Raffica" = "Wind_speed_gust",
                          "Velocita Vento Raffica_check" = "Wind_speed_gust_check",
                          .default = WPar)

    url <- switch(WPar,
                  ##### Temperature
                  Temperature = dplyr::case_when(Year >= 2021 ~ "https://www.dati.lombardia.it/resource/w9wd-u6jh.json",
                                                 Year %in% 2011:2020 ~ "https://www.dati.lombardia.it/resource/d4kj-kbpj.json",
                                                 Year <= 2010 ~ "https://www.dati.lombardia.it/resource/6eu4-4tja.json"),
                  Temperature_check = dplyr::case_when(Year >= 2021 ~ "https://www.dati.lombardia.it/Ambiente/Temperatura-dal-2021/w9wd-u6jh/about_data",
                                                       Year %in% 2011:2020 ~ "https://www.dati.lombardia.it/Ambiente/Temperatura-dal-2011-al-2020/d4kj-kbpj/about_data",
                                                       Year <= 2010 ~ "https://www.dati.lombardia.it/Ambiente/Temperatura-fino-al-2010/6eu4-4tja/about_data"),
                  ##### Wind speed
                  Wind_speed = dplyr::case_when(Year >= 2021 ~ "https://www.dati.lombardia.it/resource/hu5q-68e3.json",
                                                Year %in% 2011:2020 ~ "https://www.dati.lombardia.it/resource/x9gp-z9xx.json",
                                                Year <= 2010 ~ "https://www.dati.lombardia.it/resource/ed5f-uim8.json"),
                  Wind_speed_check = dplyr::case_when(Year >= 2021 ~ "https://www.dati.lombardia.it/Ambiente/Velocit-del-vento-dal-2021/hu5q-68e3/about_data",
                                                      Year %in% 2011:2020 ~ "https://www.dati.lombardia.it/Ambiente/Velocit-del-vento-dal-2011-al-2020/x9gp-z9xx/about_data",
                                                      Year <= 2010 ~ "https://www.dati.lombardia.it/Ambiente/Velocit-del-vento-fino-al-2010/ed5f-uim8/about_data"),
                  ##### Wind direction
                  Wind_direction = dplyr::case_when(Year >= 2021 ~ "https://www.dati.lombardia.it/resource/purm-rsjf.json",
                                                    Year %in% 2011:2020 ~ "https://www.dati.lombardia.it/resource/ypty-e75m.json",
                                                    Year <= 2010 ~ "https://www.dati.lombardia.it/resource/cptm-adt5.json"),
                  Wind_direction_check = dplyr::case_when(Year >= 2021 ~ "https://www.dati.lombardia.it/Ambiente/Direzione-del-vento-dal-2021/purm-rsjf/about_data",
                                                          Year %in% 2011:2020 ~ "https://www.dati.lombardia.it/Ambiente/Direzione-del-vento-dal-2011-al-2020/ypty-e75m/about_data",
                                                          Year <= 2010 ~ "https://www.dati.lombardia.it/Ambiente/Direzione-del-vento-fino-al-2010/cptm-adt5/about_data"),
                  ##### Wind gust
                  Wind_speed_gust = dplyr::case_when(Year >= 2021 ~ "https://www.dati.lombardia.it/resource/gi9u-2aez.json",
                                                     Year <= 2020 ~ "https://www.dati.lombardia.it/resource/ppwf-2tzh.json"),
                  Wind_speed_gust_check = dplyr::case_when(Year >= 2021 ~ "https://www.dati.lombardia.it/Ambiente/Raffica-del-vento-dal-2021/gi9u-2aez/about_data",
                                                           Year <= 2020 ~ "https://www.dati.lombardia.it/Ambiente/Raffica-del-vento-dal-2012-al-2020/ppwf-2tzh/about_data"),
                  ##### Relative humidity
                  Relative_humidity = dplyr::case_when(Year >= 2021 ~ "https://www.dati.lombardia.it/resource/823w-fh4c.json",
                                                       Year %in% 2011:2020 ~ "https://www.dati.lombardia.it/resource/xpun-8722.json",
                                                       Year <= 2010 ~ "https://www.dati.lombardia.it/resource/nure-mhrg.json"),
                  Relative_humidity_check = dplyr::case_when(Year >= 2021 ~ "https://www.dati.lombardia.it/Ambiente/Umidit-relativa-dal-2021/823w-fh4c/about_data",
                                                             Year %in% 2011:2020 ~ "https://www.dati.lombardia.it/Ambiente/Umidit-relativa-dal-2011-al-2020/xpun-8722/about_data",
                                                             Year <= 2010 ~ "https://www.dati.lombardia.it/Ambiente/Umidit-relativa-fino-al-2010/nure-mhrg/about_data"),
                  ##### Rainfall
                  Rainfall = dplyr::case_when(Year >= 2021 ~ "https://www.dati.lombardia.it/resource/pstb-pga6.json",
                                              Year %in% 2011:2020 ~ "https://www.dati.lombardia.it/resource/2kar-pnuk.json",
                                              Year <= 2010 ~ "https://www.dati.lombardia.it/resource/e7r2-7m84.json"),
                  Rainfall_check = dplyr::case_when(Year >= 2021 ~ "https://www.dati.lombardia.it/Ambiente/Precipitazioni-dal-2021/pstb-pga6/about_data",
                                                    Year %in% 2011:2020 ~ "https://www.dati.lombardia.it/Ambiente/Precipitazioni-dal-2011-al-2020/2kar-pnuk/about_data",
                                                    Year <= 2010 ~ "https://www.dati.lombardia.it/Ambiente/Precipitazioni-fino-al-2010/e7r2-7m84/about_data"),
                  ##### Global solar radiation
                  Global_radiation = dplyr::case_when(Year >= 2021 ~ "https://www.dati.lombardia.it/resource/cxym-eps2.json",
                                                       Year <= 2020 ~ "https://www.dati.lombardia.it/resource/63ns-e4tv.json"),
                  Global_radiation_check = dplyr::case_when(Year >= 2021 ~ "https://www.dati.lombardia.it/Ambiente/Radiazione-Globale-dal-2021/cxym-eps2/about_data",
                                                            Year <= 2020 ~ "https://www.dati.lombardia.it/Ambiente/Radiazione-Globale-fino-al-2020/63ns-e4tv/about_data"),
                  ##### Water height
                  Water_height = dplyr::case_when(Year >= 2021 ~ "https://www.dati.lombardia.it/resource/3e8b-w7ay.json",
                                                  Year %in% 2011:2020 ~ "https://www.dati.lombardia.it/resource/gsyu-uxt3.json",
                                                  Year <= 2010 ~ "https://www.dati.lombardia.it/resource/xubc-puka.json"),
                  Water_height_check = dplyr::case_when(Year >= 2021 ~ "https://www.dati.lombardia.it/Ambiente/Livello-idrometrico-dal-2021/3e8b-w7ay/about_data",
                                                        Year %in% 2011:2020 ~ "https://www.dati.lombardia.it/Ambiente/Livello-idrometrico-dal-2011-al-2020/gsyu-uxt3/about_data",
                                                        Year <= 2010 ~ "https://www.dati.lombardia.it/Ambiente/Livello-idrometrico-fino-al-2010/xubc-puka/about_data"),
                  ##### Snow height
                  Snow_height = dplyr::case_when(Year >= 2021 ~ "https://www.dati.lombardia.it/resource/uqbu-tt6m.json",
                                                 Year %in% 2011:2020 ~ "https://www.dati.lombardia.it/resource/id9e-7wf3.json",
                                                 Year <= 2010 ~ "https://www.dati.lombardia.it/resource/5kig-ayub.json"),
                  Snow_height_check = dplyr::case_when(Year >= 2021 ~ "https://www.dati.lombardia.it/Ambiente/Altezza-neve-dal-2021/uqbu-tt6m/about_data",
                                                       Year %in% 2011:2020 ~ "https://www.dati.lombardia.it/Ambiente/Altezza-neve-dal-2011-al-2020/id9e-7wf3/about_data",
                                                       Year <= 2010 ~ "https://www.dati.lombardia.it/it/Ambiente/Altezza-neve-fino-al-2010/5kig-ayub/about_data"),
                  NA_character_)

    ### Modified on 2026-06-15: guarantee a scalar character return value.
    if (length(url) != 1 || is.na(url)) {
      return(NA_character_)
    }

    return(url)
  }

#' @keywords internal
#' @noRd

W_metadata_reshape <-
  function() {

    '%notin%' <- Negate('%in%')

    ##### Check online availability for weather metadata
    temp <- tempfile()
    res <- suppressWarnings(try(curl::curl_fetch_disk("https://www.dati.lombardia.it/resource/nf78-nj6b.csv", temp), silent = TRUE))
    if(res$status_code != 200) {
      message(paste0("The internet resource for weather stations metadata is not available at the moment. Status code: ",res$status_code,".\nPlease, try later. If the problem persists, please contact the package maintainer."))
      return(invisible(NULL))
    } else {
      ### Modified on 2026-06-15: replace socratadata::soc_read with the internal JSON/SODA reader.
      Metadata <- ARPAL_read_socrata_json(url = "https://www.dati.lombardia.it/resource/nf78-nj6b.json")
    }

    Metadata <- Metadata %>%
      dplyr::rename(IDSensor = .data$idsensore, IDStation = .data$idstazione, NameStation = .data$nomestazione,
                    Measure = .data$tipologia, Unit_meas = .data$unit_dimisura,
                    Altitude = .data$quota, Province = .data$provincia,
                    DateStart = .data$datastart, DateStop = .data$datastop,
                    Latitude = .data$lat, Longitude = .data$lng) %>%
      ### Modified on 2026-06-15: parse Socrata JSON datetime strings robustly.
      ### The JSON endpoint may return values such as "YYYY-MM-DDT00:00:00.000",
      ### which are not handled reliably by lubridate::ymd().
      dplyr::mutate(Altitude = as.numeric(.data$Altitude),
                    DateStart = as.Date(dplyr::coalesce(lubridate::ymd_hms(.data$DateStart, quiet = TRUE),
                                                        as.POSIXct(lubridate::ymd(.data$DateStart, quiet = TRUE)))),
                    DateStop = as.Date(dplyr::coalesce(lubridate::ymd_hms(.data$DateStop, quiet = TRUE),
                                                       as.POSIXct(lubridate::ymd(.data$DateStop, quiet = TRUE))))) %>%
      dplyr::select(.data$IDSensor, .data$IDStation, .data$Measure, .data$Unit_meas, .data$NameStation, .data$Altitude,
                    .data$Province, .data$DateStart, .data$DateStop, .data$Latitude, .data$Longitude) %>%
      dplyr::mutate(Measure = dplyr::recode(.data$Measure,
                                            "Altezza Neve" = "Snow_height",
                                            "Direzione Vento" = "Wind_direction",
                                            "Livello Idrometrico" = "Water_height",
                                            "Precipitazione" = "Rainfall",
                                            "Radiazione Globale" = "Global_radiation",
                                            "Temperatura" = "Temperature"),
                    Measure = ifelse(.data$Measure == rlang::as_utf8_character("Umidit\u00e0 Relativa"),"Relative_humidity",.data$Measure),
                    Measure = ifelse(.data$Measure == rlang::as_utf8_character("Velocit\u00e0 Vento"),"Wind_speed",.data$Measure))

    ### Name stations
    # dplyr::across(c(.data$NameStation), ~ stringi::stri_trans_general(str = .x, id="Latin-ASCII")),
    Metadata <- Metadata %>%
      dplyr::mutate(dplyr::across(c(.data$NameStation), toupper),
                    dplyr::across(c(.data$NameStation), ~ gsub("\\-", " ", .x)),
                    dplyr::across(c(.data$NameStation), ~ stringr::str_replace_all(.x, c("S\\."="San ",
                                                                                         "V\\."="Via ",
                                                                                         "V\\.LE" = "Viale",
                                                                                         " D\\`" = " D\\' ",
                                                                                         " D\\` " = " D\\'",
                                                                                         "D\\`" = " D\\'",
                                                                                         "D\\'" = " D\\' ",
                                                                                         "P\\.ZZA" = "Piazza",
                                                                                         "C\\.SO" = "Corso",
                                                                                         "LOC\\." = "Localita"))),
                    dplyr::across(c(.data$NameStation), tm::removePunctuation),
                    dplyr::across(c(.data$NameStation), tm::removeNumbers),
                    dplyr::across(c(.data$NameStation), tm::stripWhitespace),
                    dplyr::across(c(.data$NameStation), stringr::str_to_title),
                    dplyr::across(c(.data$NameStation), ~ stringr::str_replace_all(.x, c(" D " = " D\\'",
                                                                                         " Xi " = " XI ",
                                                                                         " Xxv " = " XXV ",
                                                                                         " Xxiv " = " XXIV ",
                                                                                         " Via " = " - Via ",
                                                                                         " Viale " = " - Viale ",
                                                                                         " Corso " = " - Corso ",
                                                                                         " Localita " = rlang::as_utf8_character(" - Localit\u00e0 "),
                                                                                         " Piazza " = " - Piazza ",
                                                                                         " Via Le " = " Viale ",
                                                                                         "Smr" = "SMR",
                                                                                         "Bagolino Sp" = "Bagolino SP669",
                                                                                         "Borgoforte Ss" = "Borgoforte SS12",
                                                                                         "Capriolo Sp" = "Capriolo SP12",
                                                                                         "G Matteotti" = "Matteotti",
                                                                                         "Via Novembre" = "Via IV Novembre",
                                                                                         "Clusone Sp" = "Clusone SP671",
                                                                                         "Cremona Sp" = "Cremona SP10",
                                                                                         "Darfo Boario Terme Ss" = "Darfo Boario Terme SS42",
                                                                                         "Gavardo Sp" = "Gavardo SP116",
                                                                                         "Idro Ss" = "Idro SS237",
                                                                                         "Lainate Sp" = "Lainate SP109",
                                                                                         "Lomello Ss" = "Lomello SS211",
                                                                                         "Orio Litta Ss" = "Orio Litta SS234",
                                                                                         "Pavia Ss" = "Pavia SS35",
                                                                                         "Pralboino Sp" = "Pralboino SP64",
                                                                                         "Rivolta D'Adda Sp" = "Rivolta d'Adda SP4",
                                                                                         "Salerano Sul Lambro Sp" = "Salerano sul Lambro SP115",
                                                                                         "Sermide E Felonica Sp" = "Sermide e Felonica SP91",
                                                                                         "Vigevano Ss" = "Vigevano SS494"))))
    Metadata <- Metadata %>%
      filter(.data$IDStation %notin% c(891))
    # Isola San Antonio 891 (AL) --> Fuori regione

    return(Metadata)
  }

#' Check if a given object belongs to class 'ARPALdf_AQ_mun'
#'
#' @description 'is_ARPALdf_AQ_mun' checks if the input object belongs to the class 'ARPALdf_AQ_mun'
#'
#' @param Data Object to check if the class of a dataframe is 'ARPALdf_AQ_mun', i.e. ARPAL
#' dataframe for air quality data at municipal level (See 'get_ARPA_Lombardia_AQ_municipal_data'.
#' command).
#'
#' @return The function returns 'True' if the object is of class 'ARPALdf_AQ_mun' and it returns
#' 'False' if the object isn't of class 'ARPALdf_AQ_mun'
#'
#' @examples
#' d <- get_ARPA_Lombardia_AQ_registry()
#' is_ARPALdf_AQ_mun(d)
#'
#' @export

is_ARPALdf_AQ_mun <- function(Data) {
  is.element("ARPALdf_AQ_mun", attr(Data,"class"))
}

#' Download metadata (registry) on air quality monitoring stations at municipal level from ARPA Lombardia website
#'
#' @description 'get_ARPA_Lombardia_AQ_municipal_registry' returns the registry (list) of all the air quality sensors
#' owned by ARPA Lombardia for each municipality of Lombardy. The information reported are: ID of each sensor and
#' station, starting date and ending date. The column 'NameStation' identifies the name of each municipality.
#' The column 'IDStation' is an ID code (assigned from ARPA) uniquely identifying each municipality.
#' For more information, visit the section 'Stime comunali sull'aria' at the website:
#' https://www.dati.lombardia.it/stories/s/auv9-c2sj
#'
#' @return A data frame of class 'data.frame' and 'ARPALdf'. The object is fully compatible with Tidyverse.
#'
#' @examples get_ARPA_Lombardia_AQ_municipal_registry()
#'
#' @export

get_ARPA_Lombardia_AQ_municipal_registry <-
  function() {
    Metadata <- AQ_municipal_metadata_reshape()
    structure(list(Metadata = Metadata))
    attr(Metadata, "class") <- c("ARPALdf","ARPALdf_AQ_mun","tbl_df","tbl","data.frame")
    return(Metadata)
  }

#' @keywords internal
#' @noRd

Excess_na_converter <- function(grouped_data,verbose=T) {
  grouped_data2 <- grouped_data %>%
    dplyr::mutate(dplyr::across(tidyselect::contains(c("NO2","NOx","NO","Ozone","CO")), ~
                                  dplyr::case_when(mean(is.na(.x)) >= 0.25 ~ NA_real_,
                                                   mean(is.na(.x)) < 0.25 ~ .x)))

  if (verbose==T) {
    cat("Before aggregation: converting to NA all the obs. belonging to a group with more than 75% missing values \n")
    for (j in 1:dim(grouped_data2)[2]) {
      cat(paste0("Number of obs. converted to NA for ",colnames(grouped_data2)[j],": ",
                 sum(is.na(grouped_data2[,j]) - is.na(grouped_data[,j])),"\n"))
    }
  }

  return(grouped_data2)
}

#' Aggregate any ARPALdf object (with higher temporal frequency) to hourly, daily, weekly, monthly and yearly
#' temporal frequencies.
#'
#' @description Starting from an ARPALdf object with high frequency (e.g., 10mins or hourly), 'Time_aggregate'
#' aggregates the dataset to lower temporal frequencies (e.g., hourly, daily, weekly, monthly and yearly) by station.
#' The output is an ARPALdf object with observations having hourly, daily, weekly, monthly or yearly frequency.
#' The function can be applied only to ARPALdf objects.
#' User can indicate specific variables to aggregate and an aggregation function among
#' mean, median, sum (cumulated), min, max, quantiles, and variability metrics for each variable.
#' It is possible to specify different aggregation functions on the same variable
#' by repeating the name of the variable in 'Var_vec' and specifying the functions in 'Fns_vec'.
#'
#' @param Dataset ARPALdf dataframe to aggregate.
#' @param Frequency Temporal aggregation frequency. It can be "hourly", "daily", "weekly",
#' "monthly" or "yearly.
#' @param Var_vec Vector of variables to aggregate. If NULL (default) all the variables are averaged,
#' expect for 'Temperature' and 'Snow_height' which are summed.
#' @param Fns_vec Vector of aggregation functions to apply to the selected variables. Available functions
#' are 'mean', 'median', 'min', 'max', 'sum', 'qPP' (PP-th percentile), 'sd', 'var', 'vc' (variability coefficient),
#' 'skew' (skewness) and 'kurt' (kurtosis). Attention: for Wind Speed and Wind Speed Gust only mean, min
#' and max are available; for Wind Direction and Wind Direction Gust only mean is available.
#' @param verbose Logical value (TRUE or FALSE). Toggle warnings and messages. If 'verbose=T' (default) the function
#' prints on the screen some messages describing the progress of the tasks. If 'verbose=F' any message about
#' the progression is suppressed.
#'
#' @return A data frame
#'
#' @examples
#' \donttest{
#' ### Modified on 2026-06-15: reduce Time_aggregate examples to January 2020
#' ### to keep CRAN checks lightweight while preserving the same workflow.
#' ## Download hourly observed concentrations during January 2020 for station 501
#' data <- get_ARPA_Lombardia_AQ_data(ID_station=501, Date_begin = "2020-01-01",
#'                                    Date_end = "2020-01-31", Frequency="hourly")
#' ## Aggregate all the data to daily frequency
#' Time_aggregate(Dataset=data,Frequency="daily",Var_vec=NULL,Fns_vec=NULL)
#' ## Aggregate NO2 to weekly maximum concentrations and NOx to weekly minimum concentrations.
#' Time_aggregate(Dataset=data,Frequency="weekly",Var_vec=c("NO2","NOx"),Fns_vec=c("max","min"))
#' }
#'
#' @export

Time_aggregate <- function(Dataset, Frequency, Var_vec = NULL, Fns_vec = NULL, verbose = T) {

  ### Checks
  stopifnot("Data is not of class 'ARPALdf'" = is_ARPALdf(Data = Dataset) == T)

  if (is.null(Var_vec) & is.null(Fns_vec)) {
    vv <- c("Ammonia","Arsenic","Benzene","Benzo_a_pyrene","BlackCarbon","Cadmium",
            "CO","Lead","Nikel","NO","NO2","NOx","Ozone","PM_tot","PM10","PM2.5","Sulfur_dioxide",
            "Rainfall","Temperature","Relative_humidity","Global_radiation","Water_height",
            "Snow_height","Wind_speed","Wind_speed_max","Wind_direction","Wind_direction_max",
            "NO2_mean","NO2_max_day","Ozone_max_8h","Ozone_max_day","PM10_mean","PM2.5_mean")
    vv <- vv[vv %in% names(Dataset)]
    fv <- ifelse(vv == "Rainfall" | vv == "Snow_height", "sum", "mean")
  } else {
    vv <- Var_vec
    fv <- Fns_vec
  }

  data_aggr <- switch(Frequency,
                      yearly = {
                        # Aggregation to yearly data
                        Dataset %>%
                          data.frame() %>%
                          dplyr::mutate(Y = lubridate::year(.data$Date)) %>%
                          dplyr::group_by(.data$Y, .data$NameStation, .data$IDStation) %>%
                          Custom_summarise(vv, fv) %>%
                          dplyr::ungroup() %>%
                          dplyr::mutate(Date = lubridate::make_datetime(year = .data$Y)) %>%
                          dplyr::select(-c(.data$Y)) %>%
                          dplyr::relocate(.data$Date,.data$IDStation) %>%
                          dplyr::mutate(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric),
                                                      ~ ifelse(is.nan(.), NA, .))) %>%
                          dplyr::mutate(dplyr::across(tidyselect::matches(c("Wind_direction","Wind_direction_max")), ~ round(.x,0)))},
                      monthly = {
                        # Aggregation to monthly data
                        Dataset %>%
                          data.frame() %>%
                          dplyr::mutate(Y = lubridate::year(.data$Date),
                                        M = lubridate::month(.data$Date)) %>%
                          dplyr::group_by(.data$Y, .data$M, .data$NameStation, .data$IDStation) %>%
                          Custom_summarise(vv, fv) %>%
                          dplyr::ungroup() %>%
                          dplyr::mutate(Date = lubridate::make_datetime(year = .data$Y,
                                                                        month = .data$M)) %>%
                          dplyr::select(-c(.data$Y,.data$M)) %>%
                          dplyr::relocate(.data$Date,.data$IDStation) %>%
                          dplyr::mutate(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric),
                                                      ~ ifelse(is.nan(.), NA, .))) %>%
                          dplyr::mutate(dplyr::across(tidyselect::matches(c("Wind_direction","Wind_direction_max")), ~ round(.x,0)))},
                      weekly = {
                        # Aggregation to weekly data
                        Dataset %>%
                          data.frame() %>%
                          dplyr::mutate(Y = lubridate::year(.data$Date),
                                        M = lubridate::month(.data$Date),
                                        D = lubridate::day(.data$Date),
                                        W = aweek::date2week(ISOdate(year = .data$Y,
                                                                     month = .data$M,
                                                                     day = .data$D),factor = T)) %>%
                          dplyr::group_by(.data$W, .data$NameStation, .data$IDStation) %>%
                          Custom_summarise(vv, fv) %>%
                          dplyr::ungroup() %>%
                          dplyr::mutate(Date = aweek::week2date(.data$W),
                                        Date = lubridate::make_datetime(year = lubridate::year(.data$Date),
                                                                        month = lubridate::month(.data$Date),
                                                                        day = lubridate::day(.data$Date))) %>%
                          dplyr::select(-c(.data$W)) %>%
                          dplyr::relocate(.data$Date,.data$IDStation) %>%
                          dplyr::mutate(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric),
                                                      ~ ifelse(is.nan(.), NA, .))) %>%
                          dplyr::mutate(dplyr::across(tidyselect::matches(c("Wind_direction","Wind_direction_max")), ~ round(.x,0)))},
                      # Aggregation to daily data
                      daily = {
                        Dataset %>%
                          data.frame() %>%
                          dplyr::mutate(Y = lubridate::year(.data$Date),
                                        M = lubridate::month(.data$Date),
                                        D = lubridate::day(.data$Date)) %>%
                          dplyr::group_by(.data$Y, .data$M, .data$D, .data$NameStation,.data$IDStation) %>%
                          Excess_na_converter(verbose=verbose) %>%
                          Custom_summarise(vv, fv) %>%
                          dplyr::ungroup() %>%
                          dplyr::mutate(Date = lubridate::make_datetime(year = .data$Y,
                                                                        month = .data$M,
                                                                        day = .data$D)) %>%
                          dplyr::select(-c(.data$Y,.data$M,.data$D)) %>%
                          dplyr::relocate(.data$Date,.data$IDStation) %>%
                          dplyr::mutate(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric),
                                                      ~ ifelse(is.nan(.), NA, .))) %>%
                          dplyr::mutate(dplyr::across(tidyselect::matches(c("Wind_direction","Wind_direction_max")), ~ round(.x,0)))},
                      # Aggregation to hourly data
                      hourly = {
                        Dataset %>%
                          data.frame() %>%
                          dplyr::mutate(Y = lubridate::year(.data$Date),
                                        M = lubridate::month(.data$Date),
                                        D = lubridate::day(.data$Date),
                                        H = lubridate::hour(.data$Date)) %>%
                          dplyr::group_by(.data$Y, .data$M, .data$D, .data$H,
                                          .data$NameStation, .data$IDStation) %>%
                          Custom_summarise(vv, fv) %>%
                          dplyr::ungroup() %>%
                          dplyr::mutate(Date = lubridate::make_datetime(year = .data$Y,
                                                                        month = .data$M,
                                                                        day = .data$D,
                                                                        hour = .data$H)) %>%
                          dplyr::select(-c(.data$Y,.data$M,.data$D,.data$H)) %>%
                          dplyr::relocate(.data$Date,.data$IDStation) %>%
                          dplyr::mutate(dplyr::across(tidyselect::vars_select_helpers$where(is.numeric),
                                                      ~ ifelse(is.nan(.), NA, .))) %>%
                          dplyr::mutate(dplyr::across(tidyselect::matches(c("Wind_direction","Wind_direction_max")), ~ round(.x,0)))})

  attr(data_aggr, "frequency") <- Frequency
  freq_unit <- dplyr::case_when(Frequency == "hourly" ~ "hours",
                                Frequency == "daily" ~ "days",
                                Frequency == "weekly" ~ "weeks",
                                Frequency == "monthly" ~ "months",
                                Frequency == "yearly" ~ "years")
  attr(data_aggr, "units") <- freq_unit

  structure(list(data_aggr = data_aggr))
  if (is_ARPALdf_AQ(Dataset)==TRUE) {
    attr(data_aggr, "class") <- c("ARPALdf","ARPALdf_AQ","tbl_df","tbl","data.frame")
  }
  if (is_ARPALdf_AQ_mun(Dataset)==TRUE) {
    attr(data_aggr, "class") <- c("ARPALdf","ARPALdf_AQ_mun","tbl_df","tbl","data.frame")
  }
  if (is_ARPALdf_W(Dataset)==TRUE) {
    attr(data_aggr, "class") <- c("ARPALdf","ARPALdf_W","tbl_df","tbl","data.frame")
  }

  return(data_aggr)
}

#' ARPALData Package
#'
#' Contains functions for downloading and managing air quality and weather data from the Regione Lombardia open database.
#'     Data are collected by ARPA Lombardia (Lombardia Environmental Protection Agency), Italy.
#'
#' @docType package
#'
#' @author Paolo Maranzano \email{pmaranzano.ricercastatistica@gmail.com}
#'
#' @name ARPALData
NULL

#' @keywords internal
#' @noRd

ARPAL_extdata_path <- function(filename) {
  ### Modified on 2026-06-26: retrieve package-bundled support files from
  ### inst/extdata instead of downloading them from the ARPALData GitHub
  ### repository at runtime. This keeps mapping and metadata helpers available
  ### offline once the package is installed.
  path <- system.file("extdata", filename, package = "ARPALData")

  if (!nzchar(path) || !file.exists(path)) {
    stop(
      "Required package file '", filename, "' was not found in inst/extdata. ",
      "Please reinstall ARPALData or report a bug.",
      call. = FALSE
    )
  }

  return(path)
}

#' Generate a map of summary statistics for a given ARPALdf data.frame
#'
#' @description 'ARPALdf_Summary_map' represents on a map (polygon of Lombardy) the data contained in a data frame
#' of class 'ARPALdf' containing the values or the descriptive statistics by station. Data can be either
#' a ARPALdf of observed data (from 'get_ARPA_Lombardia_xxx' commands) and an ARPALdf obtained as summary descriptive
#' statistic (from 'ARPALdf_Summary' command).
#'
#' @param Data Dataset of class 'ARPALdf' containing the values or the descriptive statistics to plot on the map.
#' Data can be either a ARPALdf of observed data (from 'get_ARPA_Lombardia_xxx' commands) and an ARPALdf obtained
#' as summary descriptive statistic (from 'ARPALdf_Summary' command).
#' @param Title_main Title of the plot.
#' @param Title_legend Title fo the legend
#' @param Variable Summary variable to represent
#' @param prov_line_type Linetype for Lombardy provinces. Default is 1.
#' @param prov_line_size Size of the line for Lombardy provinces. Default is 1.
#' @param col_scale Vector indicating the minimum, the middle and the average point colors.
#' Default is c("green","yellow","red").
#' @param val_midpoint Numeric. Value associated to the middle-point scale color.
#' Default is NULL (midpoint is set equal to the average of the variable to represent).
#' @param xlab x-axis label. Default is 'Longitude'.
#' @param ylab y-axis label. Default is 'Latitude'.
#'
#' @return A map of selected stations across the Lombardy region
#'
#' @examples
#' \donttest{
#' ## Download daily air quality data from all the stations for January 2020
#' d <- get_ARPA_Lombardia_AQ_data(ID_station = NULL, Date_begin = "2020-01-01",
#'         Date_end = "2020-01-31", Frequency = "daily")
#' ## Summarising observed data
#' s <- ARPALdf_Summary(Data = d)
#' ## Mapping of the average NO2 in January 2020 at several stations
#' ARPALdf_Summary_map(Data = s$Descr_by_IDStat$Mean_by_stat,
#'         Title_main = "Mean NO2 by station in January 2020", Variable = "NO2")
#' }
#'
#' @export

ARPALdf_Summary_map <- function(Data, Title_main, Title_legend = "Variable", Variable,
                                prov_line_type = 1, prov_line_size = 1,
                                col_scale = c("#00FF00","#FFFF00","#FF0000"),
                                val_midpoint = NULL,xlab = "Longitude", ylab = "Latitude") {

  ### Checks
  stopifnot("Data is not of class ARPALdf or it does not contains the column IDStation" = is_ARPALdf(Data = Data) == T &
              sum(colnames(Data) == "IDStation") == 1)

  if (is_ARPALdf_AQ_mun(Data = Data) == T) {
    Stats <- get_ARPA_Lombardia_AQ_municipal_registry()
    NUTS_level <- "LAU"
  } else {
    NUTS_level <- "NUTS3"
  }
  Lombardia <- get_Lombardia_geospatial(NUTS_level)
  if (is.null(Lombardia)) {
    message("The map will not include the ground layer with Lombardy's shapefile. Only points/coordinates will be plot.")
  }

  if (is_ARPALdf_AQ(Data = Data) == T) {
    Stats <- get_ARPA_Lombardia_AQ_registry()
  } else if (is_ARPALdf_W(Data = Data) == T) {
    Stats <- get_ARPA_Lombardia_W_registry()
  }


  if (is_ARPALdf_AQ_mun(Data = Data) == T) {
    Data$var <- as.numeric(dplyr::pull(Data[,Variable]))
    ### Scale color midpoint
    if (is.null(val_midpoint)==T) {
      val_midpoint <- mean(Data$var,na.rm=T)
    }
    Data <- dplyr::left_join(Data,Lombardia,by=c("NameStation"="City"))
    Data <- Data %>%
      sf::st_as_sf()

    geo_plot <- Data %>%
      ggplot2::ggplot() +
      ggplot2::geom_sf(aes(fill = .data$var)) +
      ggplot2::scale_fill_gradient2(Title_legend, na.value = NA,
                                    low = col_scale[1],
                                    mid = col_scale[2],
                                    midpoint = val_midpoint,
                                    high = col_scale[3]) +
      ggplot2::guides(size = FALSE, scale = "none") +
      ggplot2::labs(title = Title_main) +
      ggplot2::theme_bw() +
      ggplot2::scale_x_continuous(labels = function(x) paste0(x, '\u00B0', "E")) +
      ggplot2::scale_y_continuous(labels = function(x) paste0(x, '\u00B0', "N"))

  } else {
    Stats <- Stats %>%
      dplyr::filter(.data$IDStation %in% unique(Data$IDStation)) %>%
      dplyr::distinct(.data$IDStation,.data$Latitude,.data$Longitude)
    Data <- dplyr::left_join(Data,Stats,by=c("IDStation"))
    Data$var <- as.numeric(dplyr::pull(Data[,Variable]))
    ### Scale color midpoint
    if (is.null(val_midpoint)==T) {
      val_midpoint <- mean(Data$var,na.rm=T)
    }
    Data <- Data %>%
      sf::st_as_sf(coords = c("Longitude", "Latitude"),crs = 4326)

    geo_plot <- Lombardia %>%
      ggplot2::ggplot() +
      ggplot2::geom_sf(linetype = prov_line_type, size = prov_line_size) +
      ggplot2::geom_sf(data = Data, aes(size = .data$var, col = .data$var)) +
      ggplot2::scale_color_gradient2(Title_legend, na.value = NA,
                                    low = col_scale[1],
                                    mid = col_scale[2],
                                    midpoint = val_midpoint,
                                    high = col_scale[3]) +
      ggplot2::guides(size = FALSE, scale = "none") +
      ggplot2::labs(title = Title_main) +
      ggplot2::theme_bw() +
      ggplot2::scale_x_continuous(labels = function(x) paste0(x, '\u00B0', "E")) +
      ggplot2::scale_y_continuous(labels = function(x) paste0(x, '\u00B0', "N"))
  }

  print(geo_plot)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/map_Lombardia_stations.R
\name{map_Lombardia_stations}
\alias{map_Lombardia_stations}
\title{Generate a map of the selected stations}
\usage{
map_Lombardia_stations(
  data,
  title = "Map of ARPA stations in Lombardy",
  prov_line_type = 1,
  prov_line_size = 1,
  col_points = "blue",
  xlab = "Longitude",
  ylab = "Latitude"
)
}
\arguments{
\item{data}{Dataset of class 'ARPALdf' containing the stations to plot on the map. It can be either a
ARPALdf of observed data (from 'get_ARPA_Lombardia_xxx' commands) and an ARPALdf obtained as registry
(from 'get_ARPA_Lombardia_xxx_registry' command).}

\item{title}{Title of the plot. Deafult is 'Map of ARPA stations in Lombardy'}

\item{prov_line_type}{Linetype for Lombardy provinces. Default is 1.}

\item{prov_line_size}{Size of the line for Lombardy provinces. Default is 1.}

\item{col_points}{Color of the points. Default is 'blue'.}

\item{xlab}{x-axis label. Default is 'Longitude'.}

\item{ylab}{y-axis label. Default is 'Latitude'.}
}
\value{
A map of selected stations across the Lombardy region
}
\description{
'get_ARPA_Lombardia_AQ_data' represents on a map (geometries/polygon of Lombardy) the location of
the stations contained in a data frame of class 'ARPALdf'. Data can be either a ARPALdf of observed data
(from 'get_ARPA_Lombardia_xxx' commands) and an ARPALdf obtained as registry
(from 'get_ARPA_Lombardia_xxx_registry' command).
}
\examples{
\donttest{
## Map network from a dataset of measurements
# Download daily concentrations observed at all the stations in January 2020.
d <- get_ARPA_Lombardia_AQ_data(ID_station = NULL, Date_begin = "2020-01-01",
                                Date_end = "2020-01-31", Frequency = "daily")
# Map the stations included in 'd'
map_Lombardia_stations(data = d, title = "Air quality stations in Lombardy")
## Map network from a registry dataset
# Download registry for all the AQ stations in 2020.
r <- get_ARPA_Lombardia_AQ_registry()
# Map the stations included in 'r'
map_Lombardia_stations(data = r, title = "Air quality stations in Lombardy")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Time_aggregate.R
\name{Time_aggregate}
\alias{Time_aggregate}
\title{Aggregate any ARPALdf object (with higher temporal frequency) to hourly, daily, weekly, monthly and yearly
temporal frequencies.}
\usage{
Time_aggregate(Dataset, Frequency, Var_vec = NULL, Fns_vec = NULL, verbose = T)
}
\arguments{
\item{Dataset}{ARPALdf dataframe to aggregate.}

\item{Frequency}{Temporal aggregation frequency. It can be "hourly", "daily", "weekly",
"monthly" or "yearly.}

\item{Var_vec}{Vector of variables to aggregate. If NULL (default) all the variables are averaged,
expect for 'Temperature' and 'Snow_height' which are summed.}

\item{Fns_vec}{Vector of aggregation functions to apply to the selected variables. Available functions
are 'mean', 'median', 'min', 'max', 'sum', 'qPP' (PP-th percentile), 'sd', 'var', 'vc' (variability coefficient),
'skew' (skewness) and 'kurt' (kurtosis). Attention: for Wind Speed and Wind Speed Gust only mean, min
and max are available; for Wind Direction and Wind Direction Gust only mean is available.}

\item{verbose}{Logical value (TRUE or FALSE). Toggle warnings and messages. If 'verbose=T' (default) the function
prints on the screen some messages describing the progress of the tasks. If 'verbose=F' any message about
the progression is suppressed.}
}
\value{
A data frame
}
\description{
Starting from an ARPALdf object with high frequency (e.g., 10mins or hourly), 'Time_aggregate'
aggregates the dataset to lower temporal frequencies (e.g., hourly, daily, weekly, monthly and yearly) by station.
The output is an ARPALdf object with observations having hourly, daily, weekly, monthly or yearly frequency.
The function can be applied only to ARPALdf objects.
User can indicate specific variables to aggregate and an aggregation function among
mean, median, sum (cumulated), min, max, quantiles, and variability metrics for each variable.
It is possible to specify different aggregation functions on the same variable
by repeating the name of the variable in 'Var_vec' and specifying the functions in 'Fns_vec'.
}
\examples{
\donttest{
### Modified on 2026-06-15: reduce Time_aggregate examples to January 2020
### to keep CRAN checks lightweight while preserving the same workflow.
## Download hourly observed concentrations during January 2020 for station 501
data <- get_ARPA_Lombardia_AQ_data(ID_station=501, Date_begin = "2020-01-01",
                                   Date_end = "2020-01-31", Frequency="hourly")
## Aggregate all the data to daily frequency
Time_aggregate(Dataset=data,Frequency="daily",Var_vec=NULL,Fns_vec=NULL)
## Aggregate NO2 to weekly maximum concentrations and NOx to weekly minimum concentrations.
Time_aggregate(Dataset=data,Frequency="weekly",Var_vec=c("NO2","NOx"),Fns_vec=c("max","min"))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is_ARPALdf.R
\name{is_ARPALdf}
\alias{is_ARPALdf}
\title{Check if a given object belongs to class 'ARPALdf'}
\usage{
is_ARPALdf(Data)
}
\arguments{
\item{Data}{Object to check if the class of a dataframe is 'ARPALdf', i.e. ARPAL dataframe.}
}
\value{
The function returns 'True' if the object is of class 'ARPALdf' and it returns 'False' if the
    object isn't of class 'ARPALdf'
}
\description{
'is_ARPALdf' checks if the input object belongs to the class 'ARPALdf'
}
\examples{
d <- get_ARPA_Lombardia_AQ_registry()
is_ARPALdf(d)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_ARPA_Lombardia_zoning.R
\name{get_ARPA_Lombardia_zoning}
\alias{get_ARPA_Lombardia_zoning}
\title{Download ARPA Lombardia zoning geometries}
\usage{
get_ARPA_Lombardia_zoning(
  plot_map = TRUE,
  title = "ARPA Lombardia zoning",
  line_type = 1,
  line_size = 1,
  xlab = "Longitude",
  ylab = "Latitude"
)
}
\arguments{
\item{plot_map}{Logical value (FALSE or TRUE). If plot_map = TRUE, the ARPA Lombardia zoning is represented
on a map, if plot_mat = FALSE only the geometry (polygon shapefile) is stored in the output.
Default is plot_map = TRUE.}

\item{title}{Title of the plot. Deafult is 'ARPA Lombardia zoning'}

\item{line_type}{Linetype for the zones' borders. Default is 1.}

\item{line_size}{Size of the line for the zones. Default is 1.}

\item{xlab}{x-axis label. Default is 'Longitude'.}

\item{ylab}{y-axis label. Default is 'Latitude'.}
}
\value{
The function returns an object of class 'sf' containing the polygon borders of the seven zones used by
ARPA Lombardia to classify the regional territory. If plot_map = 1, it also returns a map of the zoning.
}
\description{
'get_ARPA_Lombardia_zoning' returns the geometries (polygonal shape file) and a map of
the ARPA zoning of Lombardy. The zoning reflects the main orographic characteristics of the territory.
Lombardy region is classified into seven type of areas: large urbanized areas, urbanized areas in rural
contexts, rural areas, mountainous areas and valley bottom.
For more information, visit the section 'Zonizzazione ARPA Lombardia' at the websites
https://www.arpalombardia.it/temi-ambientali/aria/rete-di-rilevamento/classificazione-zone/ and
https://www.arpalombardia.it/temi-ambientali/aria/mappa-della-zonizzazione/
}
\examples{
zones <- get_ARPA_Lombardia_zoning(plot_map = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_ARPA_Lombardia_W_registry.R
\name{get_ARPA_Lombardia_W_registry}
\alias{get_ARPA_Lombardia_W_registry}
\title{Download metadata (registry) on weather monitoring stations from ARPA Lombardia website}
\usage{
get_ARPA_Lombardia_W_registry()
}
\value{
A data frame of class 'data.frame' and 'ARPALdf'. The object is fully compatible with Tidyverse.
}
\description{
'get_ARPA_Lombardia_W_registry' returns the registry (list) of all the weather sensors
and stations belonging to the ARPA Lombardia network. The information reported are: ID of each sensor and
station, geo-location (coordinates in degrees), altitude (mt), starting date and ending date.
The column 'NameStation' identifies the name of each station, while 'IDStation' is an ID code (assigned
from ARPA) uniquely identifying each station.
For more information, visit the section 'Meteo' at the websites:
https://www.dati.lombardia.it/stories/s/auv9-c2sj and
https://www.dati.lombardia.it/Ambiente/Stazioni-Meteorologiche/nf78-nj6b
}
\examples{
get_ARPA_Lombardia_W_registry()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ARPALdf_Summary_map.R
\name{ARPALdf_Summary_map}
\alias{ARPALdf_Summary_map}
\title{Generate a map of summary statistics for a given ARPALdf data.frame}
\usage{
ARPALdf_Summary_map(
  Data,
  Title_main,
  Title_legend = "Variable",
  Variable,
  prov_line_type = 1,
  prov_line_size = 1,
  col_scale = c("#00FF00", "#FFFF00", "#FF0000"),
  val_midpoint = NULL,
  xlab = "Longitude",
  ylab = "Latitude"
)
}
\arguments{
\item{Data}{Dataset of class 'ARPALdf' containing the values or the descriptive statistics to plot on the map.
Data can be either a ARPALdf of observed data (from 'get_ARPA_Lombardia_xxx' commands) and an ARPALdf obtained
as summary descriptive statistic (from 'ARPALdf_Summary' command).}

\item{Title_main}{Title of the plot.}

\item{Title_legend}{Title fo the legend}

\item{Variable}{Summary variable to represent}

\item{prov_line_type}{Linetype for Lombardy provinces. Default is 1.}

\item{prov_line_size}{Size of the line for Lombardy provinces. Default is 1.}

\item{col_scale}{Vector indicating the minimum, the middle and the average point colors.
Default is c("green","yellow","red").}

\item{val_midpoint}{Numeric. Value associated to the middle-point scale color.
Default is NULL (midpoint is set equal to the average of the variable to represent).}

\item{xlab}{x-axis label. Default is 'Longitude'.}

\item{ylab}{y-axis label. Default is 'Latitude'.}
}
\value{
A map of selected stations across the Lombardy region
}
\description{
'ARPALdf_Summary_map' represents on a map (polygon of Lombardy) the data contained in a data frame
of class 'ARPALdf' containing the values or the descriptive statistics by station. Data can be either
a ARPALdf of observed data (from 'get_ARPA_Lombardia_xxx' commands) and an ARPALdf obtained as summary descriptive
statistic (from 'ARPALdf_Summary' command).
}
\examples{
\donttest{
## Download daily air quality data from all the stations for January 2020
d <- get_ARPA_Lombardia_AQ_data(ID_station = NULL, Date_begin = "2020-01-01",
        Date_end = "2020-01-31", Frequency = "daily")
## Summarising observed data
s <- ARPALdf_Summary(Data = d)
## Mapping of the average NO2 in January 2020 at several stations
ARPALdf_Summary_map(Data = s$Descr_by_IDStat$Mean_by_stat,
        Title_main = "Mean NO2 by station in January 2020", Variable = "NO2")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is_ARPALdf_W.R
\name{is_ARPALdf_W}
\alias{is_ARPALdf_W}
\title{Check if a given object belongs to class 'ARPALdf_W'}
\usage{
is_ARPALdf_W(Data)
}
\arguments{
\item{Data}{Object to check if the class of a dataframe is 'ARPALdf_W', i.e. ARPAL
dataframe for weather data.}
}
\value{
The function returns 'True' if the object is of class 'ARPALdf_W' and it returns 'False' if the
    object isn't of class 'ARPALdf_W'
}
\description{
'is_ARPALdf_W' checks if the input object belongs to the class 'ARPALdf_W'
}
\examples{
d <- get_ARPA_Lombardia_W_registry()
is_ARPALdf_W(d)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_ARPA_Lombardia_AQ_data.R
\name{get_ARPA_Lombardia_AQ_data}
\alias{get_ARPA_Lombardia_AQ_data}
\title{Download air quality data from ARPA Lombardia website}
\usage{
get_ARPA_Lombardia_AQ_data(
  ID_station = NULL,
  Date_begin = "2022-01-01",
  Date_end = "2022-12-31",
  Frequency = "hourly",
  Var_vec = NULL,
  Fns_vec = NULL,
  by_sensor = FALSE,
  verbose = TRUE,
  parallel = FALSE,
  parworkers = NULL,
  parfuturetype = "multisession"
)
}
\arguments{
\item{ID_station}{Numeric value. ID of the station to consider. Using ID_station = NULL, all the available
stations are selected. Default is ID_station = NULL.}

\item{Date_begin}{Character vector of the first date-time to download. Format can be either "YYYY-MM-DD" or "YYYY-MM-DD hh:mm:ss". Default is Date_begin = "2022-01-01".}

\item{Date_end}{Character vector of the last date-time to download. Format can be either "YYYY-MM-DD" or "YYYY-MM-DD hh:mm:ss". Default is Date_end = "2022-12-31".}

\item{Frequency}{Temporal aggregation frequency. It can be "hourly", "daily", "weekly", "monthly" or "yearly".
Default is Frequency = "hourly".}

\item{Var_vec}{Character vector of variables to aggregate. If NULL (default) all the variables are averaged.}

\item{Fns_vec}{Character vector of aggregation function to apply to the selected variables.
Available functions are mean, median, min, max, sum, qPP (PP-th percentile), sd, var,
vc (variability coefficient), skew (skewness) and kurt (kurtosis).}

\item{by_sensor}{Logical value (TRUE or FALSE). If 'by_sensor = TRUE', the function returns the observed concentrations
by sensor code, while if 'by_sensor = FALSE' (default) it returns the observed concentrations by station.}

\item{verbose}{Logical value (TRUE or FALSE). Toggle warnings and messages. If 'verbose = TRUE' (default) the function
prints on the screen some messages describing the progress of the tasks. If 'verbose = FALSE' any message about
the progression is suppressed.}

\item{parallel}{Logical value (TRUE or FALSE). If 'parallel = FALSE' (default), data downloading is performed using a sequential/serial approach and additional parameters 'parworkers' and 'parfuturetype' are ignored.
When 'parallel = TRUE', data downloading is performed using parallel computing through the Futureverse setting.
More detailed information about parallel computing in the Futureverse is available at the following websites:
https://future.futureverse.org/ and https://cran.r-project.org/web/packages/future.apply/vignettes/future.apply-1-overview.html}

\item{parworkers}{Numeric integer value. If 'parallel = TRUE' (parallel mode active), the user can declare the number of parallel workers to be activated using 'parworkers = integer number'. By default ('parworkers = NULL'), the number of active workers is half of the available local cores.}

\item{parfuturetype}{Character vector. If 'parallel = TRUE' (parallel mode active), the user can declare the parallel strategy to be used according to the Futureverse syntax through 'parfuturetype'. By default, the 'multisession' (background R sessions on local machine) is used. In alternative, the 'multicore' (forked R processes on local machine. Not supported by Windows and RStudio) setting can be used.}
}
\value{
A data frame of class 'data.frame' and 'ARPALdf'. The object is fully compatible with Tidyverse.
}
\description{
'get_ARPA_Lombardia_AQ_data' returns observed air quality measurements collected by
ARPA Lombardia ground detection system for Lombardy region in Northern Italy.
Available airborne pollutant concentrations are: NO2, NOx, PM10, PM2.5, Ozone, Arsenic, Benzene,
Benzo-a-pirene, Ammonia, Sulfur Dioxide, Black Carbon, CO, Nikel, Cadmium and Lead.
Data are available from 1968 and are updated up to the current date.
For more information, visit the section 'Monitoraggio aria' at the website:
https://www.dati.lombardia.it/stories/s/auv9-c2sj
}
\examples{
\donttest{
## Download hourly air quality data for 2024 at station 501.
get_ARPA_Lombardia_AQ_data(ID_station=501, Date_begin = "2024-01-01",
                           Date_end = "2024-12-31", Frequency="hourly", parallel = TRUE)
## Download (parallel) monthly data for NOx and NO2 observed between May and
## August 2021 for all the stations active on the network. For NOx is computed
## the 25th percentile, while for NO2 is computed the maximum concentration observed.
get_ARPA_Lombardia_AQ_data(ID_station=NULL,Date_begin = "2021-05-01",
                           Date_end = "2021-08-01", Frequency="monthly",Var_vec=c("NOx","NO2"),
                           Fns_vec=c("q25","max"), parallel = TRUE)
## Download hourly air quality data by sensor from February to June 2026 at station 501.
get_ARPA_Lombardia_AQ_data(ID_station=501,Date_begin = "2026-02-01 00:00:00",
                           Date_end = "2026-06-01 23:00:00", by_sensor = TRUE)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_ARPA_Lombardia_W_data.R
\name{get_ARPA_Lombardia_W_data}
\alias{get_ARPA_Lombardia_W_data}
\title{Download weather/meteorological data from ARPA Lombardia website}
\usage{
get_ARPA_Lombardia_W_data(
  ID_station = NULL,
  Date_begin = "2021-01-01",
  Date_end = "2022-12-31",
  Frequency = "10mins",
  Var_vec = NULL,
  Fns_vec = NULL,
  by_sensor = FALSE,
  verbose = TRUE,
  parallel = FALSE,
  parworkers = NULL,
  parfuturetype = "multisession"
)
}
\arguments{
\item{ID_station}{Numeric value. ID of the station to consider. Using ID_station = NULL, all the available
stations are selected. Default is ID_station = NULL.}

\item{Date_begin}{Character vector of the first date-time to download. Format can be either "YYYY-MM-DD" or "YYYY-MM-DD hh:mm:ss". Default is Date_begin = "2022-01-01".}

\item{Date_end}{Character vector of the last date-time to download. Format can be either "YYYY-MM-DD" or "YYYY-MM-DD hh:mm:ss". Default is Date_end = "2022-12-31".}

\item{Frequency}{Temporal aggregation frequency. It can be "10mins", "hourly", "daily", "weekly",
"monthly". Default is Frequency = "10mins"}

\item{Var_vec}{Character vector of variables to aggregate. If NULL (default) all the variables are averaged,
except for 'Temperature' and 'Snow_height', which are cumulated.}

\item{Fns_vec}{Character vector of aggregation function to apply to the selected variables.
Available functions are mean, median, min, max, sum, qPP (PP-th percentile), sd, var,
vc (variability coefficient), skew (skewness) and kurt (kurtosis). Attention: for Wind Speed and
Wind Speed Gust only mean, min and max are available; for Wind Direction and Wind Direction Gust
only mean is available.}

\item{by_sensor}{Logical value (TRUE or FALSE). If 'by_sensor = TRUE', the function returns the observed concentrations
by sensor code, while if 'by_sensor = FALSE' (default) it returns the observed concentrations by station.}

\item{verbose}{Logical value (TRUE or FALSE). Toggle warnings and messages. If 'verbose = TRUE' (default) the function
prints on the screen some messages describing the progress of the tasks. If 'verbose = FALSE' any message about
the progression is suppressed.}

\item{parallel}{Logical value (TRUE or FALSE). If 'parallel = FALSE' (default), data downloading is performed using a sequential/serial approach and additional parameters 'parworkers' and 'parfuturetype' are ignored.
When 'parallel = TRUE', data downloading is performed using parallel computing through the Futureverse setting.
More detailed information about parallel computing in the Futureverse is available at the following websites:
https://future.futureverse.org/ and https://cran.r-project.org/web/packages/future.apply/vignettes/future.apply-1-overview.html}

\item{parworkers}{Numeric integer value. If 'parallel = TRUE' (parallel mode active), the user can declare the number of parallel workers to be activated using 'parworkers = integer number'. By default ('parworkers = NULL'), the number of active workers is half of the available local cores.}

\item{parfuturetype}{Character vector. If 'parallel = TRUE' (parallel mode active), the user can declare the parallel strategy to be used according to the Futureverse syntax through 'parfuturetype'. By default, the 'multisession' (background R sessions on local machine) is used. In alternative, the 'multicore' (forked R processes on local machine. Not supported by Windows and RStudio) setting can be used.}
}
\value{
A data frame of class 'data.frame' and 'ARPALdf'. The object is fully compatible with Tidyverse.
}
\description{
'get_ARPA_Lombardia_W_data' returns observed air weather measurements collected by
ARPA Lombardia ground detection system for Lombardy region in Northern Italy.
Available meteorological variables are: temperature (Celsius degrees), rainfall (mm), wind speed (m/s),
wind direction (degrees), relative humidity (%), global solar radiation (W/m2), and snow height (cm).
Data are available from 1989 and are updated up to the current date.
For more information, visit the section 'Idro-Nivo-Meteo' at the website:
https://www.dati.lombardia.it/stories/s/auv9-c2sj
}
\details{
More detailed description.
}
\examples{
\donttest{
## Download all the (10 minutes frequency) weather measurements at station 100
## between August 2022 and April 2024.
get_ARPA_Lombardia_W_data(ID_station = 100, Date_begin = "2022-08-01",
                          Date_end = "2024-04-30", Frequency = "10mins")
## Download all the (daily frequency) weather measurements at station 1974 during 2025 and 2026
get_ARPA_Lombardia_W_data(ID_station = 1974, Date_begin = "2025-01-01",
                          Date_end = "2026-12-31", Frequency = "daily")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_ARPA_Lombardia_AQ_municipal_registry.R
\name{get_ARPA_Lombardia_AQ_municipal_registry}
\alias{get_ARPA_Lombardia_AQ_municipal_registry}
\title{Download metadata (registry) on air quality monitoring stations at municipal level from ARPA Lombardia website}
\usage{
get_ARPA_Lombardia_AQ_municipal_registry()
}
\value{
A data frame of class 'data.frame' and 'ARPALdf'. The object is fully compatible with Tidyverse.
}
\description{
'get_ARPA_Lombardia_AQ_municipal_registry' returns the registry (list) of all the air quality sensors
owned by ARPA Lombardia for each municipality of Lombardy. The information reported are: ID of each sensor and
station, starting date and ending date. The column 'NameStation' identifies the name of each municipality.
The column 'IDStation' is an ID code (assigned from ARPA) uniquely identifying each municipality.
For more information, visit the section 'Stime comunali sull'aria' at the website:
https://www.dati.lombardia.it/stories/s/auv9-c2sj
}
\examples{
get_ARPA_Lombardia_AQ_municipal_registry()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/registry_KNN_dist.R
\name{registry_KNN_dist}
\alias{registry_KNN_dist}
\title{Identifies the K-nearest neighbors (stations) to all the monitoring sites included in a given ARPALdf
registry data.frame. The neighbors are identified computing the Euclidean distance among the sites'
coordinates.}
\usage{
registry_KNN_dist(reg_X, reg_Y, k = 1)
}
\arguments{
\item{reg_X}{Dataset of class 'ARPALdf' containing the stations list obtained as registry
(from 'get_ARPA_Lombardia_xxx_registry' command). The object must contain the following
colums: 'IDStation','NameStation','Longitude' and 'Latitude'.}

\item{reg_Y}{Dataset of class 'ARPALdf' containing the stations list obtained as registry
(from 'get_ARPA_Lombardia_xxx_registry' command). The object must contain the following
colums: 'IDStation','NameStation','Longitude' and 'Latitude'.}

\item{k}{Integer value. Represents the number of neighbors the user wants to identify.}
}
\value{
A data.frame object having the same length of reg_X. For each row (stations in reg_X)
it contains the name and the IDStation code for the k-nearest neighbors.
}
\description{
For each element included in reg_X, it identifies the k-nearest neighboring locations
(among those included in reg_Y) according to an Euclidean distance metric. reg_X and reg_Y
must be two 'ARPALdf' objects obtained using get_ARPA_Lombardia_xxx_registry'.
}
\examples{
if (require("tidyverse")) {
    regAQ <- get_ARPA_Lombardia_AQ_registry()
    regAQ <- regAQ \%>\% filter(Pollutant \%in\% c("PM10","Ammonia"))
    regW <- get_ARPA_Lombardia_W_registry()
    registry_KNN_dist(regAQ,regW,k=2)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ARPALdf_Summary.R
\name{ARPALdf_Summary}
\alias{ARPALdf_Summary}
\title{Summary statistics for a data frame of class 'ARPALdf'}
\usage{
ARPALdf_Summary(
  Data,
  by_IDStat = TRUE,
  by_Year = TRUE,
  gap_length = TRUE,
  correlation = TRUE,
  histogram = FALSE,
  density = FALSE,
  outlier = FALSE,
  verbose = TRUE
)
}
\arguments{
\item{Data}{Dataset of class 'ARPALdf' containing the data to be summarised.}

\item{by_IDStat}{Logical value (TRUE or FALSE). Use TRUE (default) to compute summary statistics by Station ID.}

\item{by_Year}{Logical value (TRUE or FALSE). Use TRUE (default) to compute summary statistics by year.}

\item{gap_length}{Logical value (TRUE or FALSE). Use TRUE (default) to compute summary statistics for the gap length of each variable.}

\item{correlation}{Logical value (TRUE or FALSE). Use TRUE (default) to compute linear correlation of available variables.}

\item{histogram}{Logical value (TRUE or FALSE). Use TRUE to plot the histogram of each variable. Default is FALSE.}

\item{density}{Logical value (TRUE or FALSE). Use TRUE to plot the kernel density plot of each variable. Default is FALSE.}

\item{outlier}{Logical value (TRUE or FALSE). Use TRUE to analyze extreme values of each variable
(boxplot and Hampel filter). Default is FALSE.}

\item{verbose}{Logical value (TRUE or FALSE). Toggle warnings and messages. If 'verbose = TRUE' (default) the function
prints on the screen some messages describing the progress of the tasks. If 'verbose = FALSE' any message about
the progression is suppressed.}
}
\value{
A list of data.frames containing summary descriptive statistics for a data frame of class 'ARPALdf'.
Summary statistics are computed for the overall sample (Descr), by Station ID (Descr_by_IDStat) and by
year (Descr_by_Year). Available statistics are: number of NAs, % of NAs over the total sample, number of null values,
number of negative values, minimum, mean, maximum and standard deviation.
}
\description{
'ARPALdf_Summary' returns many descriptive statistics summaring the data contained in a data frame
of class ARPALdf. Statistics are calculated at overall level (full sample), by station ID and by year.
For each variable are reported the basic positioning indices (min, max, mean, median, quantile) and
variability indices (range, standard deviation). Other reported statistics are the Pearson's linear correlation
by station and some graphical representation of the distribution (kernel density plot, histogram, boxplot).
In addition, the function returns useful data-quality information, such as gap length statistics (i.e. number of
missing observations for each variable by station and by year) and outlier detection tools
(e.g., Hampel filter and boxplot rule)
}
\examples{
\donttest{
## Download daily air quality data from all the stations for January 2020
d <- get_ARPA_Lombardia_AQ_data(ID_station = NULL, Date_begin = "2020-01-01",
        Date_end = "2020-01-31", Frequency = "daily")
## Summarising observed data
sum_stats <- ARPALdf_Summary(Data = d)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_Lombardia_geospatial.R
\name{get_Lombardia_geospatial}
\alias{get_Lombardia_geospatial}
\title{Download geospatial data of Lombardy from Eurostat}
\usage{
get_Lombardia_geospatial(NUTS_level = "LAU")
}
\arguments{
\item{NUTS_level}{NUTS level required: use "NUTS2" for regional geometries, "NUTS3" for provincial geometries,
or "LAU" for municipal geometries. Default NUTS_level = "LAU".}
}
\value{
A data frame of class 'data.frame', "sf" and 'ARPALdf'.
}
\description{
'get_Lombardia_geospatial' returns the polygonal (shape file) object containing the geometries
of Lombardy. Shapefile are available at different NUTS levels (https://ec.europa.eu/eurostat/web/nuts/background):
'LAU' for the shapefile of municipalities of Lombardy, 'NUTS3' for the shapefile of provinces of Lombardy
and 'NUTS2' for the shapefile of Lombardy.
}
\examples{
\donttest{
shape <- get_Lombardia_geospatial(NUTS_level = "LAU")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ARPALData.R
\docType{package}
\name{ARPALData}
\alias{ARPALData-package}
\alias{ARPALData}
\title{ARPALData Package}
\description{
Contains functions for downloading and managing air quality and weather data from the Regione Lombardia open database.
    Data are collected by ARPA Lombardia (Lombardia Environmental Protection Agency), Italy.
}
\author{
Paolo Maranzano \email{pmaranzano.ricercastatistica@gmail.com}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_ARPA_Lombardia_AQ_municipal_data.R
\name{get_ARPA_Lombardia_AQ_municipal_data}
\alias{get_ARPA_Lombardia_AQ_municipal_data}
\title{Download air quality data at manucipal from ARPA Lombardia website}
\usage{
get_ARPA_Lombardia_AQ_municipal_data(
  ID_station = NULL,
  Date_begin = "2021-01-01",
  Date_end = "2022-12-31",
  Frequency = "daily",
  Var_vec = NULL,
  Fns_vec = NULL,
  by_sensor = FALSE,
  verbose = TRUE,
  parallel = FALSE,
  parworkers = NULL,
  parfuturetype = "multisession"
)
}
\arguments{
\item{ID_station}{Numeric value. ID of the station to consider. Using ID_station = NULL, all the available
stations are selected. Default is ID_station = NULL.}

\item{Date_begin}{Character vector of the first date-time to download. Format can be either "YYYY-MM-DD" or "YYYY-MM-DD hh:mm:ss". Default is Date_begin = "2022-01-01".}

\item{Date_end}{Character vector of the last date-time to download. Format can be either "YYYY-MM-DD" or "YYYY-MM-DD hh:mm:ss". Default is Date_end = "2022-12-31".}

\item{Frequency}{Temporal aggregation frequency. It can be "daily", "weekly", "monthly" or "yearly".
Default is Frequency = "daily".}

\item{Var_vec}{Character vector of variables to aggregate. If NULL (default) all the variables are averaged.}

\item{Fns_vec}{Character vector of aggregation function to apply to the selected variables.
Available functions are mean, median, min, max, sum, qPP (PP-th percentile), sd, var,
vc (variability coefficient), skew (skewness) and kurt (kurtosis).}

\item{by_sensor}{Logical value (TRUE or FALSE). If 'by_sensor = TRUE', the function returns the observed concentrations by sensor code, while
if 'by_sensor = FALSE' (default) it returns the observed concentrations by station.}

\item{verbose}{Logical value (TRUE or FALSE). Toggle warnings and messages. If 'verbose = TRUE' (default) the function
prints on the screen some messages describing the progress of the tasks. If 'verbose = FALSE' any message about
the progression is suppressed.}

\item{parallel}{Logical value (TRUE or FALSE). If 'parallel = FALSE' (default), data downloading is performed using a sequential/serial approach and additional parameters 'parworkers' and 'parfuturetype' are ignored.
When 'parallel = TRUE', data downloading is performed using parallel computing through the Futureverse setting.
More detailed information about parallel computing in the Futureverse is available at the following websites:
https://future.futureverse.org/ and https://cran.r-project.org/web/packages/future.apply/vignettes/future.apply-1-overview.html}

\item{parworkers}{Numeric integer value. If 'parallel = TRUE' (parallel mode active), the user can declare the number of parallel workers to be activated using 'parworkers = integer number'. By default ('parworkers = NULL'), the number of active workers is half of the available local cores.}

\item{parfuturetype}{Character vector. If 'parallel = TRUE' (parallel mode active), the user can declare the parallel strategy to be used according to the Futureverse syntax through 'parfuturetype'. By default, the 'multisession' (background R sessions on local machine) is used. In alternative, the 'multicore' (forked R processes on local machine. Not supported by Windows and RStudio) setting can be used.}
}
\value{
A data frame of class 'data.frame' and 'ARPALdf'. The object is fully compatible with Tidyverse.
The column 'NameStation' identifies the name of each municipality. The column 'IDStation' is an ID code
(assigned from ARPA) uniquely identifying each municipality.
}
\description{
'get_ARPA_Lombardia_AQ_municipal_data' returns the air quality levels at municipal level estimated by
ARPA Lombardia using a physico-chemical model which simulates air quality based on weather and geo-physical
variables. For each municipality of Lombardy, ARPA estimates the average (NO2_mean) and maximum daily (NO2_max_day)
level of NO2, the daily maximum (Ozone_max_day) and the 8-hours moving window maximum (Ozone_max_8h) of Ozone
and the average levels of PM10 (PM10_mean) and PM2.5 (PM2.5_mean).
Data are available from 2011 and are updated up to the current date.
For more information, visit the section 'Stime comunali dell'aria' at the website:
https://www.dati.lombardia.it/stories/s/auv9-c2sj
}
\details{
More detailed description.
}
\examples{
\donttest{
### Modified on 2026-06-15: keep examples small to avoid stressing the
### public Open Data Lombardia API during CRAN checks.
## Download daily municipal concentrations in January 2025 for city number 100451.
get_ARPA_Lombardia_AQ_municipal_data(ID_station=100451,Date_begin = "2025-01-01",
                                     Date_end = "2025-01-07", Frequency="daily")
## Download monthly concentrations of NO2 (average and maximum) observed in early 2026
## at city number 100451.
get_ARPA_Lombardia_AQ_municipal_data(ID_station=100451,
                                     Date_begin = "2026-01-01", Date_end = "2026-03-31",
                                     Frequency="monthly",
                                     Var_vec=c("NO2_mean","NO2_mean"), Fns_vec=c("mean","max"))
## Download daily concentrations observed in March 2023 at city number 100451.
## Data are reported by sensor.
get_ARPA_Lombardia_AQ_municipal_data(ID_station=100451,
                                     Date_begin = "2023-03-01", Date_end = "2023-03-07",
                                     by_sensor = TRUE)
}

}
